#pragma once

#include <algorithm>
#include <dune/common/exceptions.hh>
#include <memory>
#include <utility>
#include <vector>

#include "arpack_geneo_wrapper.hh"
#include "raes/raes.hh"

#ifdef DUNE_FT_HAVE_FEAST
#include "feast_wrapper.hh"
#endif

#ifdef DUNE_FT_HAVE_SPECTRA
#include <Eigen/Eigen>
#include <Eigen/SparseCore>
#include <Spectra/MatOp/SparseSymMatProd.h>
#include <Spectra/MatOp/SymShiftInvert.h>
#include <Spectra/SymGEigsShiftSolver.h>
#endif

struct EigensolverParameters {
  std::string type;       // Either arpack, raes or feast (if feast is available) or spectra (if spectra is available)
  std::size_t nev_target; // How many eigenvectors should be computed ideally (in the adaptive setting, this is also the size of the initial search space)
  std::size_t nev_max;    // How many eigenvectors should be computed at most
  double tolerance;       // Error tolerance (might be handled slightly differently for each algorithm)
  double shift;           // In arpack, raes and spectra, the problem Ax=lBx is converted to (A + s*B)^-1 x = mx; the shift is s
  double a, b;            // The search interval (currently only required for feast)
  double threshold;       // If this is positive, only eigenvalues below threshold will be included (ignored in feast)
  double regularisation;  // regularisation parameter used in the random eigensolver; is added to the diagonal of A
};

template <class Mat, class Vec> class EigenSolver {
public:
  EigenSolver(const Mat &A, EigensolverParameters params) : params(params), eigenvectors(params.nev_target, Vec(A.N())), eigenvalues(params.nev_target) {}

  virtual std::pair<std::vector<Vec>, std::vector<double>> solve() = 0;
  virtual ~EigenSolver() = default;

protected:
  EigensolverParameters params;
  std::vector<Vec> eigenvectors;
  std::vector<double> eigenvalues;
};

template <class Mat, class Vec> class RAESWrapper : public EigenSolver<Mat, Vec> {
public:
  RAESWrapper(std::shared_ptr<Mat> A, std::shared_ptr<Mat> B, EigensolverParameters params, int verbose) : EigenSolver<Mat, Vec>(*A, params), A(A), B(B) {}

  virtual std::pair<std::vector<Vec>, std::vector<double>> solve() override {
    if (this->params.threshold > 0) {
      // In the adaptive setting, the eigensolver expects the eigenvectors and eigenvalues array to be of length "nev_max"
      this->eigenvalues.resize(this->params.nev_max);
      auto vec = this->eigenvectors[0];
      this->eigenvectors.resize(this->params.nev_max);
      std::fill(this->eigenvectors.begin(), this->eigenvectors.end(), vec);
      GeneralizedInverseAdaptive(*A, *B, this->params.shift, this->params.regularisation, this->params.tolerance, 100 * this->params.nev_target, this->eigenvalues, this->eigenvectors,
                                 this->params.threshold, this->params.nev_target, 0, 12345);
    } else
      GeneralizedInverse(*A, *B, this->params.shift, this->params.regularisation, this->params.tolerance, 100 * this->params.nev_target, this->params.nev_target, this->eigenvalues, this->eigenvectors,
                         0, 12345);

    return {this->eigenvectors, this->eigenvalues};
  }

private:
  std::shared_ptr<Mat> A, B;
};

#ifdef DUNE_FT_HAVE_FEAST
template <class Mat, class Vec> class FeastWrapper : public EigenSolver<Mat, Vec> {
public:
  FeastWrapper(std::shared_ptr<Mat> A, std::shared_ptr<Mat> B, EigensolverParameters params, int verbose) : EigenSolver<Mat, Vec>(*A, params), feast(*A, *B, verbose), verbose{verbose} {}

  virtual std::pair<std::vector<Vec>, std::vector<double>> solve() override {
    if (auto ret = feast.solve(this->eigenvectors, this->eigenvalues, this->params.nev_target, this->params.a, this->params.b, this->params.tolerance) != 0) {
      std::cout << "FEAST eigensolver failed, using constant function as eigenvector (Error code " << ret << ")\n";
      this->eigenvectors.resize(1);
      this->eigenvectors[0] = 1;
      this->eigenvalues.resize(1);
      this->eigenvalues[0] = 0;
    }

    return {this->eigenvectors, this->eigenvalues};
  }

private:
  Feast<Mat> feast;
  int verbose;
};
#endif

template <class Mat, class Vec> class ArpackWrapper : public EigenSolver<Mat, Vec> {
public:
  ArpackWrapper(std::shared_ptr<Mat> A, std::shared_ptr<Mat> B, EigensolverParameters params, int verbose) : EigenSolver<Mat, Vec>(*A, params), B(B), arpack(*A) {}

  virtual std::pair<std::vector<Vec>, std::vector<double>> solve() override {
    if (this->params.threshold > 0) {
      // In the adaptive setting, the arpack wrapper expects the eigenvectors and eigenvalues array to be of length "nev_max"
      this->eigenvalues.resize(this->params.nev_max);
      auto vec = this->eigenvectors[0];
      this->eigenvectors.resize(this->params.nev_max);
      std::fill(this->eigenvectors.begin(), this->eigenvectors.end(), vec);
      arpack.computeGenSymShiftInvertMinMagnitudeAdaptive(*B, this->params.tolerance, this->eigenvectors, this->eigenvalues, this->params.shift, this->params.threshold, this->params.nev_target);
    } else
      arpack.computeGenSymShiftInvertMinMagnitude(*B, this->params.tolerance, this->eigenvectors, this->eigenvalues, this->params.shift);

    return {this->eigenvectors, this->eigenvalues};
  }

private:
  std::shared_ptr<Mat> B;
  ArpackMLGeneo::ArPackPlusPlus_Algorithms<Mat, Vec> arpack;
};

#ifdef DUNE_FT_HAVE_SPECTRA
template <class Mat, class Vec> class SpectraWrapper : public EigenSolver<Mat, Vec> {
public:
  SpectraWrapper(std::shared_ptr<Mat> A, std::shared_ptr<Mat> B, EigensolverParameters params, int verbose)
      : EigenSolver<Mat, Vec>(*A, params), verbose(verbose), A_(A->N(), A->N()), B_(A->N(), A->N()) {
    using Triplet = Eigen::Triplet<double>;
    std::vector<Triplet> triplets;
    triplets.reserve(A->N() * 5);
    Dune::flatMatrixForEach(*A, [&](auto &&entry, std::size_t i, std::size_t j) {
      triplets.push_back({static_cast<int>(i), static_cast<int>(j), entry});
    });
    A_.setFromTriplets(triplets.begin(), triplets.end());

    triplets.clear();
    Dune::flatMatrixForEach(*B, [&](auto &&entry, std::size_t i, std::size_t j) {
      triplets.push_back({static_cast<int>(i), static_cast<int>(j), entry});
    });
    B_.setFromTriplets(triplets.begin(), triplets.end());

    A_.makeCompressed();
    B_.makeCompressed();
  }

  virtual std::pair<std::vector<Vec>, std::vector<double>> solve() override {
    using OpType = Spectra::SymShiftInvert<double, Eigen::Sparse, Eigen::Sparse>;
    using BOpType = Spectra::SparseSymMatProd<double>;
    using Matrix = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>;
    using Vector = Eigen::Matrix<double, Eigen::Dynamic, 1>;

    OpType op(A_, B_);
    BOpType Bop(B_);

    bool done = this->params.threshold < 0 ? true : false; // In the non-adaptive case, we're done after 1 iteration
    int curr_nev_target = this->params.nev_target;

    do {
      // TODO: Setting this up in each iteration might cause a new factorisation in each iteration
      Spectra::SymGEigsShiftSolver<OpType, BOpType, Spectra::GEigsMode::ShiftInvert> geigs(op, Bop, curr_nev_target, 2 * curr_nev_target, -1 * this->params.shift);
      geigs.init();

      // Find largest eigenvalue of the shifted problem (which corresponds to the smallest of the original problem)
      // with max. 1000 iterations and sort the resulting eigenvalues from small to large.
      int nconv = geigs.compute(Spectra::SortRule::LargestMagn, 1000, this->params.tolerance, Spectra::SortRule::SmallestAlge);

      if (geigs.info() == Spectra::CompInfo::Successful) {
        const auto evalues = geigs.eigenvalues();

        if (evalues[nconv - 1] >= this->params.threshold or curr_nev_target >= this->params.nev_max) {
          // Either we reached the threshold or we reached nev_max eigenvalues
          const auto evalues = geigs.eigenvalues();
          const auto evecs = geigs.eigenvectors();

          this->eigenvalues.resize(nconv);
          std::copy(evalues.begin(), evalues.end(), this->eigenvalues.begin());

          auto vec = this->eigenvectors[0];
          this->eigenvectors.resize(nconv);
          std::fill(this->eigenvectors.begin(), this->eigenvectors.end(), vec);

          for (int i = 0; i < nconv; ++i) {
            for (int j = 0; j < evecs.rows(); ++j) {
              this->eigenvectors[i][j] = evecs(j, i);
            }
          }

          done = true;
        }
      }

      curr_nev_target *= 1.6;
    } while (!done);

    return {this->eigenvectors, this->eigenvalues};
  }

private:
  int verbose;

  Eigen::SparseMatrix<double> A_, B_;
};
#endif
