#pragma once

#include <cmath>
#include <dune/common/exceptions.hh>
#include <dune/common/timer.hh>
#include <dune/istl/bccsmatrixinitializer.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/foreach.hh>

#include "matrixutilities.hh"

#include <algorithm>
#include <array>
#include <cassert>
#include <dune/istl/io.hh>

extern "C" {
#include <feast.h>
#include <feast_sparse.h>
}

template <class Mat> class Feast {
public:
  Feast(const Mat &A, const Mat &B, int verbose = 0) : verbose{verbose} {
    feastinit(fpm);

    if (A.N() != B.N() || A.N() != A.M())
      DUNE_THROW(Dune::Exception, "Matrices are not compatible or not square");

    Dune::Timer timer;
    timer.start();
    N = B.N(); // TODO: This assumes a block size of 1
    DuneToCSR(A, N, Ai, Aj, Av, false);
    DuneToCSR(B, N, Bi, Bj, Bv, false);
    timer.stop();
    if (verbose > 1)
      std::cout << "Matrix conversion for FEAST took: " << timer.elapsed() << "s\n";
  }

  // Returns FEAST error code
  template <class Vec> int solve(std::vector<Vec> &evecs, std::vector<double> &evals, int nev, double a, double b, double tolerance, int max_its = 20, int n_try = 1) {
    if (N != evecs[0].N())
      DUNE_THROW(Dune::Exception, "Mismatch between matrix and vectors");

    char uplo = 'F';
    double epsout;
    int found, loop, info;

    std::array<double, 2> emid{{(a + b) / 2, 0}};
    double radius = (b - a) / 2;

    if (verbose > 1)
      fpm[0] = 1; // Print info to screen
    fpm[2] = -std::log10(tolerance);   // error tolerance 10^-fpm[2]
    // fpm[17] = 30;  // ellipse ratio
    fpm[14] = 2;   // Compute only right eigenvectors
    fpm[3] = max_its;

    std::vector<double> x(2 * nev * N); // computed eigenvectors
    std::vector<double> e(2 * nev);     // computed eigenvalues
    std::vector<double> res(2 * nev);   // residuals

    // Define custom integration contours
    // We use a trapecoidal shape as the contour that looks like this and encloses the search interval.
    // We put more emphasis on the area around zero, hoping that this will increase the chance to find
    // the near zero eigenvalues more robustly.
    /*
            ^ Im
           _|_
          | | \_
          | |   \
      -----------|-------> Re
          | |  __/
          |_|_/
            |
     */
    int pieces = 4;
    double z_edge[2 * 4] = {
        a,      -2, // first point (in the complex plane)
        a,      2,  // second point
        b, 1,  // third point
        b, -1, // fourth point
    };
    int t_edge[4] = {0, 0, 0, 0}; // all contour pieces are lines
    int n_edge[4] = {8, 4, 4, 8}; // 8 integration points per line
    int nc = 0;
    for (auto n : n_edge) nc += n;
    fpm[7] = nc; // contour points

    std::vector<double> zne(2 * nc), wne(2 * nc);

    zfeast_customcontour(&nc, &pieces, n_edge, t_edge, z_edge, zne.data(), wne.data());
    dfeast_gcsrgvx(&N, Av.data(), Ai.data(), Aj.data(), Bv.data(), Bi.data(), Bj.data(), fpm, &epsout, &loop, emid.data(), &radius, &nev, e.data(), x.data(), &found, res.data(), &info, zne.data(), wne.data());

    if (info != 6 and info != 0){
      if (n_try < max_tries && info == 3) {
        // Search subspace was too small, try again with a larger one
        // TODO: Store the eigenvectors that were already computed
        if (verbose > 0) std::cout << "FEAST: Subspace too small, trying again\n";
        return solve(evecs, evals, 2 * nev, a, b, tolerance, max_its, n_try + 1);
      } else if (n_try < max_tries && info == 2) {
        // Eigensolver did not converge yet, try again with more iterations
        // TODO: Store the eigenvectors that were already computed
        if (n_try < 2) {
          // Sometimes the eigensolver reports error code 2 (not converged yet), even though the actual problem
          // is a too small subspace. So first try increasing the subspace (by a large factor, just be sure)
          // and only when that fails with the same error, increase the max iteration count.
          if (verbose > 0) std::cout << "FEAST: Eigensolver did not converge yet, trying again with larger subspace.\n";
          return solve(evecs, evals, 3 * nev, a, b, tolerance, max_its, n_try + 1);
        } else {
          // Just to be sure, we also increase the subspace size here by a bit
          if (verbose > 0) std::cout << "FEAST: Eigensolver did not converge yet, trying again with more iterations.\n";
          return solve(evecs, evals, 1.5 * nev, a, b, tolerance, 2 * max_its, n_try + 1);
        }
      } else {
        std::cout << "FEAST eigensolver did not succeed. Error code " << info << "\n";
        return info;
      }
    } else {
      nev = std::min(found, (int)evecs.size());
      std::vector<int> index(nev, 0);
      std::iota(index.begin(), index.end(), 0);
      std::sort(index.begin(), index.end(), [&](int a, int b) {
        return e[2 * a] < e[2 * b];
      });

      evals.resize(std::min(found, (int)evecs.size()));
      for (int i = 0; i < evals.size(); ++i) {
        evals[i] = e[2 * index[i]];
        for (int j = 0; j < N; ++j)
          evecs[i][j] = x[2 * j + 2 * index[i] * N];
      }
      return 0;
    }
  }

private:
  int fpm[64]; // feast parameters

  int verbose = 0;

  int N;
  double sigma = 1e-6;
  std::size_t p;

  std::vector<int> Ai;
  std::vector<int> Aj;
  std::vector<typename Mat::field_type> Av;

  std::vector<int> Bi;
  std::vector<int> Bj;
  std::vector<typename Mat::field_type> Bv;

  const int max_tries = 4;
};
