#ifndef Udune_ftworth_multilevel_geneo_preconditioner_HH
#define Udune_ftworth_multilevel_geneo_preconditioner_HH

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <queue>
#include <thread>
#include <mutex>

#include "subdomainutilities.hh"
#include "arpack_geneo_wrapper.hh"
#include "eigensolver.hh"
#include "raes/raes.hh"

//#define USE_CHOLMOD 1

#define MULTITHREADED 1

// #define HAVE_SUITESPARSE_UMFPACK 1
// #define HAVE_EIGEN 1

#if (HAVE_SUITESPARSE_UMFPACK && HAVE_EIGEN) || DOXYGEN

#include <dune/istl/umfpack.hh>
#include <dune/istl/cholmod.hh>
#include <Eigen/Dense>
#include <Eigen/Eigenvalues>

template <typename field_type>
std::vector<std::vector<field_type>>
get_partition_of_unity_standard(const std::vector<std::vector<size_t>> &local_to_global,
                                const std::vector<std::map<size_t, size_t>> &global_to_local,
                                const std::vector<std::vector<bool>> &boundary_dofs)
{
  // number of subdomains we have
  auto subdomains = local_to_global.size();

  // make a global vector holding the number of subdomains where dof is NOT on the boundary
  std::vector<int> k(global_to_local.size(), 0);
  for (unsigned p = 0; p < subdomains; p++)
    for (unsigned i = 0; i < local_to_global[p].size(); i++)
      if (!boundary_dofs[p][i])
        k[local_to_global[p][i]] += 1;

  // now compute partition of unity
  std::vector<std::vector<field_type>> pu(subdomains);
  for (unsigned p = 0; p < subdomains; p++)
  {
    pu[p].resize(local_to_global[p].size());
    for (unsigned i = 0; i < local_to_global[p].size(); i++)
      pu[p][i] = (boundary_dofs[p][i]) ? 0.0 : (((field_type)1.0) / k[local_to_global[p][i]]);
  }

  // return
  return pu;
}

template <typename MAT>
std::vector<std::vector<typename MAT::field_type>>
get_partition_of_unity_distance(const std::vector<std::shared_ptr<MAT>> &matrices,
                                const std::vector<std::vector<size_t>> &local_to_global,
                                const std::vector<std::map<size_t, size_t>> &global_to_local,
                                const std::vector<std::vector<bool>> &boundary_dofs,
                                size_t shrink)
{
  // types
  using field_type = typename MAT::field_type;

  // number of subdomains we have
  auto subdomains = local_to_global.size();

  // compute local distance from boundary in each subdomain
  const int maxdistance = 1 << 24;
  std::vector<std::vector<int>> localdistance(subdomains);
  for (unsigned p = 0; p < subdomains; p++)
  {
    // dof blocks in subdomain p
    auto n = local_to_global[p].size();

    // allocate subdomain vector and initialize with a distance never reached
    localdistance[p].resize(n, maxdistance);

    // initialize boundary degrees of freedom and interior degrees of freedom
    for (unsigned i = 0; i < n; i++)
    {
      if (boundary_dofs[p][i])
        localdistance[p][i] = 0; // boundary case
      if (global_to_local[local_to_global[p][i]].size() == 1)
        localdistance[p][i] = -1; // interior dof case
    }
    // non boundary and non interior dofs are now at maxdistance

    // get the subdomain matrix
    MAT &A = *matrices[p];

    // loop until nothing changes
    bool changed = true;
    while (changed)
    {
      // nothing changed
      changed = false;

      // update for local distance
      std::vector<int> update(localdistance[p]);

      // compute update
      for (unsigned i = 0; i < n; i++)
      {
        // this dof is interior, nothing changes
        if (localdistance[p][i] < 0)
          continue;

        // else we iterate over the row and compute minimum distance of neighbors and myself
        auto cIt = A[i].begin();
        auto cEndIt = A[i].end();
        int dist = maxdistance;
        for (; cIt != cEndIt; ++cIt)
          if (localdistance[p][cIt.index()] >= 0) // it is a neighbor (or myself) not in the interior
            dist = std::min(dist, localdistance[p][cIt.index()]);

        if (dist + 1 < localdistance[p][i])
        {
          update[i] = dist + 1;
          changed = true;
        }
      }

      if (changed)
        localdistance[p] = update;
    }

    // apply shrink to simulate oversampling, i.e. we subtract shrink from all the distances
    for (unsigned i = 0; i < n; i++)
      if (localdistance[p][i] >= 0) // skip interior
      {
        if (localdistance[p][i] > shrink)
          localdistance[p][i] -= shrink;
        else
          localdistance[p][i] = 0;
      }

    // now we have computed the distance function for the dofs in the overlap including the shrink
  }

  // now compute global distance
  std::vector<int> globaldistance(global_to_local.size(), 0);
  for (unsigned p = 0; p < subdomains; p++)
    for (unsigned i = 0; i < local_to_global[p].size(); i++)
      if (localdistance[p][i] >= 0)
        globaldistance[local_to_global[p][i]] += localdistance[p][i];

  // now compute partition of unity
  std::vector<std::vector<field_type>> pu(subdomains);
  for (unsigned p = 0; p < subdomains; p++)
  {
    pu[p].resize(local_to_global[p].size());
    for (unsigned i = 0; i < local_to_global[p].size(); i++)
    {
      if (localdistance[p][i] >= 0 && globaldistance[local_to_global[p][i]] == 0)
        std::cout << "compute PU zero global distance! p=" << p << " i=" << i << " g=" << local_to_global[p][i] << " local distance=" << localdistance[p][i] << " global distance=" << globaldistance[local_to_global[p][i]] << std::endl;
      if (localdistance[p][i] < 0)
        pu[p][i] = 1.0;
      else
        pu[p][i] = ((field_type)localdistance[p][i]) / globaldistance[local_to_global[p][i]];
    }
  }

  // return
  return pu;
}

template <typename field_type>
std::vector<std::vector<field_type>>
get_subdomain_boundary_mask(const std::vector<std::vector<bool>> &boundary_dofs)
{
  auto subdomains = boundary_dofs.size();
  std::vector<std::vector<field_type>> boundary_mask(subdomains);
  for (size_t p = 0; p < subdomains; p++)
  {
    auto n = boundary_dofs[p].size(); // dof blocks in subdomain p
    boundary_mask[p].resize(n);
    for (size_t i = 0; i < n; i++)
      boundary_mask[p][i] = (boundary_dofs[p][i]) ? 0.0 : 1.0;
  }
  return boundary_mask;
}

/** \brief Multi-level GenEO preconditioner
 *
 * \author Peter Bastian
 *
 * \tparam MAT a "Dune::BCRSMatrix<Dune::FieldMatrix>" type
 * \tparam VEC a "Dune::BlockVector<Dune::FieldVector>" type
 *
 */
template <typename MAT, typename VEC>
class MultiLevelGenEO : public Dune::Preconditioner<VEC, VEC>
{
public:
  //! \brief The domain type of the preconditioner.
  using domain_type = VEC;
  //! \brief The range type of the preconditioner.
  using range_type = VEC;
  //! \brief The field type of the preconditioner.
  using field_type = typename VEC::field_type;

  // define the category
  virtual Dune::SolverCategory::Category category() const
  {
    return Dune::SolverCategory::sequential;
  }

private:
  // types use din this class
  using VectorBlockType = typename VEC::block_type;
  using CBLOCKMAT = Dune::FieldMatrix<typename VEC::field_type, 1, 1>;
  using CBLOCKVEC = Dune::FieldVector<typename VEC::field_type, 1>;
  using DenseMat = Dune::DynamicMatrix<field_type>; // type for coarse snippet matrices
  using CMAT = Dune::BCRSMatrix<CBLOCKMAT>;         // type for coarse level matrix
  using CVEC = Dune::BlockVector<CBLOCKVEC>;        // type for coarse level vector
  using CHOLMODSOLVER = Dune::Cholmod<VEC>;
  using CCHOLMODSOLVER = Dune::Cholmod<CVEC>;
  using UMFPACKSOLVER = Dune::UMFPack<MAT>;
  using CUMFPACKSOLVER = Dune::UMFPack<CMAT>;

  enum
  {
    blocksize = VectorBlockType::dimension
  };

  enum class EigenproblemType {
    GenEO,
    AHarmonic
  };

  // the local data of this class

  // input to constructor
  MPI_Comm comm;
  std::shared_ptr<MAT> pAglobal;
  std::shared_ptr<MAT> pA_plus_global;
  std::shared_ptr<VEC> pglobal_dirichlet_mask;
  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pAvolume_snippet_matrices;
  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pAskeleton_snippet_matrices;
  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pA_plus_volume_snippet_matrices;
  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pA_plus_skeleton_snippet_matrices;
  std::shared_ptr<DomainDecompositionDOFMapper> pdddm;
  std::vector<size_t> nsubdomains;

  // local data
  size_t number_of_levels_;                                                  // how many levels do we have. Note 0 is the *finest* level and number_of_levels_-1 is the *coarsest* level!
  std::vector<std::shared_ptr<DomainDecomposition>> ddhierarchy;            // hierarchy of domain decomposition information
  std::vector<std::shared_ptr<DomainDecompositionDOFMapper>> dddmhierarchy; // hierarchy of dof information on the levels

  // system information
  std::vector<std::vector<std::vector<bool>>> subdomain_boundary_dofs;                     // subdomain_boundary_dofs[l][p][i] is true if dof i in subdomain p on level l is on the boundary
  std::vector<std::vector<std::vector<field_type>>> subdomain_boundary_mask;               // subdomain_boundary_dofs converted to a mask
  std::vector<std::vector<std::vector<field_type>>> partition_of_unity;                    // partition_of_unity[l][p][i] is in [0,1] for dof i in subdomain p on level l
  std::vector<std::vector<VEC>> fine_subdomain_basis_vectors;                              //  basis vectors computed in each subdomain on the finest level
  std::vector<std::vector<std::vector<CVEC>>> coarse_subdomain_basis_vectors;              //  basis vectors computed in each subdomain on the coarser level coarse_subdomain_basis_vectors[l][p][k]
  std::vector<std::vector<std::vector<CVEC>>> coarse_kerAB_vectors;                        //  basis vectors computed in each subdomain on the coarser level coarse_subdomain_basis_vectors[l][p][k]
  std::vector<std::shared_ptr<CMAT>> global_coarse_matrix;                                 // the global systems on the coarse levels
  std::vector<std::vector<std::shared_ptr<DenseMat>>> projected_volume_snippet_matrices;   // volume snippet matrices projected to subdomain basis vectors
  std::vector<std::vector<std::shared_ptr<DenseMat>>> projected_skeleton_snippet_matrices; // skeleton snippet matrices projected to subdomain basis vectors
  std::vector<std::vector<std::shared_ptr<CMAT>>> volume_snippet_matrices;                 // volume snippet matrices on coarser levels; from these the neumann matrices are assembled
  std::vector<std::vector<std::shared_ptr<CMAT>>> skeleton_snippet_matrices;               // skeleton snippet matrices on coarser levels; from these the neumann matrices are assembled

  // subdomain level solves
#ifdef USE_CHOLMOD
  std::vector<std::shared_ptr<CHOLMODSOLVER>> finesubdomainsolvers;                 // the LU decomposition of the subdomain problems on the finest level
  std::vector<std::shared_ptr<CCHOLMODSOLVER>> globalcoarsesolvers;                 // the LU decomposition of the coarse level problems
  std::vector<std::vector<std::shared_ptr<CCHOLMODSOLVER>>> coarsesubdomainsolvers; // the subdomain solvers on coarse levels
#else
  std::vector<std::shared_ptr<UMFPACKSOLVER>> finesubdomainsolvers;                 // the LU decomposition of the subdomain problems on the finest level
  std::vector<std::shared_ptr<CUMFPACKSOLVER>> globalcoarsesolvers;                 // the LU decomposition of the coarse level problems
  std::vector<std::vector<std::shared_ptr<CUMFPACKSOLVER>>> coarsesubdomainsolvers; // the subdomain solvers on coarse levels
#endif
  std::vector<std::shared_ptr<CVEC>> dcoarse; //  restricted defect
  std::vector<std::shared_ptr<CVEC>> vcoarse; //  prolongated corrections

  // timing
  std::vector<std::vector<double>> setup_time_procs;      // time spent on level l in proc p
  std::vector<std::vector<double>> setup_time_setup_gevp; // time spent on level l in proc p
  std::vector<std::vector<double>> setup_time_projection; // time spent on level l in proc p
  std::vector<std::vector<double>> setup_time_gevp;       // time spent on level l in proc p
  std::vector<std::vector<double>> setup_time_umfpack;    // time spent on level l in proc p
  std::vector<double> setup_time_parallelizable;          // times now sequential but probably parallelizable on each level
  double setup_time_sequential;                           // really sequential
  double setup_time_total;

  // parameters
  int nthreads_setup;            // number of threads to be used in this rank
  int nthreads_solve;            // number of threads to be used in this rank
  std::string pum;               // pu to be used
  std::string fineGEVPrhs;       // variant of B: use "pu", "1-pu" with "pu" being the default
  std::string coarseGEVPrhs;     // variant of B: use "pu", "1-pu" with "pu" being the default
  size_t shrink;                 // simulates oversampling in connection with distance based pu
  size_t coarse_overlap;         // overlap to be used when building coarse level domain decomposition
  std::string coarseeigensolver; // can take either "eigen" or "arpack" with eigen as the default
  size_t n_eigenvectors_coarse_computed;
  size_t n_eigenvectors_coarse_used;
  field_type eigenvalue_coarse_threshold; // <=0 means fixed size, >0 means take only evs with eiganvale<=threshold
  field_type abs_zero_ker;
  field_type regularization_ker;
  bool merge_disconnected;
  std::string cycle;
  int verbose; // 0=no output during creation, 1=minimal information
  bool issymmetric; // is the underlying problem symmetric?
  EigenproblemType eigenproblem_type;
  EigensolverParameters eigensolver_params;

  // for parallelization
  std::queue<size_t> task_queue; // queue of subdomain numbers
  std::mutex task_mutex;         // mutex to guard access to task queue
  std::mutex update_mutex;       // mutex to guard access to task queue
#ifdef MULTITHREADED
  std::vector<std::thread> threads; // thread pool
#endif
  double prec_total_time;

  Dune::Timer solver_timer;

public:
  //! \brief Constructor for symmetric case. It does all the work of setting up the components
  /**
   * \param comm_                         an MPI communicator (only needed for parmetis)
   * \param pAglobal_                     shared_ptr to the global system matrix
   * \param pglobal_dirichlet_mask_       global dof vector with 0 in dirichlet dofs and 1 else
   * \param pAvolume_snippet_matrices_    assembled volume snippet matrices on finest level
   * \param pAskeleton_snippet_matrices_  assembled skeleton snippet matrices on finest level
   * \param pdddm_                        dof mapper and domain decomposition information on finest level
   * \param nsubdomains                   number of subdomains on the different levels
   * \param coarse_overlap_               overlap to be used when building coarse level domain decomposition
   * \param pum                           partition of unity method to be used on finest level
   */
  MultiLevelGenEO(MPI_Comm comm_,                                                                   // an MPI communicator (only needed for parmetis)
                  int nthreads_setup_,                                                              // number of threads to be used in this rank
                  int nthreads_solve_,                                                              // number of threads to be used in this rank
                  std::shared_ptr<MAT> pAglobal_,                                                   // the global stiffness matrix
                  std::shared_ptr<VEC> pglobal_dirichlet_mask_,                                     // global dof vector with 0 in dirichlet dofs and 1 else
                  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pAvolume_snippet_matrices_,    // assembled volume snippet matrices on finest level
                  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pAskeleton_snippet_matrices_,  // assembled skeleton snippet matrices on finest level
                  std::shared_ptr<DomainDecompositionDOFMapper> pdddm_,                             // dof mapper and domain decomposition information on finest level
                  const std::vector<size_t> &nsubdomains_,                                          // desired subdomain numbers on all levels
                  std::string pum_,                                                                 // pu to be used
                  std::string fineGEVPrhs_,                                                         // variant of B: use "pu", "1-pu" with "pu" being the default
                  std::string coarseGEVPrhs_,                                                       // variant of B: use "pu", "1-pu" with "pu" being the default
                  size_t shrink_,                                                                   // shrink factor for partition of unity to simulate oversampling
                  size_t coarse_overlap_,                                                           // overlap to be used when building coarse level domain decomposition
                  std::string eigenproblem_type,                                                    // either "geneo" or "aharmonic"
                  EigensolverParameters fine_eigensolver_params,                                    // parameters for the fine-level eigensolver
                  std::string coarseeigensolver_,                                                   // can take either "eigen" or "arpack" with eigen as the default
                  size_t n_eigenvectors_coarse_computed_,
                  size_t n_eigenvectors_coarse_used_,
                  field_type eigenvalue_coarse_threshold_,                                          // <=0 means fixed size, >0 means take only evs with eiganvale<=threshold
                  field_type abs_zero_ker_,
                  field_type regularization_ker_,
                  bool merge_disconnected_,
                  std::string cycle_,
                  int verbose_                                                                      // 0=no output during creation, 1=minimal information
                  )
      : comm(comm_),
        nthreads_setup(nthreads_setup_),
        nthreads_solve(nthreads_solve_),
        pAglobal(pAglobal_),
        pA_plus_global(pAglobal_),
        pglobal_dirichlet_mask(pglobal_dirichlet_mask_),
        pAvolume_snippet_matrices(pAvolume_snippet_matrices_),
        pAskeleton_snippet_matrices(pAskeleton_snippet_matrices_),
        pA_plus_volume_snippet_matrices(pAvolume_snippet_matrices_),
        pA_plus_skeleton_snippet_matrices(pAskeleton_snippet_matrices_),
        finesubdomainsolvers(nsubdomains_[0]),
        fine_subdomain_basis_vectors(nsubdomains_[0]),
        pdddm(pdddm_),
        nsubdomains(nsubdomains_),
        pum(pum_),                     // pu to be used
        fineGEVPrhs(fineGEVPrhs_),     // variant of B: use "pu", "1-pu" with "pu" being the default
        coarseGEVPrhs(coarseGEVPrhs_), // variant of B: use "pu", "1-pu" with "pu" being the default
        shrink(shrink_),
        coarse_overlap(coarse_overlap_),       // overlap to be used when building coarse level domain decomposition
        eigenproblem_type(eigenproblem_type_for_string(eigenproblem_type)), // Convert string to EigenproblemType enum
        eigensolver_params(fine_eigensolver_params),
        coarseeigensolver(coarseeigensolver_), // can take either "eigen" or "arpack" with eigen as the default
        n_eigenvectors_coarse_computed(n_eigenvectors_coarse_computed_),
        n_eigenvectors_coarse_used(n_eigenvectors_coarse_used_),
        eigenvalue_coarse_threshold(eigenvalue_coarse_threshold_), // <=0 means fixed size, >0 means take only evs with eiganvale<=threshold
        abs_zero_ker(abs_zero_ker_),
        regularization_ker(regularization_ker_),
        merge_disconnected(merge_disconnected_),
        cycle(cycle_),
        verbose(verbose_), // 0=no output during creation, 1=minimal information
        issymmetric(true),
#ifdef MULTITHREADED
        threads(std::max(nthreads_setup_, nthreads_solve_)), // create non joinable threads
#endif
        prec_total_time(0.0)
  {
    set_up_preconditioner();
  }

  //! \brief Constructor for asymmetric case. It does all the work of setting up the components
  /**
   * \param comm_                               an MPI communicator (only needed for parmetis)
   * \param pAglobal_                           shared_ptr to the asymmetric global system matrix, A
   * \param pA_plus_global_                     shared_ptr to the symmetric global system matrix, A_plus
   * \param pglobal_dirichlet_mask_             global dof vector with 0 in dirichlet dofs and 1 else
   * \param pAvolume_snippet_matrices_          assembled volume snippet matrices on finest level for A
   * \param pAskeleton_snippet_matrices_        assembled skeleton snippet matrices on finest level for A
   * \param pA_plus_volume_snippet_matrices_    assembled volume snippet matrices on finest level for A_plus
   * \param pA_plus_skeleton_snippet_matrices_  assembled skeleton snippet matrices on finest level for A_plus
   * \param pdddm_                              dof mapper and domain decomposition information on finest level
   * \param nsubdomains                         number of subdomains on the different levels
   * \param coarse_overlap_                     overlap to be used when building coarse level domain decomposition
   * \param pum                                 partition of unity method to be used on finest level
   */
  MultiLevelGenEO(MPI_Comm comm_,                                                                         // an MPI communicator (only needed for parmetis)
                  int nthreads_setup_,                                                                    // number of threads to be used in this rank
                  int nthreads_solve_,                                                                    // number of threads to be used in this rank
                  std::shared_ptr<MAT> pAglobal_,                                                         // the asymmetric global stiffness matrix, A
                  std::shared_ptr<MAT> pA_plus_global_,                                                   // the symmetric global stiffness matrix, A_plus
                  std::shared_ptr<VEC> pglobal_dirichlet_mask_,                                           // global dof vector with 0 in dirichlet dofs and 1 else
                  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pAvolume_snippet_matrices_,          // assembled volume snippet matrices on finest level for A
                  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pAskeleton_snippet_matrices_,        // assembled skeleton snippet matrices on finest level for A
                  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pA_plus_volume_snippet_matrices_,    // assembled volume snippet matrices on finest level for A_plus
                  std::shared_ptr<std::vector<std::shared_ptr<MAT>>> pA_plus_skeleton_snippet_matrices_,  // assembled skeleton snippet matrices on finest level for A_plus
                  std::shared_ptr<DomainDecompositionDOFMapper> pdddm_,                                   // dof mapper and domain decomposition information on finest level
                  const std::vector<size_t> &nsubdomains_,                                                // desired subdomain numbers on all levels
                  std::string pum_,                                                                       // pu to be used
                  std::string fineGEVPrhs_,                                                               // variant of B: use "pu", "1-pu" with "pu" being the default
                  std::string coarseGEVPrhs_,                                                             // variant of B: use "pu", "1-pu" with "pu" being the default
                  size_t shrink_,                                                                         // shrink factor for partition of unity to simulate oversampling
                  size_t coarse_overlap_,                                                                 // overlap to be used when building coarse level domain decomposition
                  std::string eigenproblem_type,                                                          // either "geneo" or "aharmonic"
                  EigensolverParameters fine_eigensolver_params,                                          // parameters for the fine-level eigensolver
                  std::string coarseeigensolver_,                                                         // can take either "eigen" or "arpack" with eigen as the default
                  size_t n_eigenvectors_coarse_computed_,
                  size_t n_eigenvectors_coarse_used_,
                  field_type eigenvalue_coarse_threshold_, // <=0 means fixed size, >0 means take only evs with eiganvale<=threshold
                  field_type abs_zero_ker_,
                  field_type regularization_ker_,
                  bool merge_disconnected_,
                  std::string cycle_,
                  int verbose_ // 0=no output during creation, 1=minimal information
                  )
      : comm(comm_),
        nthreads_setup(nthreads_setup_),
        nthreads_solve(nthreads_solve_),
        pAglobal(pAglobal_),
        pA_plus_global(pA_plus_global_),
        pglobal_dirichlet_mask(pglobal_dirichlet_mask_),
        pAvolume_snippet_matrices(pAvolume_snippet_matrices_),
        pAskeleton_snippet_matrices(pAskeleton_snippet_matrices_),
        pA_plus_volume_snippet_matrices(pA_plus_volume_snippet_matrices_),
        pA_plus_skeleton_snippet_matrices(pA_plus_skeleton_snippet_matrices_),
        finesubdomainsolvers(nsubdomains_[0]),
        fine_subdomain_basis_vectors(nsubdomains_[0]),
        pdddm(pdddm_),
        nsubdomains(nsubdomains_),
        pum(pum_),                     // pu to be used
        fineGEVPrhs(fineGEVPrhs_),     // variant of B: use "pu", "1-pu" with "pu" being the default
        coarseGEVPrhs(coarseGEVPrhs_), // variant of B: use "pu", "1-pu" with "pu" being the default
        shrink(shrink_),
        coarse_overlap(coarse_overlap_),       // overlap to be used when building coarse level domain decomposition
        eigenproblem_type(eigenproblem_type_for_string(eigenproblem_type)), // Convert string to EigenproblemType enum
        eigensolver_params(fine_eigensolver_params),
        coarseeigensolver(coarseeigensolver_), // can take either "eigen" or "arpack" with eigen as the default
        n_eigenvectors_coarse_computed(n_eigenvectors_coarse_computed_),
        n_eigenvectors_coarse_used(n_eigenvectors_coarse_used_),
        eigenvalue_coarse_threshold(eigenvalue_coarse_threshold_), // <=0 means fixed size, >0 means take only evs with eiganvale<=threshold
        abs_zero_ker(abs_zero_ker_),
        regularization_ker(regularization_ker_),
        merge_disconnected(merge_disconnected_),
        cycle(cycle_),
        verbose(verbose_), // 0=no output during creation, 1=minimal information
        issymmetric(false),
#ifdef MULTITHREADED
        threads(std::max(nthreads_setup_, nthreads_solve_)), // create non joinable threads
#endif
        prec_total_time(0.0)
  {
    set_up_preconditioner();
  }

  void set_up_preconditioner()
  {
    Dune::Timer timer_total, timer, timer2; // two timers, one for total time, one for general use
    timer_total.reset();

    // some analysis of the input
    if (verbose > -1)
      std::cout << "starting level " << 0 << std::endl;
    size_t totaldofs = blocksize * pglobal_dirichlet_mask->N(), constraineddofs = 0;
    for (size_t i = 0; i < pglobal_dirichlet_mask->N(); i++)
      for (size_t j = 0; j < blocksize; j++)
        if ((*pglobal_dirichlet_mask)[i][j] == 0.0)
          constraineddofs++;
    if (verbose > 0)
      std::cout << "ML Geneo blocksize=" << blocksize << " constraineddofs=" << constraineddofs << " out of " << totaldofs << std::endl;

    // how many levels do we have. Note 0 is the *finest* level and Number_Of_Levels-1 is the *coarsest* level!
    number_of_levels_ = nsubdomains.size();

    // check input
    if (number_of_levels_ < 2)
      DUNE_THROW(Dune::Exception, "number of levels must be at least 2.");
    if (nsubdomains[number_of_levels_ - 1] != 1)
      DUNE_THROW(Dune::Exception, "number of subdomains on coarsest level must be 1.");

    //====================
    // We start on the finest level
    //====================

    timer.reset();

    // allocate time slots
    setup_time_procs.resize(number_of_levels_);
    setup_time_setup_gevp.resize(number_of_levels_);
    setup_time_projection.resize(number_of_levels_);
    setup_time_gevp.resize(number_of_levels_);
    setup_time_umfpack.resize(number_of_levels_);
    setup_time_procs[0].resize(nsubdomains[0], 0.0);
    setup_time_umfpack[0].resize(nsubdomains[0], 0.0);
    setup_time_parallelizable.resize(number_of_levels_, 0.0);
    setup_time_sequential = 0.0;
    setup_time_total = 0.0;

    // initialize multilevel domain decomposition and associated dof information
    ddhierarchy.resize(number_of_levels_);
    dddmhierarchy.resize(number_of_levels_);
    ddhierarchy[0] = pdddm->pdd;
    ddhierarchy[0]->print_info(verbose);
    dddmhierarchy[0] = pdddm;
    dddmhierarchy[0]->print_info(verbose);

    // for the solver application and the partition of unity we need the subdomain boundary dof information
    subdomain_boundary_dofs.resize(number_of_levels_);
    subdomain_boundary_dofs[0] = get_subdomain_boundary_dofs(*pAglobal, pdddm->subdomainnumber_local_to_global, pdddm->subdomainnumber_global_to_local);
    if (verbose > 2) // the following loop reports number of boundary dofs in each subdomain
      for (size_t p = 0; p < nsubdomains[0]; ++p)
      {
        size_t sdboundarydofs = 0;
        for (size_t i = 0; i < subdomain_boundary_dofs[0][p].size(); i++)
          if (subdomain_boundary_dofs[0][p][i])
            sdboundarydofs += blocksize;
        std::cout << "ML Geneo p=" << p << " boundarydofs=" << sdboundarydofs << std::endl;
      }

    // convert boundary dof information to a mask
    subdomain_boundary_mask.resize(number_of_levels_);
    subdomain_boundary_mask[0] = get_subdomain_boundary_mask<field_type>(subdomain_boundary_dofs[0]);

    // build partition of unity on finest level
    // the rest of this comment seems to be outdated PB 10 Aug 23
    // we can handle only the standard one here; otherwise we would have to assemble all subdomain matrices at once
    // this fits also better to the coarser levels which can only used the standard pu
    partition_of_unity.resize(number_of_levels_);
    if (pum == "distance")
    {
      // we need all matrices, this is costly!
      auto matrices = set_up_local_matrices(*pAglobal, pdddm->subdomainnumber_local_to_global, pdddm->subdomainnumber_global_to_local);
      partition_of_unity[0] = get_partition_of_unity_distance(matrices, pdddm->subdomainnumber_local_to_global, pdddm->subdomainnumber_global_to_local, subdomain_boundary_dofs[0], shrink);
    }
    else
      partition_of_unity[0] = get_partition_of_unity_standard<field_type>(pdddm->subdomainnumber_local_to_global, pdddm->subdomainnumber_global_to_local, subdomain_boundary_dofs[0]);
    setup_time_parallelizable[0] += timer.elapsed();

    // now lets work on the finest level: assemble subdomain matrices, etc
    finesubdomainsolvers.resize(nsubdomains[0]); // we will store the UMFPack solver objects here
    fine_subdomain_basis_vectors.resize(nsubdomains[0]);

    auto elapsed = timer.elapsed();
    if (verbose > -1)
      std::cout << "TIME prior to subdomain loop on level 0: " << elapsed << " seconds" << std::endl;

    // it suffices to assemble subdomain problems one by one!
    //==========================================
    // THIS IS THE BIG LOOP OVER ALL SUBDOMAINS!
    //==========================================
    std::vector<bool> disconnected_subdomains(nsubdomains[0], false); // true when subdomain is not connected

    timer.reset();
#ifdef MULTITHREADED
    for (size_t p = 0; p < nsubdomains[0]; ++p)
      task_queue.push(p);
    for (int rank = 0; rank < nthreads_setup; ++rank)
      threads[rank] = std::thread{&MultiLevelGenEO<MAT, VEC>::fine_subdomain_setup_task, this, rank};
    for (int rank = 0; rank < nthreads_setup; ++rank)
      threads[rank].join();
#else
    for (size_t p = 0; p < nsubdomains[0]; ++p)
      fine_subdomain_setup(0, p);
#endif
    elapsed = timer.elapsed();
    if (verbose > -1)
      std::cout << "TIME for subdomain loop on level 0: " << elapsed << " seconds" << std::endl;

    // Now that we have produced all basis vectors we may produce the global system on the next coarser level.
    // We would not actually need that for the intermediate levels 1...number_of_levels_-2 but it is definitely
    // easier and allows us to do the iterated two grid method

    timer.reset();

    // first add the basisvector local<->global information to the DofMapper on this level
    std::vector<size_t> number_of_basis_vectors(nsubdomains[0]);
    for (size_t p = 0; p < nsubdomains[0]; ++p)
      number_of_basis_vectors[p] = fine_subdomain_basis_vectors[p].size();                 // we take from finest level
    pdddm->fill_basisvector_information(number_of_basis_vectors, disconnected_subdomains); // computes basis vector global<=> local

    // set up global coarse level matrix
    {
      auto n_coarse = pdddm->basisvector_global_to_local.size(); // now we know the size of the coarse problem
      size_t nz_coarse = 0;                                      // compute number of nonzeroes
      for (size_t p = 0; p < pdddm->p_subdomain_graph->graph.size(); p++)
        for (auto q : pdddm->p_subdomain_graph->graph[p])
          nz_coarse += pdddm->basisvector_local_to_global[p].size() * pdddm->basisvector_local_to_global[q].size();
      if (verbose > -1)
        std::cout << "coarse matrix " << n_coarse << "x" << n_coarse << " with " << nz_coarse << " nonzeroes" << std::endl;
      global_coarse_matrix.resize(number_of_levels_);
      global_coarse_matrix[1] = std::shared_ptr<CMAT>(new CMAT(n_coarse, n_coarse, nz_coarse, CMAT::row_wise));   // allocate matrix in rowwise creation mode
      for (auto row = global_coarse_matrix[1]->createbegin(); row != global_coarse_matrix[1]->createend(); ++row) // fill nonzeroes
      {
        auto i = row.index();                                                       // this is the global row index
        auto p = pdddm->basisvector_global_to_subdomainnumber[i];                   // the subdomain this coarse vector comes from
        for (auto q : pdddm->p_subdomain_graph->graph[p])                           // loop over neighboring subdomains
          for (size_t k = 0; k < pdddm->basisvector_local_to_global[q].size(); k++) // loop over coarse vectors in neighboring subdomain
          {
            auto j = pdddm->basisvector_local_to_global[q][k]; // the global number of this coarse vector
            row.insert(j);
          }
      }
    }

    elapsed = timer.elapsed();
    if (verbose > -1)
      std::cout << "TIME for creating level 1 system " << elapsed << " seconds" << std::endl;
    timer.reset();

    // assemble global coarse matrix
#ifdef MULTITHREADED
    for (size_t p = 0; p < nsubdomains[0]; ++p)
      task_queue.push(p);
    for (int rank = 0; rank < nthreads_setup; ++rank)
      threads[rank] = std::thread{&MultiLevelGenEO<MAT, VEC>::fine_assemble_coarse_system_task, this, rank};
    for (int rank = 0; rank < nthreads_setup; ++rank)
      threads[rank].join();
#else
    for (size_t p = 0; p < nsubdomains[0]; ++p)
      fine_assemble_coarse_system(0, p);
#endif
    elapsed = timer.elapsed();
    if (verbose > -1)
      std::cout << "TIME for assembling level 1 system " << elapsed << " seconds" << std::endl;
    if (verbose > 2)
      std::cout << "computed entries of coarse system on level " << 1 << std::endl;

    timer.reset();

    // at this point we can compute the Galerkin projection of the volume and skeleton snippets
    projected_volume_snippet_matrices.resize(number_of_levels_); // to store all the matrices on the different levels
    if (pAskeleton_snippet_matrices != nullptr)
      projected_skeleton_snippet_matrices.resize(number_of_levels_);

    // projection of snippet matrices on subspace spanned by first coarse level
    if (number_of_levels_ > 2) // only then we will need projected snippet matrices
    {
      // resize containers for the result
      projected_volume_snippet_matrices[0].resize(ddhierarchy[0]->number_to_volumesnippet.size()); // resize to number of snippets
      if (pAskeleton_snippet_matrices != nullptr)
        projected_skeleton_snippet_matrices[0].resize(ddhierarchy[0]->number_to_skeletonsnippet.size()); // resize to number of snippets

        // now do the projection in parallel
#ifdef MULTITHREADED
      for (int rank = 0; rank < nthreads_setup; ++rank)
        threads[rank] = std::thread{&MultiLevelGenEO<MAT, VEC>::snippet_projection_finest_to_first_coarse, this, rank};
      for (int rank = 0; rank < nthreads_setup; ++rank)
        threads[rank].join();
#else
      nthreads_setup = 1;
      snippet_projection_finest_to_first_coarse(0);
#endif
    }

    elapsed = timer.elapsed();
    if (verbose > -1)
      std::cout << "TIME for projection of snippets " << elapsed << " seconds" << std::endl;

    setup_time_parallelizable[0] += timer.elapsed();

    //====================
    // Now we can proceed to the coarse levels; this is only needed when there are more than two levels!
    //====================

    // resize to hold level wise information
    volume_snippet_matrices.resize(number_of_levels_);
    if (pAskeleton_snippet_matrices != nullptr)
      skeleton_snippet_matrices.resize(number_of_levels_);
    coarse_subdomain_basis_vectors.resize(number_of_levels_);
    coarse_kerAB_vectors.resize(number_of_levels_);
    coarsesubdomainsolvers.resize(number_of_levels_);
    globalcoarsesolvers.resize(number_of_levels_);

    // loop over coarse levels
    for (size_t targetlevel = 1; targetlevel < number_of_levels_; targetlevel++)
    {
      setup_time_procs[targetlevel].resize(nsubdomains[targetlevel], 0.0);
      setup_time_setup_gevp[targetlevel].resize(nsubdomains[targetlevel], 0.0);
      setup_time_projection[targetlevel].resize(nsubdomains[targetlevel], 0.0);
      setup_time_gevp[targetlevel].resize(nsubdomains[targetlevel], 0.0);
      setup_time_umfpack[targetlevel].resize(nsubdomains[targetlevel], 0.0);

      // std::cout << "XXX " << targetlevel << " " << setup_time_parallelizable[targetlevel] << std::endl;

      // now we entered a coarse level
      // on the finer level we already have produced the basis vectors in each subdomain
      if (verbose > -1)
        std::cout << "starting coarse level " << targetlevel << std::endl;

      // cluster nodes of subdomain graph on the finer level
      timer.reset();
      if (verbose > 1)
        std::cout << "coarse subdomain graph " << std::endl;
      auto coarsening = ddhierarchy[targetlevel - 1]->p_subdomain_graph->coarsen_graph(comm, nsubdomains[targetlevel], coarse_overlap, dddmhierarchy[targetlevel - 1]->disconnected_subdomains, merge_disconnected);

      // from the coarsening we can deduce the domain decomposition on the target level
      if (verbose > 1)
        std::cout << "create domain decomposition information " << std::endl;
      ddhierarchy[targetlevel] = std::shared_ptr<DomainDecomposition>(new DomainDecomposition(ddhierarchy[targetlevel - 1], coarsening));
      ddhierarchy[targetlevel]->print_info(verbose);

      // now build dof information on this level
      if (verbose > 1)
        std::cout << "create dof information " << std::endl;
      dddmhierarchy[targetlevel] = std::shared_ptr<DomainDecompositionDOFMapper>(new DomainDecompositionDOFMapper(ddhierarchy[targetlevel], dddmhierarchy[targetlevel - 1]));
      dddmhierarchy[targetlevel]->print_info(verbose);
      setup_time_parallelizable[targetlevel] += timer.elapsed();

      elapsed = timer.elapsed();
      if (verbose > -1)
        std::cout << "TIME preparing level " << targetlevel << " is " << elapsed << " seconds" << std::endl;

      // if this is the coarsest level we are done; otherwise we need to compute basis vectors for the next coarser level ...
      if (targetlevel == number_of_levels_ - 1) {
        // factorize coarsest level system (wich already exists)
        std::cout << "factorizing coarse system on level " << targetlevel << std::endl;
        timer.reset();
#ifdef USE_CHOLMOD
        globalcoarsesolvers[targetlevel] = std::make_shared<CCHOLMODSOLVER>();
        globalcoarsesolvers[targetlevel]->setMatrix(*global_coarse_matrix[targetlevel]);
#else
        globalcoarsesolvers[targetlevel] = std::make_shared<CUMFPACKSOLVER>(*global_coarse_matrix[targetlevel], false);
#endif
        elapsed = timer.elapsed();
        setup_time_procs[targetlevel][0] += elapsed;
        if (verbose > -1)
          std::cout << "TIME factorizing coarsest system is " << elapsed << " seconds" << std::endl;

        break; // exit loop over setting up coarse levels
      }

      // now here we are NOT on the coarsest level
      // extract references for ease of writing
      auto &dd = *ddhierarchy[targetlevel];
      auto &dddm = *dddmhierarchy[targetlevel];
      auto &finedd = *ddhierarchy[targetlevel - 1];
      auto &finedddm = *dddmhierarchy[targetlevel - 1];

      timer.reset();
      //=======================
      // Step 1: assemble volume snippet matrices on this level from projected fine grid snippet matrices
      volume_snippet_matrices[targetlevel].resize(dd.number_to_volumesnippet.size()); // space for the pointers
      if (verbose > 0)
        std::cout << "assemble volume snippet matrices on level " << targetlevel << std::endl;
      for (size_t cvsn = 0; cvsn < dd.volumesnippet_coarse_to_fine.size(); cvsn++) // loop over all volume snippets on this level
      {
        // this is our coarse volume snippet
        auto &cvs = dd.number_to_volumesnippet[cvsn];
        if (verbose > 1)
          std::cout << "coarse volume snippet " << dd.vs_to_str(cvs);

        // count nonzeroes in snippet matrix
        std::map<size_t, std::set<size_t>> coupling;            // coupling of the subdomains
        for (auto fvsn : dd.volumesnippet_coarse_to_fine[cvsn]) // loop over volume snippets on fine level for this volume snippet
        {
          auto &fvs = finedd.number_to_volumesnippet[fvsn];
          // all subdomains of this snippet are fully coupled
          for (auto p : fvs)
            for (auto q : fvs)
              coupling[p].insert(q);
        }
        if (pAskeleton_snippet_matrices != nullptr)
          for (auto fssn : dd.volumesnippet_coarse_to_fine_skeletonsnippetnumber[cvsn]) // loop over skeleton snippets on fine level contained in volume snippet
          {
            // get snippets
            auto &fss = finedd.number_to_skeletonsnippet[fssn];
            auto it = fss.begin(); // points to first v snippet number
            size_t left_volumesnippetnumber = *it;
            ++it;
            size_t right_volumesnippetnumber = *it; // so now we have the numbers of the two volume snippets to left and right
            auto &left_volumesnippet = finedd.number_to_volumesnippet[left_volumesnippetnumber];
            auto &right_volumesnippet = finedd.number_to_volumesnippet[right_volumesnippetnumber];

            // add inter snippet couplings
            auto s = dd.set_intersection(left_volumesnippet, right_volumesnippet);
            for (auto p : s)
              for (auto q : s)
                coupling[p].insert(q);
          }

        size_t nz = 0; // compute nonzeroes from coupled subdomains
        for (auto &entry : coupling)
          for (auto &q : entry.second)
            nz += finedddm.n_basis_vectors_per_subdomain[entry.first] * finedddm.n_basis_vectors_per_subdomain[q];
        size_t n = dddm.volumesnippetnumber_local_to_global[cvsn].size(); // number of degrees of freedom
        if (verbose > 1)
          std::cout << " ndof=" << n << " nz=" << nz << std::endl;

        // set up sparse matrix for volume snippet
        volume_snippet_matrices[targetlevel][cvsn] = std::shared_ptr<CMAT>(new CMAT(n, n, nz, CMAT::row_wise)); // allocate matrix in rowwise creation mode
        auto &coarsesnippetA = *volume_snippet_matrices[targetlevel][cvsn];                                     // reference
        for (auto row = coarsesnippetA.createbegin(); row != coarsesnippetA.createend(); ++row)                 // loop over rows
        {
          auto i = row.index();                                                    // local in volume snippet
          auto iglobal = dddm.volumesnippetnumber_local_to_global[cvsn][i];        // global dof number on this level
          auto p = finedddm.basisvector_global_to_subdomainnumber[iglobal];        // the subdomain this coarse vector comes from
          for (auto q : coupling[p])                                               // loop over neighboring subdomains
            for (size_t l = 0; l < finedddm.n_basis_vectors_per_subdomain[q]; l++) // loop over basis vectors in neighboring subdomain
            {
              auto jglobal = finedddm.basisvector_local_to_global[q][l];        // the global number of this coarse vector
              auto j = dddm.volumesnippetnumber_global_to_local[jglobal][cvsn]; //
              // std::cout << "p=" << p << " q=" << q << " l=" << l << " jg=" << jglobal << " j=" << j << std::endl;
              row.insert(j);
            }
        }
        coarsesnippetA = 0.0; // clear matrix
        // Dune::printmatrix(std::cout, coarsesnippetA, "coarsesnippetA", "");

        // assemble from fine volume snippets contained in coarse volume snippet
        if (verbose > 1)
          std::cout << "assemble volume->volume " << targetlevel << std::endl;
        for (auto fvsn : dd.volumesnippet_coarse_to_fine[cvsn]) // loop over volume snippets on fine level for this volume snippet
        {
          auto &fvs = finedd.number_to_volumesnippet[fvsn];
          if (verbose > 1)
            std::cout << "  fine volume snippet " << dd.vs_to_str(fvs) << std::endl;

          // get projected fine level matrix for this fine snippet
          auto &finesnippetA = *projected_volume_snippet_matrices[targetlevel - 1][fvsn];

          // need locol_to_local dof mapping from fine snippet to coarse snippet
          std::vector<size_t> local_to_local(finesnippetA.N());
          size_t count = 0;
          for (auto p : fvs)                                                       // loop over all subdomains intersected by the volume snippet
            for (size_t k = 0; k < finedddm.n_basis_vectors_per_subdomain[p]; k++) // loop over all basis vectors in that subdomain
            {
              auto iglobal = finedddm.basisvector_local_to_global[p][k];                 // global dof number of basis vector [p][k] on this level
              auto ilocal_cvs = dddm.volumesnippetnumber_global_to_local[iglobal][cvsn]; // local number of that dof in coarse snippet
              auto ilocal_fvs = count++;                                                 // local number in fine snippet
              local_to_local[ilocal_fvs] = ilocal_cvs;
            }

          // now assemble fine snippet to coarse snippet
          for (auto rIt = finesnippetA.begin(); rIt != finesnippetA.end(); ++rIt)
          {
            auto cIt = rIt->begin();
            auto cEndIt = rIt->end();
            for (; cIt != cEndIt; ++cIt)
            {
              // std::cout << "(" << rIt.index() << "," << cIt.index() << ") -> (" << local_to_local[rIt.index()] << "," << local_to_local[cIt.index()] << ")" << std::endl;
              coarsesnippetA[local_to_local[rIt.index()]][local_to_local[cIt.index()]] += (*cIt);
            }
          }
        }

        // assemble from fine skeleton snippets contained inside the coarse volume snippet
        if (pAskeleton_snippet_matrices != nullptr)
        {
          if (verbose > 1)
            std::cout << "assemble skeleton->volume " << targetlevel << std::endl;
          for (auto fssn : dd.volumesnippet_coarse_to_fine_skeletonsnippetnumber[cvsn]) // loop over skeleton snippets on fine level inside this coarse volume snippet
          {
            auto &fss = finedd.number_to_skeletonsnippet[fssn];
            auto it = fss.begin(); // points to first v snippet number
            size_t left_volumesnippetnumber = *it;
            ++it;
            size_t right_volumesnippetnumber = *it; // so now we have the numbers of the two volume snippets to left and right
            auto &left_volumesnippet = finedd.number_to_volumesnippet[left_volumesnippetnumber];
            auto &right_volumesnippet = finedd.number_to_volumesnippet[right_volumesnippetnumber];
            auto s = dd.set_intersection(left_volumesnippet, right_volumesnippet); // subdomains that are both to left and right
            if (verbose > 1)
              std::cout << "  fine skeleton snippet " << finedd.ss_to_str(fss) << " intersection " << finedd.vs_to_str(s) << std::endl;

            // get projected fine level matrix for this fine snippet
            auto &finesnippetA = *projected_skeleton_snippet_matrices[targetlevel - 1][fssn];

            // need local_to_local dof mapping from fine snippet to coarse snippet
            // this is computed on the fly
            std::vector<size_t> local_to_local(finesnippetA.N());
            size_t count = 0;
            for (auto p : left_volumesnippet)       // loop over all subdomains intersected by the left volume snippet
              if (right_volumesnippet.count(p) > 0) // if p is to left and right of snippet
                for (size_t k = 0; k < finedddm.n_basis_vectors_per_subdomain[p]; k++)
                {
                  auto iglobal = finedddm.basisvector_local_to_global[p][k];                 // global dof number of basis vector [p][k] on this level
                  auto ilocal_cvs = dddm.volumesnippetnumber_global_to_local[iglobal][cvsn]; // local number of that dof in coarse snippet
                  auto ilocal_fss = count++;                                                 // local number in fine snippet
                  local_to_local[ilocal_fss] = ilocal_cvs;
                }

            // now assemble fine snippet to coarse snippet
            for (auto rIt = finesnippetA.begin(); rIt != finesnippetA.end(); ++rIt)
            {
              auto cIt = rIt->begin();
              auto cEndIt = rIt->end();
              for (; cIt != cEndIt; ++cIt)
              {
                // std::cout << "(" << rIt.index() << "," << cIt.index() << ") -> (" << local_to_local[rIt.index()] << "," << local_to_local[cIt.index()] << ")" << std::endl;
                coarsesnippetA[local_to_local[rIt.index()]][local_to_local[cIt.index()]] += (*cIt);
              }
            }
          }
        }

      } // end loop over coarse volume snippets; now the volume snippet matrices are assembled;

      //=======================
      // Step 2 : assemble coarse skeleton snippets from fine skeleton snippets
      if (pAskeleton_snippet_matrices != nullptr)
      {
        skeleton_snippet_matrices[targetlevel].resize(dd.number_to_skeletonsnippet.size()); // space for the pointers
        if (verbose > 0)
          std::cout << "assemble skeleton snippet matrices on level " << targetlevel << std::endl;
        for (size_t cssn = 0; cssn < dd.number_to_skeletonsnippet.size(); cssn++) // loop over all skelton snippets on this level
        {
          // this is our coarse volume snippet
          auto &css = dd.number_to_skeletonsnippet[cssn];
          if (verbose > 1)
            std::cout << "COARSE skeleton snippet " << dd.ss_to_str(css) << std::endl;

          // count nonzeroes in snippet matrix
          std::map<size_t, std::set<size_t>> coupling;              // coupling of the subdomains
          for (auto fssn : dd.skeletonsnippet_coarse_to_fine[cssn]) // loop over skeleton snippets on fine level for this skeleton snippet
          {
            // get snippets
            auto &fss = finedd.number_to_skeletonsnippet[fssn];
            auto it = fss.begin(); // points to first v snippet number
            size_t left_volumesnippetnumber = *it;
            ++it;
            size_t right_volumesnippetnumber = *it; // so now we have the numbers of the two volume snippets to left and right
            auto &left_volumesnippet = finedd.number_to_volumesnippet[left_volumesnippetnumber];
            auto &right_volumesnippet = finedd.number_to_volumesnippet[right_volumesnippetnumber];

            // all subdomains of this snippet are fully coupled
            auto s = finedd.set_intersection(left_volumesnippet, right_volumesnippet);
            if (verbose > 2)
              std::cout << "   fine skeleton snippet " << finedd.ss_to_str(fss) << " is: " << finedd.vs_to_str(s) << std::endl;
            for (auto p : s)
              for (auto q : s)
                coupling[p].insert(q);
          }
          size_t nz = 0; // compute nonzeroes from coupled subdomains
          for (auto &entry : coupling)
            for (auto &q : entry.second)
              nz += finedddm.n_basis_vectors_per_subdomain[entry.first] * finedddm.n_basis_vectors_per_subdomain[q];
          size_t n = dddm.skeletonsnippetnumber_local_to_global[cssn].size(); // number of degrees of freedom
          if (verbose > 1)
            std::cout << "ndof=" << n << " nz=" << nz << std::endl;

          // set up sparse matrix for skeleton snippet
          skeleton_snippet_matrices[targetlevel][cssn] = std::shared_ptr<CMAT>(new CMAT(n, n, nz, CMAT::row_wise)); // allocate matrix in rowwise creation mode
          auto &coarsesnippetA = *skeleton_snippet_matrices[targetlevel][cssn];                                     // reference
          for (auto row = coarsesnippetA.createbegin(); row != coarsesnippetA.createend(); ++row)                   // loop over rows
          {
            auto i = row.index();                                                    // local in skeleton snippet
            auto iglobal = dddm.skeletonsnippetnumber_local_to_global[cssn][i];      // global dof number on this level
            auto p = finedddm.basisvector_global_to_subdomainnumber[iglobal];        // the subdomain this coarse vector comes from
            for (auto q : coupling[p])                                               // loop over neighboring subdomains
              for (size_t l = 0; l < finedddm.n_basis_vectors_per_subdomain[q]; l++) // loop over basis vectors in neighboring subdomain
              {
                auto jglobal = finedddm.basisvector_local_to_global[q][l];          // the global number of this coarse vector
                auto j = dddm.skeletonsnippetnumber_global_to_local[jglobal][cssn]; //
                // std::cout << "p=" << p << " q=" << q << " l=" << l << " jg=" << jglobal << " j=" << j << std::endl;
                row.insert(j);
              }
          }
          coarsesnippetA = 0.0; // clear matrix
          // Dune::printmatrix(std::cout, coarsesnippetA, "coarsesnippetA", "");

          // assemble from fine skeleton snippets
          for (auto fssn : dd.skeletonsnippet_coarse_to_fine[cssn]) // loop over skeleton snippets on fine level for this skeleton snippet
          {
            // get snippets
            auto &fss = finedd.number_to_skeletonsnippet[fssn];
            auto it = fss.begin(); // points to first v snippet number
            size_t left_volumesnippetnumber = *it;
            ++it;
            size_t right_volumesnippetnumber = *it; // so now we have the numbers of the two volume snippets to left and right
            auto &left_volumesnippet = finedd.number_to_volumesnippet[left_volumesnippetnumber];
            auto &right_volumesnippet = finedd.number_to_volumesnippet[right_volumesnippetnumber];
            if (verbose > 1)
              std::cout << "  fine skeleton snippet " << finedd.ss_to_str(fss) << std::endl;

            // get projected fine level matrix for this fine snippet
            auto &finesnippetA = *projected_skeleton_snippet_matrices[targetlevel - 1][fssn];

            // need local_to_local dof mapping from fine snippet to coarse snippet
            std::vector<size_t> local_to_local;
            for (auto p : left_volumesnippet) // loop over all subdomains intersected by the left volume snippet
              if (right_volumesnippet.count(p) > 0)
                for (size_t k = 0; k < finedddm.n_basis_vectors_per_subdomain[p]; k++)
                {
                  auto iglobal = finedddm.basisvector_local_to_global[p][k]; // global dof number of basis vector [p][k] on this level
                  auto it = dddm.skeletonsnippetnumber_global_to_local[iglobal].find(cssn);
                  if (it == dddm.skeletonsnippetnumber_global_to_local[iglobal].end())
                    DUNE_THROW(Dune::Exception, "assemble coarse from fine skeleton snippets: access failed");
                  auto ilocal_css = it->second; // local number of that dof in coarse snippet
                  local_to_local.push_back(ilocal_css);
                }
            if (local_to_local.size() != finesnippetA.N())
              DUNE_THROW(Dune::Exception, "assemble coarse from fine skeleton snippets: size mismatch");

            // now assemble fine snippet to coarse snippet
            for (auto rIt = finesnippetA.begin(); rIt != finesnippetA.end(); ++rIt)
            {
              auto cIt = rIt->begin();
              auto cEndIt = rIt->end();
              for (; cIt != cEndIt; ++cIt)
              {
                // std::cout << "(" << rIt.index() << "," << cIt.index() << ") -> (" << local_to_local[rIt.index()] << "," << local_to_local[cIt.index()] << ")" << std::endl;
                coarsesnippetA[local_to_local[rIt.index()]][local_to_local[cIt.index()]] += (*cIt);
              }
            }
          } // loop over fine snippets
        }   // assembling coarse skeleton snippets from fine skeleton snippets
      }     // if we have skeleton snippets at all

      elapsed = timer.elapsed();
      if (verbose > -1)
        std::cout << "TIME assembling snippets on level " << targetlevel << " is " << elapsed << " seconds" << std::endl;

      // Step 3 : assemble Neumann subdomain matrices and solve eigenproblems; needs the partition of unity
      if (verbose > 0)
        std::cout << "assemble subdomain problems and eigensolves on level " << targetlevel << std::endl;

      // boundary dofs are determined by inspecting rows of the global matrix on this level
      subdomain_boundary_dofs[targetlevel] = dddm.subdomain_boundary;
      subdomain_boundary_mask[targetlevel] = get_subdomain_boundary_mask<field_type>(subdomain_boundary_dofs[targetlevel]);

      // build partition of unity
      partition_of_unity[targetlevel] = get_partition_of_unity_standard<field_type>(dddm.subdomainnumber_local_to_global, dddm.subdomainnumber_global_to_local, subdomain_boundary_dofs[targetlevel]);
      setup_time_parallelizable[targetlevel] += timer.elapsed();

      // loop over subdomains one by one!
      timer.reset();
      coarse_subdomain_basis_vectors[targetlevel].resize(nsubdomains[targetlevel]); // storage for basis vectors
      coarse_kerAB_vectors[targetlevel].resize(nsubdomains[targetlevel]);           // storage for basis vectors
      coarsesubdomainsolvers[targetlevel].resize(nsubdomains[targetlevel]);         // storage for the subdomain solvers
      std::vector<bool> disconnected_subdomains(nsubdomains[targetlevel], false);   // true when subdomain is not connected

      // this is the big subdomain loop with eigensolves on coarse levels
#ifdef MULTITHREADED
      for (size_t p = 0; p < nsubdomains[targetlevel]; ++p)
        task_queue.push(p);
      for (int rank = 0; rank < nthreads_setup; ++rank)
        threads[rank] = std::thread{&MultiLevelGenEO<MAT, VEC>::coarse_subdomain_setup_task, this, targetlevel, rank};
      for (int rank = 0; rank < nthreads_setup; ++rank)
        threads[rank].join();
#else
      for (size_t p = 0; p < nsubdomains[targetlevel]; ++p)
        coarse_subdomain_setup(targetlevel, 0, p);
#endif
      elapsed = timer.elapsed();
      if (verbose > -1)
        std::cout << "TIME big subdomain loop with eigensolves on level " << targetlevel << " is " << elapsed << " seconds" << std::endl;

      timer.reset();

      // as result there are now basis vectors on this level
      std::vector<size_t> number_of_basis_vectors(nsubdomains[targetlevel]);
      for (size_t p = 0; p < nsubdomains[targetlevel]; p++)
        number_of_basis_vectors[p] = coarse_subdomain_basis_vectors[targetlevel][p].size();
      dddm.fill_basisvector_information(number_of_basis_vectors, disconnected_subdomains); // computes basis vector global<=> local

      // and we need to construct the global coarse matrix ...
      // set up global coarse level matrix
      {
        auto n_coarse = dddm.basisvector_global_to_local.size(); // now we know the size of the coarse problem
        size_t nz_coarse = 0;                                    // compute number of nonzeroes
        for (size_t p = 0; p < dddm.p_subdomain_graph->graph.size(); p++)
          for (auto q : dddm.p_subdomain_graph->graph[p])
            nz_coarse += dddm.basisvector_local_to_global[p].size() * dddm.basisvector_local_to_global[q].size();
        if (verbose > 0)
          std::cout << "level " << targetlevel + 1 << " coarse matrix " << n_coarse << "x" << n_coarse << " with " << nz_coarse << " nonzeroes" << std::endl;
        global_coarse_matrix[targetlevel + 1] = std::shared_ptr<CMAT>(new CMAT(n_coarse, n_coarse, nz_coarse, CMAT::row_wise)); // allocate matrix in rowwise creation mode
        CMAT &matrix = *global_coarse_matrix[targetlevel + 1];
        for (auto row = matrix.createbegin(); row != matrix.createend(); ++row) // fill nonzeroes
        {
          auto i = row.index();                                                     // this is the global row index
          auto p = dddm.basisvector_global_to_subdomainnumber[i];                   // the subdomain this coarse vector comes from
          for (auto q : dddm.p_subdomain_graph->graph[p])                           // loop over neighboring subdomains
            for (size_t k = 0; k < dddm.basisvector_local_to_global[q].size(); k++) // loop over coarse vectors in neighboring subdomain
            {
              auto j = dddm.basisvector_local_to_global[q][k]; // the global number of this coarse vector
              row.insert(j);
            }
        }
      }

      //  assemble global coarse matrix
#ifdef MULTITHREADED
      for (size_t p = 0; p < nsubdomains[targetlevel]; ++p)
        task_queue.push(p);
      for (int rank = 0; rank < nthreads_setup; ++rank)
        threads[rank] = std::thread{&MultiLevelGenEO<MAT, VEC>::coarse_assemble_coarse_system_task, this, targetlevel, rank};
      for (int rank = 0; rank < nthreads_setup; ++rank)
        threads[rank].join();
#else
      for (size_t p = 0; p < nsubdomains[targetlevel]; ++p)
        coarse_assemble_coarse_system(targetlevel, 0, p);
#endif
      elapsed = timer.elapsed();
      if (verbose > -1)
        std::cout << "TIME for setting up and assembling system on level " << targetlevel + 1 << " is " << elapsed << " seconds" << std::endl;

      timer.reset();

      // at this point we can compute the Galerkin projection of the volume and skeleton snippets to the coarse basis vectors
      if (targetlevel < number_of_levels_ - 2) // if next coarse level is not the coarsest
      {
        // first, project the volume snippets
        projected_volume_snippet_matrices[targetlevel].resize(dd.number_to_volumesnippet.size()); // resize to number of snippets
        if (pAskeleton_snippet_matrices != nullptr)
          projected_skeleton_snippet_matrices[targetlevel].resize(dd.number_to_skeletonsnippet.size()); // resize to number of snippets

        if (verbose > 0)
          std::cout << "projection of volume and skeleton snippets to coarse basis vectors" << std::endl;

          // now do the projection in parallel
#ifdef MULTITHREADED
        for (int rank = 0; rank < nthreads_setup; ++rank)
          threads[rank] = std::thread{&MultiLevelGenEO<MAT, VEC>::snippet_projection_coarse_to_coarse, this, rank, targetlevel};
        for (int rank = 0; rank < nthreads_setup; ++rank)
          threads[rank].join();
#else
        nthreads_setup = 1;
        snippet_projection_coarse_to_coarse(0, targetlevel);
#endif
      }

      elapsed = timer.elapsed();
      if (verbose > -1)
        std::cout << "TIME for projecting snippets to level " << targetlevel + 1 << " is " << elapsed << " seconds" << std::endl;
    }

    // allocate defect and correction vectors for the solver
    dcoarse.resize(number_of_levels_);
    vcoarse.resize(number_of_levels_);
    for (int l = 1; l < number_of_levels_; l++) {
      dcoarse[l] = std::shared_ptr<CVEC>(new CVEC(dddmhierarchy[l]->n_global_dofs));
      vcoarse[l] = std::shared_ptr<CVEC>(new CVEC(dddmhierarchy[l]->n_global_dofs));
    }

    // print info
    for (size_t l = 0; l < number_of_levels_; l++) {
      size_t sum = 0;
      size_t maximum = 0;
      for (size_t p = 0; p < nsubdomains[l]; p++) {
        sum += dddmhierarchy[l]->subdomainnumber_local_to_global[p].size();
        maximum = std::max(maximum, dddmhierarchy[l]->subdomainnumber_local_to_global[p].size());
      }
      std::cout << "LEVEL " << l << ": " << dddmhierarchy[l]->subdomainnumber_global_to_local.size() << " dofs in " << nsubdomains[l] << " subdomains";
      std::cout << " max=" << maximum;
      std::cout << " avg=" << ((double)sum) / nsubdomains[l];
      std::cout << " ideal=" << ((double)dddmhierarchy[l]->subdomainnumber_global_to_local.size()) / nsubdomains[l];
      std::cout << std::endl;
    }
  }

  // grab as many subdomains as possible and process them
  void fine_assemble_coarse_system_task(size_t rank)
  {
    while (true)
    {
      size_t p; // the task to process in this iteration
      {
        std::lock_guard<std::mutex> lock{task_mutex}; // get exclusive access to task queue
        if (task_queue.empty())
          break;                // exit loop
        p = task_queue.front(); // read task
        task_queue.pop();       // remove task from queue
        if (verbose > 1)
          std::cout << "thread " << rank << " processing subdomain " << p << std::endl;
      }                                     // free lock
      fine_assemble_coarse_system(rank, p); // process task
    }
  }

  void fine_assemble_coarse_system(size_t rank, size_t prow)
  {
    // we need to set up the subdomain matrix again because it is not stores
    std::shared_ptr<MAT> pA = set_up_local_matrix(*pAglobal, prow, pdddm->subdomainnumber_local_to_global, pdddm->subdomainnumber_global_to_local);

    // and fill it with entries of the global matrix
    // this is needed to get the correct entries on the boundary of the subdomain
    MAT &Asubdomain = *pA;                      // get a reference for ease of writing
    for (size_t i = 0; i < Asubdomain.N(); i++) // loop over rows
    {
      auto iglobal = pdddm->subdomainnumber_local_to_global[prow][i];
      auto cIt = Asubdomain[i].begin();
      auto cEndIt = Asubdomain[i].end();
      for (; cIt != cEndIt; ++cIt) // loop over columns
      {
        auto j = cIt.index();
        auto jglobal = pdddm->subdomainnumber_local_to_global[prow][j];
        (*cIt) = (*pAglobal)[iglobal][jglobal];
      }
    }

    // now do the triple matrix products
    VEC v1(Asubdomain.N()); // two vectors on the subdomain to compute scalar products
    VEC v2(Asubdomain.N());

    // check other subdomains we overlap with by doing the scalar product of partition of unities
    for (size_t i = 0; i < Asubdomain.N(); i++)
      v1[i] = partition_of_unity[0][prow][i]; // partition of unity in subdomain prow
    std::set<size_t> neighbors;
    for (auto pcol : pdddm->p_subdomain_graph->graph[prow])
    {
      v2 = 0.0; // zero out
      for (size_t i = 0; i < Asubdomain.N(); i++)
      {
        auto iglobal = pdddm->subdomainnumber_local_to_global[prow][i];       // global index of this local index
        auto it = pdddm->subdomainnumber_global_to_local[iglobal].find(pcol); // test if this global dof block exists in neighboring subdomain
        if (it != pdddm->subdomainnumber_global_to_local[iglobal].end())      // if yes, we have a common entry
        {
          auto index_in_other_subdomain = it->second;                    // local index in subdomain pcol
          v2[i] = partition_of_unity[0][pcol][index_in_other_subdomain]; // restriction of partition of unity from other subdomain
        }
      }
      double sp = v2 * v1;
      if (std::abs(sp) > 1e-8)  // if we have an intersection of the partition of unities
        neighbors.insert(pcol); // then we store this subdomain as a neighbor
    }

    // now process all rows and all neighbors
    for (size_t krow = 0; krow < pdddm->basisvector_local_to_global[prow].size(); krow++) // loop over all basis vectors in this subdomain
    {
      // row number in global coarse system
      auto icoarse = pdddm->basisvector_local_to_global[prow][krow];

      // matrix vector product
      //Asubdomain.mv(fine_subdomain_basis_vectors[prow][krow], v1);
      Asubdomain.mtv(fine_subdomain_basis_vectors[prow][krow], v1);

      // now do all the scalar products with the other basis vectors
      for (auto pcol : pdddm->p_subdomain_graph->graph[prow])                                   // loop over all neighboring subdomains (including myself)
        if (neighbors.count(pcol) == 1)                                                         // process this subdomain only if we have an intersection
          for (size_t kcol = 0; kcol < pdddm->basisvector_local_to_global[pcol].size(); kcol++) // loop over all basis vectors in the neighboring subdomain
          {
            // column number in coarse system
            auto jcoarse = pdddm->basisvector_local_to_global[pcol][kcol];

            // produce restriction of basis vector from subdomain pcol to subdomain prow
            v2 = 0.0; // zero out
            for (size_t i = 0; i < Asubdomain.N(); i++)
            {
              auto iglobal = pdddm->subdomainnumber_local_to_global[prow][i];
              auto it = pdddm->subdomainnumber_global_to_local[iglobal].find(pcol); // test if this global dof block exists in neighboring subdomain
              if (it != pdddm->subdomainnumber_global_to_local[iglobal].end())      // if yes, we have a common entry
              {
                auto index_in_other_subdomain = it->second;
                v2[i] = fine_subdomain_basis_vectors[pcol][kcol][index_in_other_subdomain]; // restriction of coarse vector in subdomain pcol in subdomain prow
              }
            }

            // now we have a matrix entry
            (*global_coarse_matrix[1])[icoarse][jcoarse] = v2 * v1;
          }
    }
  }

  // grab as many subdomains as possible and process them
  void coarse_assemble_coarse_system_task(int targetlevel, size_t rank)
  {
    while (true)
    {
      size_t p; // the task to process in this iteration
      {
        std::lock_guard<std::mutex> lock{task_mutex}; // get exclusive access to task queue
        if (task_queue.empty())
          break;                // exit loop
        p = task_queue.front(); // read task
        task_queue.pop();       // remove task from queue
        if (verbose > 1)
          std::cout << "thread " << rank << " processing subdomain " << p << std::endl;
      }                                                    // free lock
      coarse_assemble_coarse_system(targetlevel, rank, p); // process task
    }
  }

  void coarse_assemble_coarse_system(int targetlevel, size_t rank, size_t p_col)
  {
    auto &dd = *ddhierarchy[targetlevel];
    auto &dddm = *dddmhierarchy[targetlevel];

    // assemble global coarse matrix
    // we need to set up the subdomain matrix again
    std::shared_ptr<CMAT> pA = set_up_local_matrix(*global_coarse_matrix[targetlevel], p_col, dddm.subdomainnumber_local_to_global, dddm.subdomainnumber_global_to_local);

    // ... by filling it with entries of the global matrix
    // this is needed to get the correct entries on the boundary of the subdomain (there, the Neumann matrix would be wrong)
    CMAT &Asubdomain = *pA;                     // get a reference for ease of writing
    for (size_t i = 0; i < Asubdomain.N(); i++) // loop over rows
    {
      auto iglobal = dddm.subdomainnumber_local_to_global[p_col][i];
      auto cIt = Asubdomain[i].begin();
      auto cEndIt = Asubdomain[i].end();
      for (; cIt != cEndIt; ++cIt) // loop over columns
      {
        auto j = cIt.index();
        auto jglobal = dddm.subdomainnumber_local_to_global[p_col][j];
        (*cIt) = (*global_coarse_matrix[targetlevel])[iglobal][jglobal];
      }
    }

    // now do the triple matrix products
    CVEC v1(Asubdomain.N()); // two vectors on the subdomain to compute scalar products
    CVEC v2(Asubdomain.N());
    for (size_t k_col = 0; k_col < dddm.basisvector_local_to_global[p_col].size(); k_col++) // loop over all basis vectors in this subdomain
    {
      // column number in global coarse system
      auto j = dddm.basisvector_local_to_global[p_col][k_col];

      // matrix vector product A*r_j
      Asubdomain.mv(coarse_subdomain_basis_vectors[targetlevel][p_col][k_col], v1);

      // now do all the scalar products with the other (row) basis vectors
      for (auto p_row : dddm.p_subdomain_graph->graph[p_col])                                   // loop over all neighboring subdomains (including myself)
        for (size_t k_row = 0; k_row < dddm.basisvector_local_to_global[p_row].size(); k_row++) // loop over all basis vectors in the neighboring subdomain
        {
          // row number in coarse system
          auto i = dddm.basisvector_local_to_global[p_row][k_row];

          // produce restriction of basis vector from subdomain p_row to subdomain p_col
          v2 = 0.0; // zero out
          for (size_t i = 0; i < Asubdomain.N(); i++)
          {
            auto iglobal = dddm.subdomainnumber_local_to_global[p_col][i];
            auto it = dddm.subdomainnumber_global_to_local[iglobal].find(p_row); // test if this global dof block exists in neighboring subdomain
            if (it != dddm.subdomainnumber_global_to_local[iglobal].end())       // if yes, we have a common entry
            {
              auto index_in_other_subdomain = it->second;
              v2[i] = coarse_subdomain_basis_vectors[targetlevel][p_row][k_row][index_in_other_subdomain]; // restriction of coarse vector in subdomain p_row in subdomain p_col
            }
          }

          // now we have a matrix entry
          (*global_coarse_matrix[targetlevel + 1])[i][j] = v2 * v1;
        }
    }
  }

  // grab as many subdomains as possible and process them
  void fine_subdomain_setup_task(size_t rank)
  {
    while (true)
    {
      size_t p; // the task to process in this iteration
      {
        std::lock_guard<std::mutex> lock{task_mutex}; // get exclusive access to task queue
        if (task_queue.empty())
          break;                // exit loop
        p = task_queue.front(); // read task
        task_queue.pop();       // remove task from queue
        if (verbose > 1)
          std::cout << "thread " << rank << " processing subdomain " << p << std::endl;
      }                              // free lock
      fine_subdomain_setup(rank, p); // process task
    }
  }

  // solve eigenvalue problem and factorize subdomain problem
  void fine_subdomain_setup(size_t rank, size_t p)
  {
    // set up sparse subdomain matrix
    std::shared_ptr<MAT> pA = set_up_local_matrix(*pAglobal, p, pdddm->subdomainnumber_local_to_global, pdddm->subdomainnumber_global_to_local);
    *pA = 0.0; // clear subdomain matrix

    // ...same for A_plus (if we're in the A-harmonic or asymmetric GenEO setting)
    std::shared_ptr<MAT> pA_plus;
    if (eigenproblem_type == EigenproblemType::AHarmonic or not issymmetric) {
      pA_plus = set_up_local_matrix(*pA_plus_global, p, pdddm->subdomainnumber_local_to_global, pdddm->subdomainnumber_global_to_local);
      *pA_plus = 0.0; // clear subdomain matrix
    }

    // assemble snippets to subdomain matrix
    // This results in Neumann subdomain matrix without any boundary conditions!
    assemble_snippets_to_subdomain_matrix(p,
                                          pdddm->pdd->subdomain_to_volumesnippetnumber,
                                          pdddm->volumesnippetnumber_local_to_global,
                                          pdddm->subdomainnumber_global_to_local,
                                          *pAvolume_snippet_matrices,
                                          pA);
    if (pAskeleton_snippet_matrices != nullptr)
      assemble_snippets_to_subdomain_matrix(p,
                                            pdddm->pdd->subdomain_to_interior_skeletonsnippetnumber,
                                            pdddm->skeletonsnippetnumber_local_to_global,
                                            pdddm->subdomainnumber_global_to_local,
                                            *pAskeleton_snippet_matrices,
                                            pA);

    // ...same for A_plus
    if (eigenproblem_type == EigenproblemType::AHarmonic or not issymmetric) {
      assemble_snippets_to_subdomain_matrix(p,
                                            pdddm->pdd->subdomain_to_volumesnippetnumber,
                                            pdddm->volumesnippetnumber_local_to_global,
                                            pdddm->subdomainnumber_global_to_local,
                                            *pA_plus_volume_snippet_matrices,
                                            pA_plus);
      if (pA_plus_skeleton_snippet_matrices != nullptr)
        assemble_snippets_to_subdomain_matrix(p,
                                              pdddm->pdd->subdomain_to_interior_skeletonsnippetnumber,
                                              pdddm->skeletonsnippetnumber_local_to_global,
                                              pdddm->subdomainnumber_global_to_local,
                                              *pA_plus_skeleton_snippet_matrices,
                                              pA_plus);
    }

    // assemble global dirichlet constraints in subdomain matrices
    bool floating = assemble_dirichlet_constraints_in_subdomain(pdddm->subdomainnumber_local_to_global[p], *pglobal_dirichlet_mask, pA);
    if (eigenproblem_type == EigenproblemType::AHarmonic or not issymmetric)
      assemble_dirichlet_constraints_in_subdomain(pdddm->subdomainnumber_local_to_global[p], *pglobal_dirichlet_mask, pA_plus);

    // At this point we have the NEUMANN subdomain matrix (with global Dirichlet conditions enforced in non-floating subdomains).
    // Now we may set up and solve the eigenvalue problems
    MAT &A = *pA;              // introduce references for ease of writing
    MAT &A_plus = *pA_plus;
    // B will be the right-hand side of our eigenproblem. In the symmetric case, it starts as a copy of A,
    // in the unsymmetric or A-harmonic case it starts as a copy of the symmetric Neumann matrix A_plus.
    MAT B;
    if (eigenproblem_type == EigenproblemType::AHarmonic or not issymmetric)
      B = A_plus;
    else
      B = A;

    // put in zero on diagonal on Dirichlet dofs; Note: the partition of unity is not zero on the global Dirichlet boundary
    for (auto row_iter = B.begin(); row_iter != B.end(); ++row_iter)
      for (auto col_iter = row_iter->begin(); col_iter != row_iter->end(); ++col_iter)
        if (col_iter.index() == row_iter.index())
        {
          auto biglobal = pdddm->subdomainnumber_local_to_global[p][row_iter.index()];
          for (size_t i = 0; i < blocksize; i++)
            (*col_iter)[i][i] *= (*pglobal_dirichlet_mask)[biglobal][i];
        }

    if (fineGEVPrhs == "geneo") {
      B = 0.0; // clear B
      assemble_snippets_to_subdomain_matrix_geneo_volume(p,
                                                         pdddm->pdd->subdomain_to_volumesnippetnumber,
                                                         pdddm->pdd->number_to_volumesnippet,
                                                         pdddm->volumesnippetnumber_local_to_global,
                                                         pdddm->subdomainnumber_global_to_local,
                                                         *pA_plus_volume_snippet_matrices,
                                                         B); // assemble without interior
      if (pA_plus_skeleton_snippet_matrices != nullptr)
        assemble_snippets_to_subdomain_matrix_geneo_skeleton(p,
                                                             pdddm->pdd->subdomain_to_interior_skeletonsnippetnumber,
                                                             pdddm->pdd->number_to_volumesnippet,
                                                             pdddm->pdd->number_to_skeletonsnippet,
                                                             pdddm->skeletonsnippetnumber_local_to_global,
                                                             pdddm->subdomainnumber_global_to_local,
                                                             *pA_plus_skeleton_snippet_matrices,
                                                             B);
      assemble_dirichlet_constraints_in_subdomain(pdddm->subdomainnumber_local_to_global[p], *pglobal_dirichlet_mask, B);
      
      for (auto row_iter = B.begin(); row_iter != B.end(); ++row_iter)
        for (auto col_iter = row_iter->begin(); col_iter != row_iter->end(); ++col_iter)
          *col_iter *= partition_of_unity[0][p][row_iter.index()] * partition_of_unity[0][p][col_iter.index()];
    } else if (fineGEVPrhs == "1-pu") {
      for (auto row_iter = B.begin(); row_iter != B.end(); ++row_iter)
        for (auto col_iter = row_iter->begin(); col_iter != row_iter->end(); ++col_iter)
          *col_iter *= (1.0 - partition_of_unity[0][p][row_iter.index()]) * (1.0 - partition_of_unity[0][p][col_iter.index()]);
    } else if (fineGEVPrhs == "pu") {
      // Now multiply B from left and right with the partition of unity.
      for (auto row_iter = B.begin(); row_iter != B.end(); ++row_iter)
        for (auto col_iter = row_iter->begin(); col_iter != row_iter->end(); ++col_iter)
          *col_iter *= partition_of_unity[0][p][row_iter.index()] * partition_of_unity[0][p][col_iter.index()];
    } else {
      DUNE_THROW(Dune::NotImplemented, "RHS of fine GEVP of type '" + fineGEVPrhs + "' is not implemented");
    }

    // These variables are only used in the A-harmonic case, they are defined here because they are needed in multiple places below
    const auto n = pdddm->subdomainnumber_local_to_global[p].size(); // size of vectors in subdomain p
    std::vector<size_t> reorder(n);
    enum DofType {
      inner = 1,
      boundary = 2,
      dirichlet = 3
    };
    std::vector<int> doftype(n); // 1=inner, 2=interior subdomain boundary, 3=global dirichlet boundary
    
    // Set up eigenproblems
    if (verbose > 0)
      std::cout << "subdomain " << p << ": solve eigenvalue problem, floating=" << floating << std::endl;

    std::unique_ptr<EigenSolver<MAT, VEC>> eigensolver;

    std::shared_ptr<MAT> GEVP_A, GEVP_B; // The matrices used in the eigenvalue problem. Will be set below
    switch (eigenproblem_type) {
    case EigenproblemType::GenEO:
      if (issymmetric) {
        GEVP_A = pA;
        GEVP_B = Dune::stackobject_to_shared_ptr(B);
      } else {
        GEVP_A = pA_plus;
        GEVP_B = Dune::stackobject_to_shared_ptr(B);
      }      
      break;

    case EigenproblemType::AHarmonic: {
      if (blocksize > 1)
        DUNE_THROW(Dune::Exception, "aharmonic coarse space only implemented for blocksize 1");
      
      size_t n_inner = 0;          // dof in subdomain interior or on global Neumann boundary
      size_t n_boundary = 0;       // dof on interior subdomain boundary
      size_t n_dirichlet = 0;      // dof on global Dirichlet boundary
      for (size_t i = 0; i < n; i++) {
        auto iglobal = pdddm->subdomainnumber_local_to_global[p][i];
        if ((*pglobal_dirichlet_mask)[iglobal][0] == 0.0) {
          doftype[i] = dirichlet;
          n_dirichlet += 1;
        } else if (subdomain_boundary_dofs[0][p][i]) {
          doftype[i] = boundary;
          n_boundary += 1;
        } else {
          doftype[i] = inner;
          n_inner += 1;
        }
      }

      // count and report
      size_t n1 = n_inner + n_boundary;
      size_t n2 = n_inner;
      size_t nhat = n1 + n2;
      if (verbose > 0)
        std::cout << "a-harmonic eigenproblem sd=" << p << " n=" << n << " n_inner=" << n_inner << " n_boundary=" << n_boundary << " n_dirichlet=" << n_dirichlet << " n1=" << n1 << " n2=" << n2
                  << " nhat=" << nhat << std::endl;

      // reorder index set: reorder[i] enumerates first all interior then all boundary and then all dirichlet dofs
      size_t ci = 0;
      size_t cb = n_inner;
      size_t cd = n_inner + n_boundary;
      for (size_t i = 0; i < n; i++)
        if (doftype[i] == inner)
          reorder[i] = ci++;
        else if (doftype[i] == boundary)
          reorder[i] = cb++;
        else
          reorder[i] = cd++;

      // now we may allocate the matrix without any dirchlet dofs
      int avg = A.nonzeroes() / n + 2;
      auto pAhat2 = std::shared_ptr<MAT>(new MAT(nhat, nhat, 2 * avg, 0.15, MAT::implicit)); // a new sparse matrix

      // set the A-harmonic constraint (asymmetric problem)
      for (auto row_iter = A.begin(); row_iter != A.end(); ++row_iter) {
        auto bi = row_iter.index();
        auto ri = reorder[bi];
        for (auto col_iter = row_iter->begin(); col_iter != row_iter->end(); ++col_iter) {
          auto bj = col_iter.index();
          auto rj = reorder[bj];
          if ((doftype[bi] == inner) && (doftype[bj] != dirichlet)) {
            // note: in DUNE, a_ij = a(\phi_j, \phi_i), switched indices compared with Ma 2023, (2.39)
            pAhat2->entry(rj, n1 + ri) = *col_iter; // A_11^T and A_12^T, B block of Ma 2023, (2.39)
            pAhat2->entry(n1 + ri, rj) = *col_iter; // A_11 and A_12, B^* block of Ma 2023, (2.39)
          }
        }
      }

      // set the remaining part corresponding to the symmetric matrix
      // TO DO: move to other for-loop? more efficient? what about memory?
      for (auto row_iter = A_plus.begin(); row_iter != A_plus.end(); ++row_iter) {
        auto bi = row_iter.index();
        auto ri = reorder[bi];
        for (auto col_iter = row_iter->begin(); col_iter != row_iter->end(); ++col_iter) {
          auto bj = col_iter.index();
          auto rj = reorder[bj];
          if ((doftype[bi] != dirichlet) && (doftype[bj] != dirichlet)) // B^+ block of Ma 2023, (2.39)
            pAhat2->entry(ri, rj) = *col_iter;
        }
      }
      auto stats = pAhat2->compress();

      // and the same for the B
      int avgB = B.nonzeroes() / n + 2;
      auto pBhat2 = std::shared_ptr<MAT>(new MAT(nhat, nhat, 2 * avgB, 0.15, MAT::implicit)); // a new sparse matrix
      for (auto row_iter = B.begin(); row_iter != B.end(); ++row_iter) {
        auto bi = row_iter.index();
        auto ri = reorder[bi];
        for (auto col_iter = row_iter->begin(); col_iter != row_iter->end(); ++col_iter) {
          auto bj = col_iter.index();
          auto rj = reorder[bj];
          if ((doftype[bi] == inner) && (doftype[bj] == inner)) // P block of Ma 2023, (2.39)
            pBhat2->entry(ri, rj) = *col_iter;
        }
      }
      stats = pBhat2->compress();

      GEVP_A = pAhat2;
      GEVP_B = pBhat2;
    } break;

    default:
      DUNE_THROW(Dune::Exception, "Unknown type of generalized eigenvalue problem (use either geneo or aharmonic)");
    }

    // Now set up and solve the eigenvalue problem with the given matrices
    if (eigensolver_params.type == "raes")
    {
      eigensolver = std::make_unique<RAESWrapper<MAT, VEC>>(GEVP_A, GEVP_B, eigensolver_params, verbose);
    }
    else if (eigensolver_params.type == "arpack")
    {
      if (nthreads_setup > 1)
        DUNE_THROW(Dune::Exception, "attempt to solve with Arpack with more than one thread");
      eigensolver = std::make_unique<ArpackWrapper<MAT, VEC>>(GEVP_A, GEVP_B, eigensolver_params, verbose);
    }
#ifdef DUNE_FT_HAVE_FEAST
    else if (eigensolver_params.type == "feast")
    {
      eigensolver = std::make_unique<FeastWrapper<MAT, VEC>>(GEVP_A, GEVP_B, eigensolver_params, verbose);
    }
#endif
#ifdef DUNE_FT_HAVE_SPECTRA
    else if (eigensolver_params.type == "spectra")
    {
      eigensolver = std::make_unique<SpectraWrapper<MAT, VEC>>(GEVP_A, GEVP_B, eigensolver_params, verbose);
    }
#endif
    else
    {
      DUNE_THROW(Dune::NotImplemented, "Eigensolver '" + eigensolver_params.type + "' is not implemented");
    }
    auto [eigenvectors, eigenvalues] = eigensolver->solve();

    // In the A-harmonic case, the first 'n' components (reordered appropriately) of the computed eigenvectors make up the actual eigenvectors
    if (eigenproblem_type == EigenproblemType::AHarmonic)
    {
      VEC vec(n);
      vec = 0.0;
      std::vector<VEC> eigenvectors_actual(eigenvectors.size(), vec);
      
      for (size_t k = 0; k < eigenvectors.size(); k++)
        for (size_t i = 0; i < n; i++)
          if (doftype[i] != dirichlet)
            eigenvectors_actual[k][i] = eigenvectors[k][reorder[i]];
          else
            eigenvectors_actual[k][i] = 0.0;

      eigenvectors = std::move(eigenvectors_actual);
    }
    
    if (verbose > 0)
    {
      for (size_t i = 0; i < eigenvalues.size(); ++i)
        std::cout << "  lambda_" << i << " = " << eigenvalues[i] << std::endl;
    }


    // analyse eigenvalues
    bool negative = false;
    for (size_t i = 0; i < eigenvalues.size(); ++i)
      if (eigenvalues[i] < -1e-4)
        negative = true;
    if (negative)
    {
      std::cout << "negative eigenvalue(s) found in subdomain " << p << " on level " << 0 << std::endl;
      for (size_t i = 0; i < eigenvalues.size(); ++i)
        if (eigenvalues[i] < -1e-4)
          std::cout << "  lambda_" << i << " = " << eigenvalues[i] << std::endl;
      // DUNE_THROW(Dune::Exception, "at least one negative eigenvalue detected");
    }
    size_t number_of_zero_eigenvalues = 0;
    // for (size_t i=0; i<eigenvalues.size(); ++i)
    //   if (std::abs(eigenvalues[i])<5e-13)
    //     number_of_zero_eigenvalues++;
    // if (number_of_zero_eigenvalues>1)
    //   disconnected_subdomains[p] = true;
    // if (number_of_zero_eigenvalues>1)
    //   {
    //     std::cout << ">1 zero eigenvalue(s) found in subdomain " << p << std::endl;
    //     for (size_t i=0; i<eigenvalues.size(); ++i)
    //       if (std::abs(eigenvalues[i])<5e-13)
    //      std::cout << "  lambda_" << i<< " = " << eigenvalues[i] << std::endl;
    //   }

    // store the eigenvectors computed
    if (eigensolver_params.threshold <= 0)
    {
      // we take a fixed number of eigenvalues
      size_t n_zero = 0;
      size_t n_positive = 0;
      for (size_t k = 0; k < eigenvalues.size(); k++) // look at all eigenvectors
      {
        if (fine_subdomain_basis_vectors[p].size() >= eigensolver_params.nev_target)
          break;

        // if (eigenvalues[k]>=1e-11 && n_zero+n_positive>=n_eigenvectors_fine_used_)
        //   break; // we have enough

        // if (eigenvalues[k]<1e-11)
        //   n_zero = 1; // count them only once
        // else
        //   n_positive++;

        // rescale
        eigenvectors[k] *= 1.0 / eigenvectors[k].two_norm();
        
        // multiply with partition of unity and zero on Dirichlet boundary
        for (size_t i = 0; i < eigenvectors[0].size(); i++)
        {
          auto iglobal = pdddm->subdomainnumber_local_to_global[p][i];
          eigenvectors[k][i] *= partition_of_unity[0][p][i]; // multiply with partition of unity, puts zero on subdomain boundary
          for (size_t j = 0; j < blocksize; j++)
            eigenvectors[k][i][j] *= (*pglobal_dirichlet_mask)[iglobal][j]; // zero out dof on global Dirichlet boundary
        }

        // now we have a basis vector in the subdomain
        fine_subdomain_basis_vectors[p].push_back(eigenvectors[k]);
      }
      // std::vector<size_t> colpermut(fine_subdomain_basis_vectors[p].size());
      // field_type th=1e-14;
      // field_type min_norm=1e100;
      // auto rank = rank_revealing_gram_schmidt(A,A,std::vector<VEC>(0),fine_subdomain_basis_vectors[p],colpermut,th,min_norm);
      // if (rank<fine_subdomain_basis_vectors[p].size())
      //   std::cout << "subdomain " << p << " has " << fine_subdomain_basis_vectors[p].size() << " basis vectors with " << rank << " linearly dependent" << std::endl;

      if (verbose > 2)
        std::cout << p << ": " << fine_subdomain_basis_vectors[p].size()
                  << " basis vectors taken, first ev: " << eigenvalues[0]
                  << " last ev: " << eigenvalues[fine_subdomain_basis_vectors[p].size() - 1]
                  << std::endl;
    }
    else
    {
      // adaptive choice of eigenvalues
      if (eigenvalues.size() <= 1)
        DUNE_THROW(Dune::Exception, "you should probably compute more than 1 eigenvalue ...");
      size_t last_taken = 0;
      for (size_t k = 0; k < eigenvalues.size() - 1; k++) // we know they are all nonnegative and ordered
        //if (eigenvalues[k] <= eigenvalue_fine_threshold)
        if (std::abs(eigenvalues[k]) <= eigensolver_params.threshold)
        {
          // rescale
          eigenvectors[k] *= 1.0 / eigenvectors[k].two_norm();

          // multiply with partition of unity and zero on Dirichlet boundary
          for (size_t i = 0; i < eigenvectors[0].size(); i++)
          {
            auto iglobal = pdddm->subdomainnumber_local_to_global[p][i];
            eigenvectors[k][i] *= partition_of_unity[0][p][i]; // multiply with partition of unity, puts zero on subdomain boundary
            for (size_t j = 0; j < blocksize; j++)
              eigenvectors[k][i][j] *= (*pglobal_dirichlet_mask)[iglobal][j]; // zero out dof on global Dirichlet boundary
          }

          /*// rescale
          eigenvectors[k] *= 1.0 / eigenvectors[k].two_norm();*/
          // now we have a basis vector in the subdomain
          //if (eigenvalues[k] > -1e3)
          {
            fine_subdomain_basis_vectors[p].push_back(eigenvectors[k]);
            last_taken = k;
          }
        }
      //if (eigenvalues[last_taken + 1] <= eigenvalue_fine_threshold)
      if (std::abs(eigenvalues[last_taken + 1]) <= eigensolver_params.threshold)
      {
        std::cout << "WARNING: number of eigenvalues computed is too small. First eigenvalue not taken: " << eigenvalues[last_taken + 1] << std::endl;
      }
      if (verbose > 2)
        std::cout << p << ": " << fine_subdomain_basis_vectors[p].size()
                  << " basis vectors taken, first ev: " << eigenvalues[0]
                  << " last ev: " << eigenvalues[fine_subdomain_basis_vectors[p].size() - 1]
                  << std::endl;
    }

    /*std::string evfile = "eigenvalues_diff6_dg1_ref2_sd";
    evfile.append(std::to_string(p));
    std::ofstream fout(evfile.append(".txt"));
    fout << std::setprecision(10);

    for(auto const& ev : eigenvalues)
        fout << ev << '\n';*/

    // and we set up the subdomain problem solvers for the apply phase
    // assemble dirichlet conditions on subdomain boundary
    for (size_t i = 0; i < A.N(); i++)
      if (subdomain_boundary_dofs[0][p][i])
      {
        // all dofs in this block are on the boundary
        auto cIt = A[i].begin();
        auto cEndIt = A[i].end();
        for (; cIt != cEndIt; ++cIt)
        {
          (*cIt) = 0.0;
          if (cIt.index() == i)
            for (size_t comp = 0; comp < blocksize; comp++)
              (*cIt)[comp][comp] = 1.0;
        }
      }
      else
      {
        // non dirchlet row; eliminate Dirichlet columns
        auto cIt = A[i].begin();
        auto cEndIt = A[i].end();
        for (; cIt != cEndIt; ++cIt)
          if (subdomain_boundary_dofs[0][p][cIt.index()])
            (*cIt) = 0.0;
      }

      // store pointer to UMFPACK solver object holding the decomposed matrix
      // hopefully it does not need the original matrix anymore ...
#ifdef USE_CHOLMOD
    finesubdomainsolvers[p] = std::make_shared<CHOLMODSOLVER>();
    finesubdomainsolvers[p]->setMatrix(A);
#else
    finesubdomainsolvers[p] = std::make_shared<UMFPACKSOLVER>(A, false);
#endif
  } // end loop over all subdomains on finest level

  // grab as many subdomains as possible and process them
  void coarse_subdomain_setup_task(size_t targetlevel, size_t rank)
  {
    while (true)
    {
      size_t p; // the task to process in this iteration
      {
        std::lock_guard<std::mutex> lock{task_mutex}; // get exclusive access to task queue
        if (task_queue.empty())
          break;                // exit loop
        p = task_queue.front(); // read task
        task_queue.pop();       // remove task from queue
        if (verbose > 1)
          std::cout << "thread " << rank << " processing subdomain " << p << std::endl;
      }                                             // free lock
      coarse_subdomain_setup(targetlevel, rank, p); // process task
    }
  }

  // solve coarse eigenproblem in one subdomain (used for multithreading)
  void coarse_subdomain_setup(size_t targetlevel, size_t rank, size_t p)
  {
    // introduce abbreviations
    auto &dd = *ddhierarchy[targetlevel];
    auto &dddm = *dddmhierarchy[targetlevel];
    auto &finedd = *ddhierarchy[targetlevel - 1];
    auto &finedddm = *dddmhierarchy[targetlevel - 1];

    // set up sparse subdomain matrix
    std::shared_ptr<CMAT> pA = set_up_local_matrix(*global_coarse_matrix[targetlevel], p, dddm.subdomainnumber_local_to_global, dddm.subdomainnumber_global_to_local);
    *pA = 0.0; // clear subdomain matrix
    if (verbose > 0)
      std::cout << "process subdomain " << p << " on level " << targetlevel << " n=" << pA->N() << std::endl;

    // assemble snippets to subdomain matrix
    // This results in Neumann subdomain matrix without any boundary conditions!
    if (verbose > 1)
      std::cout << "assembling volume->subdomain" << std::endl;
    assemble_snippets_to_subdomain_matrix(p,
                                          dd.subdomain_to_volumesnippetnumber,
                                          dddm.volumesnippetnumber_local_to_global,
                                          dddm.subdomainnumber_global_to_local,
                                          volume_snippet_matrices[targetlevel],
                                          pA);
    if (pAskeleton_snippet_matrices != nullptr)
    {
      if (verbose > 1)
        std::cout << "assembling interior skeleton->subdomain" << std::endl;
      assemble_snippets_to_subdomain_matrix(p,
                                            dd.subdomain_to_interior_skeletonsnippetnumber,
                                            dddm.skeletonsnippetnumber_local_to_global,
                                            dddm.subdomainnumber_global_to_local,
                                            skeleton_snippet_matrices[targetlevel],
                                            pA);
    }
    // there are no Dirichlet conditions on the coarse levels anymore...

    // now at this point we have the NEUMANN subdomain matrix
    // now we may set up and solve the eigenvalue problems
    CMAT &A = *pA;     // introduce reference for ease of writing
    // make_symmetric(A); // out of paranoia
    // is_symmetric(A,"Symmetry of B",true,1e-7);

    // make the sparse B matrix out of A
    CMAT B(A);
    if (coarseGEVPrhs == "geneo")
    {
      B = 0.0; // clear B
      assemble_snippets_to_subdomain_matrix_geneo_volume(p,
                                                         dd.subdomain_to_volumesnippetnumber,
                                                         dd.number_to_volumesnippet,
                                                         dddm.volumesnippetnumber_local_to_global,
                                                         dddm.subdomainnumber_global_to_local,
                                                         volume_snippet_matrices[targetlevel],
                                                         B); // assemble without interior
      if (pAskeleton_snippet_matrices != nullptr)
        assemble_snippets_to_subdomain_matrix_geneo_skeleton(p,
                                                             dd.subdomain_to_interior_skeletonsnippetnumber,
                                                             dd.number_to_volumesnippet,
                                                             dd.number_to_skeletonsnippet,
                                                             dddm.skeletonsnippetnumber_local_to_global,
                                                             dddm.subdomainnumber_global_to_local,
                                                             skeleton_snippet_matrices[targetlevel],
                                                             B);

      // std::ostringstream sBovlp;
      // sBovlp << basename << "_Bovlp_" << 1 << "_" << p << ".txt";
      // std::cout << "writing matrix file " << sBovlp.str() << std::endl;
      // writeMatrixToMatlab(B,sBovlp.str(),16);

      for (auto row_iter = B.begin(); row_iter != B.end(); ++row_iter)
        for (auto col_iter = row_iter->begin(); col_iter != row_iter->end(); ++col_iter)
          *col_iter *= partition_of_unity[targetlevel][p][row_iter.index()] * partition_of_unity[targetlevel][p][col_iter.index()];
    }
    else if (coarseGEVPrhs == "1-pu")
    {
      for (auto rIt = B.begin(); rIt != B.end(); ++rIt)
        for (auto cIt = rIt->begin(); cIt != rIt->end(); ++cIt)
          (*cIt) *= (1.0 - partition_of_unity[targetlevel][p][rIt.index()]) * (1.0 - partition_of_unity[targetlevel][p][cIt.index()]);
    }
    else
    {
      for (auto rIt = B.begin(); rIt != B.end(); ++rIt)
        for (auto cIt = rIt->begin(); cIt != rIt->end(); ++cIt)
          (*cIt) *= partition_of_unity[targetlevel][p][rIt.index()] * partition_of_unity[targetlevel][p][cIt.index()];
    }

    // writing matrices
    // std::string basename("neumann_problem");
    // std::ostringstream sA;
    // sA << basename << "_A_" << targetlevel << "_" << p << ".txt";
    // std::cout << "writing matrix file " << sA.str() << std::endl;
    // writeMatrixToMatlab(A,sA.str(),16);
    // std::ostringstream sB;
    // sB << basename << "_B_" << targetlevel << "_" << p << ".txt";
    // std::cout << "writing matrix file " << sB.str() << std::endl;
    // writeMatrixToMatlab(B,sB.str(),16);

    //------------------------------------------------------
    // set up A+B and project it to range(A+B)
    //------------------------------------------------------

    // counts number of basis vectors already put into the coarse space from this subdomain
    // since this can be done at various places, we define this variable here.
    size_t n_actually_used = 0;

    // QR decomposition of A+B
    size_t n = A.N(); // #dof in our coarse problem

    if (coarseGEVPrhs == "1-pu")
      DUNE_THROW(Dune::Exception, "straightforward application of ARPACK should only be used with partition of unity pu");

    // some additional regularization of A to make UMFPack not fail
    if (regularization_ker > 0.0)
      for (auto rIt = A.begin(); rIt != A.end(); ++rIt)
        for (auto cIt = rIt->begin(); cIt != rIt->end(); ++cIt)
          if (rIt.index() == cIt.index() && std::abs(partition_of_unity[targetlevel][p][rIt.index()]) < 1e-10)
            (*cIt) += regularization_ker;

    // make seperate vectors for arpack
    CVEC vec(n); // make a template
    vec = 0.0;
    std::vector<CVEC> eigenvectors(n_eigenvectors_coarse_computed, vec); // create vector of ISTL vectors for Arpack to compute

    // make vectors to store eigenvalues. Eigenvectors will be written directly into coarse_basis_vectors.
    std::vector<double> eigenvalues(n_eigenvectors_coarse_computed, 0.0);

    // solve GenEO eigenproblems
    if (coarseeigensolver == "raes")
    {
      GeneralizedInverse(A, B, eigensolver_params.shift, 0.0, eigensolver_params.tolerance, 100 * n_eigenvectors_coarse_used, n_eigenvectors_coarse_computed, eigenvalues, eigenvectors, 0, 12345);
    }
    else if (coarseeigensolver == "arpack")
    {
      if (nthreads_setup > 1)
        DUNE_THROW(Dune::Exception, "attempt to solve with Arpack with more than one thread");
      ArpackMLGeneo::ArPackPlusPlus_Algorithms<CMAT, CVEC> arpack(A);
      if (eigenvalue_coarse_threshold <= 0)
        arpack.computeGenSymShiftInvertMinMagnitude(B, eigensolver_params.threshold, eigenvectors, eigenvalues, eigensolver_params.shift);
      else
        arpack.computeGenSymShiftInvertMinMagnitudeAdaptive(B, eigensolver_params.threshold, eigenvectors, eigenvalues, eigensolver_params.shift, eigenvalue_coarse_threshold, n_eigenvectors_coarse_used);
    } else {
      DUNE_THROW(Dune::Exception, "Unknown coarse eigensolver");
    }

    if (verbose > 0)
      for (size_t i = 0; i < eigenvalues.size(); ++i)
        std::cout << "  coarse lambda_" << i << " = " << eigenvalues[i] << std::endl;

    // analyse EV for negative eigenvalues
    for (size_t i = 0; i < eigenvalues.size(); ++i)
      if (eigenvalues[i] < -1e-8)
      {
        CVEC Ae(n);
        A.mv(eigenvectors[i], Ae);
        CVEC Be(n);
        B.mv(eigenvectors[i], Be);
        auto X(Ae);
        X.axpy(-eigenvalues[i], Be);
        auto Y(Ae);
        Y.axpy(eigenvalues[i], Be);
        std::cout << "eigenvalue " << i << " in subdomain " << p << " is negative : "
                  << "||e||=" << eigenvectors[i].infinity_norm()
                  << " ||Ae||=" << Ae.infinity_norm()
                  << " ||Be||=" << Be.infinity_norm()
                  << " ||Ae-lambda*Be||=" << X.infinity_norm()
                  << " ||Ae+lambda*Be||=" << Y.infinity_norm()
                  << std::endl;
      }

    // store eigenvectors
    for (size_t k = 0; k < eigenvalues.size(); k++)
      if ((eigenvalue_coarse_threshold <= 0 && n_actually_used < n_eigenvectors_coarse_used) || (eigenvalue_coarse_threshold > 0 && eigenvalues[k] <= eigenvalue_coarse_threshold))
        if (eigenvalues[k] > -1e30)
        {
          // generate eigenvector
          CVEC &v = eigenvectors[k];

          // multiply with partition of unity
          for (size_t i = 0; i < n; i++)
            v[i] *= partition_of_unity[targetlevel][p][i]; // multiply with partition of unity, puts zero on subdomain boundary

          // rescale
          if (verbose > 1)
            std::cout << "norm of basisvector " << k << " is " << v.two_norm() << std::endl;
          v *= 1.0 / v.two_norm();

          // now we have a basis vector in the subdomain
          coarse_subdomain_basis_vectors[targetlevel][p].push_back(v);
          n_actually_used++;
        }
    if (verbose > 0)
      std::cout << p << ": " << coarse_subdomain_basis_vectors[targetlevel][p].size() << " basis vectors taken" << std::endl;

    // check adaptive ev threshold
    if (eigenvalue_coarse_threshold > 0)
    {
      if (n_actually_used == eigenvalues.size())
        std::cout << "WARNING: all eigenvectors used, cannot guarantee threshold" << std::endl;
      else if (eigenvalues[n_actually_used] <= eigenvalue_coarse_threshold)
        std::cout << "WARNING: number of eigenvalues not sufficient for threshold. First not taken: " << eigenvalues[n_actually_used] << std::endl;
    }

    //------------------------------------------------------
    // now make the subdomain matrix out of A and factorize it
    //------------------------------------------------------

    // assemble dirichlet conditions on subdomain boundary in A
    for (size_t i = 0; i < A.N(); i++)
      if (subdomain_boundary_dofs[targetlevel][p][i])
      {
        // boundary row
        auto cIt = A[i].begin();
        auto cEndIt = A[i].end();
        for (; cIt != cEndIt; ++cIt)
          (*cIt) = (cIt.index() == i) ? 1.0 : 0.0;
      }
      else
      {
        // non boundary row
        auto cIt = A[i].begin();
        auto cEndIt = A[i].end();
        for (; cIt != cEndIt; ++cIt)
          if (subdomain_boundary_dofs[targetlevel][p][cIt.index()])
            (*cIt) = 0.0;
      }

      // // compare with C
      // is_equal(A,C,"A=C",true,1e-12);

      // factorize subdomain system with UMFPack
#ifdef USE_CHOLMOD
    coarsesubdomainsolvers[targetlevel][p] = std::make_shared<CCHOLMODSOLVER>();
    coarsesubdomainsolvers[targetlevel][p]->setMatrix(A);
#else
    coarsesubdomainsolvers[targetlevel][p] = std::make_shared<CUMFPACKSOLVER>(A, false);
#endif
  }

  //  ZZZZ
  // project snippets in parallel
  void snippet_projection_finest_to_first_coarse(size_t rank)
  {
    // first, project the volume snippets
    if (verbose > 1)
      std::cout << "projection of volume snippets to coarse basis vectors" << std::endl;
    size_t nvsn = ddhierarchy[0]->number_to_volumesnippet.size();
    for (size_t vsn = 0; vsn < nvsn; vsn++) // loop over all volume snippets
    {
      if (vsn % nthreads_setup != rank)
        continue; // compute in parallel

      auto &vs = ddhierarchy[0]->number_to_volumesnippet[vsn]; // get the volume snippet
      if (verbose > 3)
        std::cout << "Galerkin projection of " << ddhierarchy[0]->vs_to_str(vs);
      MAT &snippetA = *(*pAvolume_snippet_matrices)[vsn]; // matrix for volume snippet on current level
      if (verbose > 3)
        std::cout << " with matrix of size " << snippetA.N() << "x" << snippetA.M();

      // we need a local to global map for the subdomain basis vectors involved with this snippet
      // this can be computed on the fly for the volume snippet
      std::map<size_t, std::map<size_t, size_t>> subdomain_local_to_snippet_local;
      size_t nc = 0;
      for (auto p : vs) // loop over all subdomains intersected by the volume snippet
        for (size_t k = 0; k < dddmhierarchy[0]->n_basis_vectors_per_subdomain[p]; k++)
          subdomain_local_to_snippet_local[p][k] = nc++;
      if (verbose > 2)
        std::cout << " nc=" << nc;

      // allocate dense matrix
      std::shared_ptr<DenseMat> pA = std::shared_ptr<DenseMat>(new DenseMat(nc, nc, 0.0)); // make a dense matrix with zeroes
      projected_volume_snippet_matrices[0][vsn] = pA;                                      // store the matrix; it will be used on the next coarser level
      DenseMat &projectedA = *pA;                                                          // use reference

      // compute projected entries
      for (auto &colentry1 : subdomain_local_to_snippet_local) // loop over all subdomains intersected by the volume snippet
        for (auto &colentry2 : colentry1.second)
        {
          auto pcol = colentry1.first;
          auto kcol = colentry2.first;
          auto jprojected = colentry2.second;

          // restrict basis vector [pcol][kcol] to snippet
          VEC v1(dddmhierarchy[0]->volumesnippetnumber_local_to_global[vsn].size());
          v1 = 0.0;
          for (size_t i = 0; i < dddmhierarchy[0]->volumesnippetnumber_local_to_global[vsn].size(); i++) // loop over all dofs in volume snippet
          {
            auto iglobal = dddmhierarchy[0]->volumesnippetnumber_local_to_global[vsn][i];       // global dof
            auto isubdomain = dddmhierarchy[0]->subdomainnumber_global_to_local[iglobal][pcol]; // local dof in subdomain pcol
            v1[i] = fine_subdomain_basis_vectors[pcol][kcol][isubdomain];                       // pick entry from basis vector
          }

          // multiply snippet matrix with column basis vector
          VEC Av1(dddmhierarchy[0]->volumesnippetnumber_local_to_global[vsn].size());
          //snippetA.mv(v1, Av1);
          snippetA.mtv(v1, Av1);

          // now multiply with the rows
          for (auto &rowentry1 : subdomain_local_to_snippet_local) // loop over all subdomains intersected by the volume snippet
            for (auto &rowentry2 : rowentry1.second)
            {
              auto prow = rowentry1.first;
              auto krow = rowentry2.first;
              auto iprojected = rowentry2.second;

              // restrict basis vector [prow][krow] to snippet
              VEC v2(dddmhierarchy[0]->volumesnippetnumber_local_to_global[vsn].size());
              v2 = 0.0;
              for (size_t i = 0; i < dddmhierarchy[0]->volumesnippetnumber_local_to_global[vsn].size(); i++) // loop over all dofs in volume snippet
              {
                auto iglobal = dddmhierarchy[0]->volumesnippetnumber_local_to_global[vsn][i];       // global dof
                auto isubdomain = dddmhierarchy[0]->subdomainnumber_global_to_local[iglobal][prow]; // local dof in subdomain prow
                v2[i] = fine_subdomain_basis_vectors[prow][krow][isubdomain];                       // pick entry from basis vector
              }

              // now we have one entry
              projectedA[iprojected][jprojected] = v2 * Av1;
            }
        }
      if (verbose > 2)
        std::cout << " ... done" << std::endl;
    } // loop over volume snippets

    // next are the skeleton snippets; it works basically the same ...
    if (pAskeleton_snippet_matrices != nullptr)
    {
      if (verbose > 1)
        std::cout << "projection of skeleton snippets to coarse basis vectors" << std::endl;
      size_t nssn = ddhierarchy[0]->number_to_skeletonsnippet.size();
      for (size_t ssn = 0; ssn < nssn; ssn++) // loop over all skeleton snippets
      {
        if (ssn % nthreads_setup != rank)
          continue;

        auto &ss = ddhierarchy[0]->number_to_skeletonsnippet[ssn]; // get the skeleton snippet
        auto it = ss.begin();                                      // points to first v snippet number
        size_t left_volumesnippetnumber = *it;
        ++it;
        size_t right_volumesnippetnumber = *it; // so now we have the numbers of the two volume snippets to left and right
        auto &left_volumesnippet = ddhierarchy[0]->number_to_volumesnippet[left_volumesnippetnumber];
        auto &right_volumesnippet = ddhierarchy[0]->number_to_volumesnippet[right_volumesnippetnumber];

        if (verbose > 3)
          std::cout << "Galerkin projection of skeleton snippet " << ddhierarchy[0]->vs_to_str(ss) << " : " << ddhierarchy[0]->vs_to_str(left_volumesnippet) << "|" << ddhierarchy[0]->vs_to_str(right_volumesnippet);
        MAT &snippetA = *(*pAskeleton_snippet_matrices)[ssn]; // matrix for skeleton snippet on current level
        if (verbose > 3)
          std::cout << " with matrix of size " << snippetA.N() << "x" << snippetA.M();

        // we need a local to global map for the subdomain basis vectors involved with this snippet
        // this can be computed on the fly for the skeleton snippet
        // NOTE: only dofs of subdomains which are BOTH left and right are involved!
        // WHY? say subdomain p is only to one side, then such a function zero (PU!) on one side and non existent on the other side
        std::map<size_t, std::map<size_t, size_t>> subdomain_local_to_snippet_local;
        size_t nc = 0;
        for (auto p : left_volumesnippet)       // loop over all subdomains intersected by the left volume snippet
          if (right_volumesnippet.count(p) > 0) // p is to left and right of skeleton snippet
            for (size_t k = 0; k < dddmhierarchy[0]->n_basis_vectors_per_subdomain[p]; k++)
              subdomain_local_to_snippet_local[p][k] = nc++;
        if (verbose > 2)
          std::cout << " nc=" << nc;

        // allocate dense matrix
        std::shared_ptr<DenseMat> pA = std::shared_ptr<DenseMat>(new DenseMat(nc, nc, 0.0)); // make a dense matrix with zeroes
        projected_skeleton_snippet_matrices[0][ssn] = pA;                                    // store the matrix; it will be used on the next coarser level
        DenseMat &projectedA = *pA;                                                          // use reference

        // compute projected entries
        for (auto &colentry1 : subdomain_local_to_snippet_local) // loop over all subdomains intersected by the snippet
          for (auto &colentry2 : colentry1.second)
          {
            auto pcol = colentry1.first;
            auto kcol = colentry2.first;
            auto jprojected = colentry2.second;

            // restrict basis vector [pcol][kcol] to snippet
            VEC v1(dddmhierarchy[0]->skeletonsnippetnumber_local_to_global[ssn].size());
            v1 = 0.0;
            for (size_t i = 0; i < dddmhierarchy[0]->skeletonsnippetnumber_local_to_global[ssn].size(); i++) // loop over all dofs in snippet
            {
              auto iglobal = dddmhierarchy[0]->skeletonsnippetnumber_local_to_global[ssn][i]; // global dof
              auto it = dddmhierarchy[0]->subdomainnumber_global_to_local[iglobal].find(pcol);
              if (it != dddmhierarchy[0]->subdomainnumber_global_to_local[iglobal].end())
                v1[i] = fine_subdomain_basis_vectors[pcol][kcol][it->second]; // pick entry from basis vector
            }

            // multiply snippet matrix with column basis vector
            VEC Av1(dddmhierarchy[0]->skeletonsnippetnumber_local_to_global[ssn].size());
            snippetA.mv(v1, Av1);

            // now multiply with the rows
            for (auto &rowentry1 : subdomain_local_to_snippet_local) // loop over all subdomains intersected by the volume snippet
              for (auto &rowentry2 : rowentry1.second)
              {
                auto prow = rowentry1.first;
                auto krow = rowentry2.first;
                auto iprojected = rowentry2.second;

                // restrict basis vector [prow][krow] to snippet
                VEC v2(dddmhierarchy[0]->skeletonsnippetnumber_local_to_global[ssn].size());
                v2 = 0.0;
                for (size_t i = 0; i < dddmhierarchy[0]->skeletonsnippetnumber_local_to_global[ssn].size(); i++) // loop over all dofs in snippet
                {
                  auto iglobal = dddmhierarchy[0]->skeletonsnippetnumber_local_to_global[ssn][i]; // global dof
                  auto it = dddmhierarchy[0]->subdomainnumber_global_to_local[iglobal].find(prow);
                  if (it != dddmhierarchy[0]->subdomainnumber_global_to_local[iglobal].end())
                    v2[i] = fine_subdomain_basis_vectors[prow][krow][it->second]; // pick entry from basis vector
                }

                // now we have one entry
                projectedA[iprojected][jprojected] = v2 * Av1;
              }
          }
        if (verbose > 2)
          std::cout << " ... done" << std::endl;
      } // loop over skeleton snippets
    }   // if there are skeleton snippets
  }

  //  ZZZZZZ
  void snippet_projection_coarse_to_coarse(size_t rank, int targetlevel)
  {
    auto &dd = *ddhierarchy[targetlevel];
    auto &dddm = *dddmhierarchy[targetlevel];

    for (size_t vsn = 0; vsn < dd.number_to_volumesnippet.size(); vsn++) // loop over all volume snippets on this level
    {
      if (vsn % nthreads_setup != rank)
        continue; // compute in parallel

      auto &vs = dd.number_to_volumesnippet[vsn]; // get the volume snippet
      if (verbose > 1)
        std::cout << "Galerkin projection of " << dd.vs_to_str(vs);
      CMAT &snippetA = *volume_snippet_matrices[targetlevel][vsn]; // matrix for volume snippet on current level
      if (verbose > 1)
        std::cout << " with matrix of size " << snippetA.N() << "x" << snippetA.M();

      // we need a local to global map for the subdomain basis vectors involved with this snippet
      // this can be computed on the fly for the volume snippet
      std::map<size_t, std::map<size_t, size_t>> subdomain_local_to_snippet_local;
      size_t nc = 0;
      for (auto p : vs) // loop over all subdomains intersected by the volume snippet
        for (size_t k = 0; k < dddm.n_basis_vectors_per_subdomain[p]; k++)
          subdomain_local_to_snippet_local[p][k] = nc++;
      if (verbose > 1)
        std::cout << " nc=" << nc;

      // allocate dense matrix
      std::shared_ptr<DenseMat> pA = std::shared_ptr<DenseMat>(new DenseMat(nc, nc, 0.0)); // make a dense matrix with zeroes
      projected_volume_snippet_matrices[targetlevel][vsn] = pA;                            // store the matrix; it will be used on the next coarser level
      DenseMat &projectedA = *pA;                                                          // use reference

      // compute projected entries
      for (auto &colentry1 : subdomain_local_to_snippet_local) // loop over all subdomains intersected by the volume snippet
        for (auto &colentry2 : colentry1.second)
        {
          auto pcol = colentry1.first;
          auto kcol = colentry2.first;
          auto jprojected = colentry2.second;

          // restrict basis vector [pcol][kcol] to snippet
          CVEC v1(dddm.volumesnippetnumber_local_to_global[vsn].size());
          v1 = 0.0;
          for (size_t i = 0; i < dddm.volumesnippetnumber_local_to_global[vsn].size(); i++) // loop over all dofs in volume snippet
          {
            auto iglobal = dddm.volumesnippetnumber_local_to_global[vsn][i];             // global dof
            auto isubdomain = dddm.subdomainnumber_global_to_local[iglobal][pcol];       // local dof in subdomain pcol
            v1[i] = coarse_subdomain_basis_vectors[targetlevel][pcol][kcol][isubdomain]; // pick entry from basis vector
          }

          // multiply snippet matrix with column basis vector
          CVEC Av1(dddm.volumesnippetnumber_local_to_global[vsn].size());
          snippetA.mv(v1, Av1);

          // now multiply with the rows
          for (auto &rowentry1 : subdomain_local_to_snippet_local) // loop over all subdomains intersected by the volume snippet
            for (auto &rowentry2 : rowentry1.second)
            {
              auto prow = rowentry1.first;
              auto krow = rowentry2.first;
              auto iprojected = rowentry2.second;

              // restrict basis vector [prow][krow] to snippet
              CVEC v2(dddm.volumesnippetnumber_local_to_global[vsn].size());
              v2 = 0.0;
              for (size_t i = 0; i < dddm.volumesnippetnumber_local_to_global[vsn].size(); i++) // loop over all dofs in volume snippet
              {
                auto iglobal = dddm.volumesnippetnumber_local_to_global[vsn][i];             // global dof
                auto isubdomain = dddm.subdomainnumber_global_to_local[iglobal][prow];       // local dof in subdomain prow
                v2[i] = coarse_subdomain_basis_vectors[targetlevel][prow][krow][isubdomain]; // pick entry from basis vector
              }

              // now we have one entry
              projectedA[iprojected][jprojected] = v2 * Av1;
            }
        }
      if (verbose > 1)
        std::cout << " ... done" << std::endl;
    } // loop over volume snippets

    // next are the skeleton snippets; it works basically the same ...
    if (pAskeleton_snippet_matrices != nullptr)
    {
      for (size_t ssn = 0; ssn < dd.number_to_skeletonsnippet.size(); ssn++) // loop over all skeleton snippets
      {
        if (ssn % nthreads_setup != rank)
          continue;

        auto &ss = dd.number_to_skeletonsnippet[ssn]; // get the skeleton snippet
        auto it = ss.begin();                         // points to first v snippet number
        size_t left_volumesnippetnumber = *it;
        ++it;
        size_t right_volumesnippetnumber = *it; // so now we have the numbers of the two volume snippets to left and right
        auto &left_volumesnippet = dd.number_to_volumesnippet[left_volumesnippetnumber];
        auto &right_volumesnippet = dd.number_to_volumesnippet[right_volumesnippetnumber];

        if (verbose > 1)
          std::cout << "Galerkin projection of skeleton snippet " << dd.vs_to_str(ss) << " : " << dd.vs_to_str(left_volumesnippet) << "|" << dd.vs_to_str(right_volumesnippet);
        CMAT &snippetA = *skeleton_snippet_matrices[targetlevel][ssn]; // matrix for skeleton snippet on current level
        if (verbose > 1)
          std::cout << " with matrix of size " << snippetA.N() << "x" << snippetA.M();

        // we need a local to global map for the subdomain basis vectors involved with this snippet
        // this can be computed on the fly for the skeleton snippet
        // NOTE: only dofs of subdomains which are BOTH left and right are involved!
        // WHY? say subdomain p is only to one side, then such a function zero (PU!) on one side and non existent on the other side
        std::map<size_t, std::map<size_t, size_t>> subdomain_local_to_snippet_local;
        size_t nc = 0;
        for (auto p : left_volumesnippet)       // loop over all subdomains intersected by the left volume snippet
          if (right_volumesnippet.count(p) > 0) // p is to left and right of skeleton snippet
            for (size_t k = 0; k < dddm.n_basis_vectors_per_subdomain[p]; k++)
              subdomain_local_to_snippet_local[p][k] = nc++;
        if (verbose > 1)
          std::cout << " nc=" << nc;

        // allocate dense matrix
        std::shared_ptr<DenseMat> pA = std::shared_ptr<DenseMat>(new DenseMat(nc, nc, 0.0)); // make a dense matrix with zeroes
        projected_skeleton_snippet_matrices[targetlevel][ssn] = pA;                          // store the matrix; it will be used on the next coarser level
        DenseMat &projectedA = *pA;                                                          // use reference

        // compute projected entries
        for (auto &colentry1 : subdomain_local_to_snippet_local) // loop over all subdomains intersected by the snippet
          for (auto &colentry2 : colentry1.second)
          {
            auto pcol = colentry1.first;
            auto kcol = colentry2.first;
            auto jprojected = colentry2.second;

            // restrict basis vector [pcol][kcol] to snippet
            CVEC v1(dddm.skeletonsnippetnumber_local_to_global[ssn].size());
            v1 = 0.0;
            for (size_t i = 0; i < dddm.skeletonsnippetnumber_local_to_global[ssn].size(); i++) // loop over all dofs in snippet
            {
              auto iglobal = dddm.skeletonsnippetnumber_local_to_global[ssn][i]; // global dof
              auto it = dddm.subdomainnumber_global_to_local[iglobal].find(pcol);
              if (it != dddm.subdomainnumber_global_to_local[iglobal].end())
                v1[i] = coarse_subdomain_basis_vectors[targetlevel][pcol][kcol][it->second]; // pick entry from basis vector
            }

            // multiply snippet matrix with column basis vector
            CVEC Av1(dddm.skeletonsnippetnumber_local_to_global[ssn].size());
            snippetA.mv(v1, Av1);

            // now multiply with the rows
            for (auto &rowentry1 : subdomain_local_to_snippet_local) // loop over all subdomains intersected by the volume snippet
              for (auto &rowentry2 : rowentry1.second)
              {
                auto prow = rowentry1.first;
                auto krow = rowentry2.first;
                auto iprojected = rowentry2.second;

                // restrict basis vector [prow][krow] to snippet
                CVEC v2(dddm.skeletonsnippetnumber_local_to_global[ssn].size());
                v2 = 0.0;
                for (size_t i = 0; i < dddm.skeletonsnippetnumber_local_to_global[ssn].size(); i++) // loop over all dofs in snippet
                {
                  auto iglobal = dddm.skeletonsnippetnumber_local_to_global[ssn][i]; // global dof
                  auto it = dddm.subdomainnumber_global_to_local[iglobal].find(prow);
                  if (it != dddm.subdomainnumber_global_to_local[iglobal].end())
                    v2[i] = coarse_subdomain_basis_vectors[targetlevel][prow][krow][it->second]; // pick entry from basis vector
                }

                // now we have one entry
                projectedA[iprojected][jprojected] = v2 * Av1;
              }
          }
        if (verbose > 1)
          std::cout << " ... done" << std::endl;
      } // loop over skeleton snippets
    }   // if there are skeleton snippets
  }     // end of function snippet_projection_coarse_to_coarse

  //! \brief get number of basis vectors in subdomain
  size_t
  number_of_basis_vectors(int subdomain)
  {
    if (subdomain < 0 || subdomain >= nsubdomains[0])
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: subdomain out of range.");
    return fine_subdomain_basis_vectors[subdomain].size();
  }

  //! \brief get number of basis vectors in subdomain
  size_t number_of_basis_vectors(int level, int subdomain)
  {
    if (level < 0 || level >= number_of_levels_ - 1)
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: level out of range.");
    if (subdomain < 0 || subdomain >= nsubdomains[level])
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: subdomain out of range.");
    if (level == 0)
      return fine_subdomain_basis_vectors[subdomain].size();
    else
      return coarse_subdomain_basis_vectors[level][subdomain].size();
  }

  //! \brief get number of basis vectors in subdomain
  size_t number_of_kerAB_vectors(int level, int subdomain)
  {
    if (level < 1 || level >= number_of_levels_ - 1)
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: level out of range.");
    if (subdomain < 0 || subdomain >= nsubdomains[level])
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: subdomain out of range.");
    return coarse_kerAB_vectors[level][subdomain].size();
  }

  //! get number of levels
  size_t number_of_levels()
  {
    return number_of_levels_;
  }

  //! get number of levels
  size_t number_of_subdomains(int level)
  {
    if (level < 0 || level >= number_of_levels_ - 1)
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: level out of range.");
    return nsubdomains[level];
  }

  //! get basis vector k in subdomain p
  VEC prolongated_basis_vector(int p, int k)
  {
    if (p < 0 && p >= nsubdomains[0])
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: subdomain out of range.");

    VEC v(dddmhierarchy[0]->subdomainnumber_global_to_local.size()); // the global vector to be returned
    v = 0.0;
    auto n = dddmhierarchy[0]->subdomainnumber_local_to_global[p].size();
    for (size_t i = 0; i < n; i++)
      v[dddmhierarchy[0]->subdomainnumber_local_to_global[p][i]] = fine_subdomain_basis_vectors[p][k][i];
    return v;
  }

  //! get partition of unity in subdomain p
  VEC prolongated_partition_of_unity(int p)
  {
    if (p < 0 && p >= nsubdomains[0])
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: subdomain out of range.");

    VEC v(dddmhierarchy[0]->subdomainnumber_global_to_local.size()); // the global vector to be returned
    v = 0.0;
    auto n = dddmhierarchy[0]->subdomainnumber_local_to_global[p].size();
    for (size_t i = 0; i < n; i++)
      v[dddmhierarchy[0]->subdomainnumber_local_to_global[p][i]] = partition_of_unity[0][p][i];
    return v;
  }

  //! get basis vector k in subdomain p on level l
  VEC prolongated_basis_vector(int l, int p, int k)
  {
    if (l < 0 || l >= number_of_levels_ - 1)
      DUNE_THROW(Dune::Exception, "prolongated_basis_vector: level out of range.");
    if (p < 0 || p >= nsubdomains[l])
      DUNE_THROW(Dune::Exception, "prolongated_basis_vector subdomain out of range.");
    if (k < 0 && k >= coarse_subdomain_basis_vectors[l][p].size())
      DUNE_THROW(Dune::Exception, "prolongated_basis_vector: k out of range.");

    // make the result
    VEC result(dddmhierarchy[0]->subdomainnumber_global_to_local.size()); // the global vector to be returned
    result = 0.0;

    if (l == 0)
    {
      // we are already on the finest level. No prolongation necessary
      auto n = dddmhierarchy[0]->subdomainnumber_local_to_global[p].size();
      for (size_t i = 0; i < n; i++)
        result[dddmhierarchy[0]->subdomainnumber_local_to_global[p][i]] = fine_subdomain_basis_vectors[p][k][i];
      return result;
    }

    // we are on a coarse level that needs prolongation.

    // make a global dof vector (v) on level l containing the basis vector p,k
    CVEC v(dddmhierarchy[l]->subdomainnumber_global_to_local.size()); // the global vector to be prolonged
    v = 0.0;
    {
      auto nsubdomain = dddmhierarchy[l]->subdomainnumber_local_to_global[p].size();
      for (size_t i = 0; i < nsubdomain; i++)
        v[dddmhierarchy[l]->subdomainnumber_local_to_global[p][i]] = coarse_subdomain_basis_vectors[l][p][k][i];
    }

    for (int targetlevel = l - 1; targetlevel >= 0; targetlevel--)
      if (targetlevel > 0)
      {
        // intermediate level
        CVEC w(dddmhierarchy[targetlevel]->subdomainnumber_global_to_local.size()); // intermediate vector on this level
        w = 0.0;

        // prolongate one level
        for (size_t i = 0; i < v.size(); i++) // loop over all dofs from targetlevel+1
        {
          auto p = dddmhierarchy[targetlevel]->basisvector_global_to_subdomainnumber[i];
          auto k = dddmhierarchy[targetlevel]->basisvector_global_to_local[i];
          // add basis vector to result
          auto n = dddmhierarchy[targetlevel]->subdomainnumber_local_to_global[p].size();
          for (size_t j = 0; j < n; j++)
            w[dddmhierarchy[targetlevel]->subdomainnumber_local_to_global[p][j]] += coarse_subdomain_basis_vectors[targetlevel][p][k][j] * v[i];
        }

        // w is the input on the next level
        v = w;
      }
      else
      {
        // targetlevel==0; prolongate to result
        for (size_t i = 0; i < v.size(); i++) // loop over all dofs from targetlevel+1
        {
          auto p = dddmhierarchy[targetlevel]->basisvector_global_to_subdomainnumber[i];
          auto k = dddmhierarchy[targetlevel]->basisvector_global_to_local[i];
          // add basis vector to result
          auto n = dddmhierarchy[targetlevel]->subdomainnumber_local_to_global[p].size();
          for (size_t j = 0; j < n; j++)
          {
            auto blub = fine_subdomain_basis_vectors[p][k][j];
            blub *= v[i][0];
            result[dddmhierarchy[targetlevel]->subdomainnumber_local_to_global[p][j]] += blub;
          }
        }
      }

    return result;
  }

  //! get basis vector k in subdomain p on level l
  VEC prolongated_kerAB_vector(int l, int p, int k)
  {
    if (l < 1 || l >= number_of_levels_ - 1)
      DUNE_THROW(Dune::Exception, "prolongated_kerAB_vector: level out of range.");
    if (p < 0 || p >= nsubdomains[l])
      DUNE_THROW(Dune::Exception, "prolongated_kerAB_vector subdomain out of range.");
    if (k < 0 && k >= coarse_kerAB_vectors[l][p].size())
      DUNE_THROW(Dune::Exception, "prolongated_kerAB_vector: k out of range.");

    // make the result
    VEC result(dddmhierarchy[0]->subdomainnumber_global_to_local.size()); // the global vector to be returned
    result = 0.0;

    // make a global dof vector (v) on level l containing the basis vector p,k
    CVEC v(dddmhierarchy[l]->subdomainnumber_global_to_local.size()); // the global vector to be prolonged
    v = 0.0;
    {
      auto nsubdomain = dddmhierarchy[l]->subdomainnumber_local_to_global[p].size();
      for (size_t i = 0; i < nsubdomain; i++)
        v[dddmhierarchy[l]->subdomainnumber_local_to_global[p][i]] = coarse_kerAB_vectors[l][p][k][i];
    }

    for (int targetlevel = l - 1; targetlevel >= 0; targetlevel--)
      if (targetlevel > 0)
      {
        // intermediate level
        CVEC w(dddmhierarchy[targetlevel]->subdomainnumber_global_to_local.size()); // intermediate vector on this level
        w = 0.0;

        // prolongate one level
        for (size_t i = 0; i < v.size(); i++) // loop over all dofs from targetlevel+1
        {
          auto p = dddmhierarchy[targetlevel]->basisvector_global_to_subdomainnumber[i];
          auto k = dddmhierarchy[targetlevel]->basisvector_global_to_local[i];
          // add basis vector to result
          auto n = dddmhierarchy[targetlevel]->subdomainnumber_local_to_global[p].size();
          for (size_t j = 0; j < n; j++)
            w[dddmhierarchy[targetlevel]->subdomainnumber_local_to_global[p][j]] += coarse_subdomain_basis_vectors[targetlevel][p][k][j] * v[i];
        }

        // w is the input on the next level
        v = w;
      }
      else
      {
        // targetlevel==0; prolongate to result
        for (size_t i = 0; i < v.size(); i++) // loop over all dofs from targetlevel+1
        {
          auto p = dddmhierarchy[targetlevel]->basisvector_global_to_subdomainnumber[i];
          auto k = dddmhierarchy[targetlevel]->basisvector_global_to_local[i];
          // add basis vector to result
          auto n = dddmhierarchy[targetlevel]->subdomainnumber_local_to_global[p].size();
          for (size_t j = 0; j < n; j++)
          {
            auto blub = fine_subdomain_basis_vectors[p][k][j];
            blub *= v[i][0];
            result[dddmhierarchy[targetlevel]->subdomainnumber_local_to_global[p][j]] += blub;
          }
        }
      }

    return result;
  }

  //! \brief export information from domain decomposition
  const std::vector<std::set<size_t>> &get_overlappingsubdomains(int l)
  {
    return ddhierarchy[l]->overlappingsubdomains;
  }

  //! \brief export information from domain decomposition
  const std::vector<size_t> &get_element_to_volumesnippetnumber(int l)
  {
    return ddhierarchy[l]->element_to_volumesnippetnumber;
  }

  //! \brief export information from domain decomposition
  const std::vector<size_t> &get_partition(int l)
  {
    return ddhierarchy[l]->partition;
  }

  //! \brief export information from domain decomposition
  const std::vector<size_t> &get_vizpartition(int l)
  {
    return ddhierarchy[l]->vizpartition;
  }

  /*!
    \brief Prepare the preconditioner.
  */
  virtual void pre(VEC &x, VEC &b)
  {
    prec_total_time = 0.0;
  }

  /*!
    \brief Apply the precondioner.
  */
  virtual void apply(VEC &v, const VEC &d)
  {
    if (cycle == "mras") // multiplicative over levels, RAS within levels
    {
      // we start with the fine level subdomain solves
      Dune::Timer prec_timer;
#ifdef MULTITHREADED
      for (size_t p = 0; p < nsubdomains[0]; ++p)
        task_queue.push(p);
      for (int rank = 0; rank < nthreads_solve; ++rank)
        threads[rank] = std::thread{&MultiLevelGenEO<MAT, VEC>::fine_subdomain_solve_task, this, rank, std::ref(v), std::cref(d)};
      for (int rank = 0; rank < nthreads_solve; ++rank)
        threads[rank].join();
#else
      for (size_t p = 0; p < nsubdomains[0]; ++p)
        fine_subdomain_solve(0, p, v, d);
#endif
      auto dt = prec_timer.elapsed();
      prec_total_time += dt;

      // // we start with the fine level subdomain solves
      // for (size_t p = 0; p < nsubdomains[0]; p++)
      // {
      //   // dof blocks in subdomain p
      //   auto n = dddmhierarchy[0]->subdomainnumber_local_to_global[p].size();

      //   // set up right hand side in subdomain
      //   VEC dlocal(n);
      //   for (size_t i = 0; i < n; i++)
      //   {
      //     dlocal[i] = d[dddmhierarchy[0]->subdomainnumber_local_to_global[p][i]];
      //     dlocal[i] *= subdomain_boundary_mask[0][p][i];
      //   }

      //   // solve subdomain problem
      //   Dune::InverseOperatorResult stat;
      //   VEC vlocal(n);
      //   vlocal = 0.0;
      //   finesubdomainsolvers[p]->apply(vlocal, dlocal, stat);

      //   // multiply with partition of unity -> restricted additive schwarz
      //   for (size_t i = 0; i < n; i++)
      //     vlocal[i] *= partition_of_unity[0][p][i];

      //   // accumulate correction from subdomain
      //   for (size_t i = 0; i < n; i++)
      //     v[dddmhierarchy[0]->subdomainnumber_local_to_global[p][i]] += vlocal[i];
      // }
      // // now the fine subdomains are completed

      // update defect
      VEC d2(d);
      pAglobal->mmv(v, d2);

      // restrict defect to level 1; defect vectors have been allocated already
      // due to different types, restriction level 0->1 has to be handled seperately
      {
        *dcoarse[1] = 0.0; // clear defect on level 1
        DomainDecompositionDOFMapper &finedddm = *dddmhierarchy[0];
        for (size_t p = 0; p < nsubdomains[0]; p++)                                   // loop over subdomains one level finer
          for (size_t k = 0; k < finedddm.basisvector_local_to_global[p].size(); k++) // loop over all basis vectors
          {
            auto icoarse = finedddm.basisvector_local_to_global[p][k];   // associated global degree of freedom on the coarser level
            auto n = finedddm.subdomainnumber_local_to_global[p].size(); // size of the basis vector
            for (size_t i = 0; i < n; i++)
              (*dcoarse[1])[icoarse] += d2[finedddm.subdomainnumber_local_to_global[p][i]] * fine_subdomain_basis_vectors[p][k][i];
          }
      }

      // perform downward part of V-cycle
      for (int targetlevel = 1; targetlevel < number_of_levels_ - 1; targetlevel++)
      {
        // info for that level
        DomainDecompositionDOFMapper &dddm = *dddmhierarchy[targetlevel];

        // clear global correction on this level
        *vcoarse[targetlevel] = 0.0;

        // and perform subdomain solves
        for (size_t p = 0; p < nsubdomains[targetlevel]; p++)
        {
          // solve in subdomain p
          auto n = dddm.subdomainnumber_local_to_global[p].size();

          // set up right hand side in subdomain (pick out)
          CVEC dlocal(n);
          for (size_t i = 0; i < n; i++)
          {
            dlocal[i] = (*dcoarse[targetlevel])[dddm.subdomainnumber_local_to_global[p][i]]; // this is a picking out operator
            dlocal[i] *= subdomain_boundary_mask[targetlevel][p][i];                         // the Dirichlet condition
          }

          // solve subdomain problem
          Dune::InverseOperatorResult stat;
          CVEC vlocal(n);
          vlocal = 0.0;
          coarsesubdomainsolvers[targetlevel][p]->apply(vlocal, dlocal, stat);

          // multiply with partition of unity -> restricted additive schwarz
          for (size_t i = 0; i < n; i++)
            vlocal[i] *= partition_of_unity[targetlevel][p][i];

          // accumulate correction from subdomain to global correction on this level
          for (size_t i = 0; i < n; i++)
            (*vcoarse[targetlevel])[dddm.subdomainnumber_local_to_global[p][i]] += vlocal[i]; // this is the extension operator
        }

        // update defect on that level
        global_coarse_matrix[targetlevel]->mmv(*vcoarse[targetlevel], *dcoarse[targetlevel]);

        // restrict defect to next coarser level (which must exist)
        *dcoarse[targetlevel + 1] = 0.0;                                          // clear defect on targetlevel
        for (size_t p = 0; p < nsubdomains[targetlevel]; p++)                     // loop over subdomains
          for (size_t k = 0; k < dddm.basisvector_local_to_global[p].size(); k++) // loop over all basis vectors in that subdomain
          {
            auto icoarse = dddm.basisvector_local_to_global[p][k];   // associated global degree of freedom on the coarser level
            auto n = dddm.subdomainnumber_local_to_global[p].size(); // size of the basis vector
            for (size_t i = 0; i < n; i++)
              (*dcoarse[targetlevel + 1])[icoarse] += (*dcoarse[targetlevel])[dddm.subdomainnumber_local_to_global[p][i]] * coarse_subdomain_basis_vectors[targetlevel][p][k][i];
          }
      }

      // coarsest level solve
      {
        // the coarsest level; there is only one subdomain
        auto targetlevel = number_of_levels_ - 1;
        Dune::InverseOperatorResult stat;
        *vcoarse[targetlevel] = 0.0;
        globalcoarsesolvers[targetlevel]->apply(*vcoarse[targetlevel], *dcoarse[targetlevel], stat);
      }

      // upward part of V-cycle
      for (int sourcelevel = number_of_levels_ - 1; sourcelevel > 1; sourcelevel--) // loop over remaining levels
      {
        DomainDecompositionDOFMapper &finedddm = *dddmhierarchy[sourcelevel - 1];     // targetlevel
        for (size_t p = 0; p < nsubdomains[sourcelevel - 1]; p++)                     // loop over subdomains one level finer
          for (size_t k = 0; k < finedddm.basisvector_local_to_global[p].size(); k++) // loop over all basis vectors
          {
            auto icoarse = finedddm.basisvector_local_to_global[p][k];   // associated global degree of freedom on the coarser level
            auto n = finedddm.subdomainnumber_local_to_global[p].size(); // size of the basis vector
            for (size_t i = 0; i < n; i++)
              (*vcoarse[sourcelevel - 1])[finedddm.subdomainnumber_local_to_global[p][i]] += (*vcoarse[sourcelevel])[icoarse][0] * coarse_subdomain_basis_vectors[sourcelevel - 1][p][k][i];
          }
      }
      // prolongation 1(=sourcelevel)->0
      {
        DomainDecompositionDOFMapper &finedddm = *dddmhierarchy[0];                   // targetlevel
        for (size_t p = 0; p < nsubdomains[0]; p++)                                   // loop over subdomains one level finer
          for (size_t k = 0; k < finedddm.basisvector_local_to_global[p].size(); k++) // loop over all basis vectors
          {
            auto icoarse = finedddm.basisvector_local_to_global[p][k];   // associated global degree of freedom on the coarser level
            auto n = finedddm.subdomainnumber_local_to_global[p].size(); // size of the basis vector
            for (size_t i = 0; i < n; i++)
              v[finedddm.subdomainnumber_local_to_global[p][i]] += (*vcoarse[1])[icoarse][0] * fine_subdomain_basis_vectors[p][k][i];
          }
      }

      return; // end of mras cycle
    }

    //===========================================
    // now the default cycle form: fully additive
    //===========================================

    solver_timer.reset();

    // we start with the fine level subdomain solves
    Dune::Timer prec_timer;
#ifdef MULTITHREADED
    for (size_t p = 0; p < nsubdomains[0]; ++p)
      task_queue.push(p);
    for (int rank = 0; rank < nthreads_solve; ++rank)
      threads[rank] = std::thread{&MultiLevelGenEO<MAT, VEC>::fine_subdomain_solve_task, this, rank, std::ref(v), std::cref(d)};
    for (int rank = 0; rank < nthreads_solve; ++rank)
      threads[rank].join();
#else
    for (size_t p = 0; p < nsubdomains[0]; ++p)
      fine_subdomain_solve(0, p, v, d);
#endif
    auto dt = prec_timer.elapsed();
    prec_total_time += dt;

    // for (size_t p = 0; p < nsubdomains[0]; p++)
    //   fine_subdomain_solve(0,p,v,d);

    // {
    //   // dof blocks in subdomain p
    //   auto n = dddmhierarchy[0]->subdomainnumber_local_to_global[p].size();

    //   // set up right hand side in subdomain
    //   VEC dlocal(n);
    //   for (size_t i = 0; i < n; i++)
    //   {
    //     dlocal[i] = d[dddmhierarchy[0]->subdomainnumber_local_to_global[p][i]];
    //     dlocal[i] *= subdomain_boundary_mask[0][p][i];
    //   }

    //   // solve subdomain problem
    //   Dune::InverseOperatorResult stat;
    //   VEC vlocal(n);
    //   vlocal = 0.0;
    //   finesubdomainsolvers[p]->apply(vlocal, dlocal, stat);

    //   // multiply with partition of unity -> restricted additive schwarz
    //   if (cycle == "aras")
    //     for (size_t i = 0; i < n; i++)
    //       vlocal[i] *= partition_of_unity[0][p][i];

    //   // accumulate correction from subdomain
    //   for (size_t i = 0; i < n; i++)
    //     v[dddmhierarchy[0]->subdomainnumber_local_to_global[p][i]] += vlocal[i];
    // }
    // now the fine subdomains are completed

    // restrict defect to all coarse levels; defect vectors have been allocated already
    // due to different types, restriction level 0->1 has to be handled seperately
    {
      *dcoarse[1] = 0.0; // clear defect on level 1
      DomainDecompositionDOFMapper &finedddm = *dddmhierarchy[0];
      for (size_t p = 0; p < nsubdomains[0]; p++)                                   // loop over subdomains one level finer
        for (size_t k = 0; k < finedddm.basisvector_local_to_global[p].size(); k++) // loop over all basis vectors
        {
          auto icoarse = finedddm.basisvector_local_to_global[p][k];   // associated global degree of freedom on the coarser level
          auto n = finedddm.subdomainnumber_local_to_global[p].size(); // size of the basis vector
          for (size_t i = 0; i < n; i++)
            (*dcoarse[1])[icoarse] += d[finedddm.subdomainnumber_local_to_global[p][i]] * fine_subdomain_basis_vectors[p][k][i];
        }
    }
    for (int targetlevel = 2; targetlevel < number_of_levels_; targetlevel++) // loop over remaining levels
    {
      *dcoarse[targetlevel] = 0.0; // clear defect on targetlevel
      DomainDecompositionDOFMapper &finedddm = *dddmhierarchy[targetlevel - 1];
      for (size_t p = 0; p < nsubdomains[targetlevel - 1]; p++)                     // loop over subdomains one level finer
        for (size_t k = 0; k < finedddm.basisvector_local_to_global[p].size(); k++) // loop over all basis vectors
        {
          auto icoarse = finedddm.basisvector_local_to_global[p][k];   // associated global degree of freedom on the coarser level
          auto n = finedddm.subdomainnumber_local_to_global[p].size(); // size of the basis vector
          for (size_t i = 0; i < n; i++)
            (*dcoarse[targetlevel])[icoarse] += (*dcoarse[targetlevel - 1])[finedddm.subdomainnumber_local_to_global[p][i]] * coarse_subdomain_basis_vectors[targetlevel - 1][p][k][i];
        }
    }

    // solve subdomain problems
    for (int targetlevel = 1; targetlevel < number_of_levels_; targetlevel++) // loop over remaining levels
      if (targetlevel == number_of_levels_ - 1)
      {
        // the coarsest level
        Dune::InverseOperatorResult stat;
        *vcoarse[targetlevel] = 0.0;
        globalcoarsesolvers[targetlevel]->apply(*vcoarse[targetlevel], *dcoarse[targetlevel], stat);
      }
      else
      {
        // clear global correection on this level
        *vcoarse[targetlevel] = 0.0;

        // an intermediate level
        for (size_t p = 0; p < nsubdomains[targetlevel]; p++)
        {
          // solve in subdomain p
          DomainDecompositionDOFMapper &dddm = *dddmhierarchy[targetlevel];
          auto n = dddm.subdomainnumber_local_to_global[p].size();

          // set up right hand side in subdomain (pick out)
          CVEC dlocal(n);
          for (size_t i = 0; i < n; i++)
          {
            dlocal[i] = (*dcoarse[targetlevel])[dddm.subdomainnumber_local_to_global[p][i]]; // this is a picking out operator
            dlocal[i] *= subdomain_boundary_mask[targetlevel][p][i];                         // the Dirichlet condition
          }

          // solve subdomain problem
          Dune::InverseOperatorResult stat;
          CVEC vlocal(n);
          vlocal = 0.0;
          coarsesubdomainsolvers[targetlevel][p]->apply(vlocal, dlocal, stat);

          // multiply with partition of unity -> restricted additive schwarz
          if (cycle == "aras")
            for (size_t i = 0; i < n; i++)
              vlocal[i] *= partition_of_unity[targetlevel][p][i];

          // accumulate correction from subdomain to global correction on this level
          for (size_t i = 0; i < n; i++)
            (*vcoarse[targetlevel])[dddm.subdomainnumber_local_to_global[p][i]] += vlocal[i]; // this is the extension operator
        }
      }

    // now we have all the correction on the coarse levels in global vectors
    // prolongation
    for (int sourcelevel = number_of_levels_ - 1; sourcelevel > 1; sourcelevel--) // loop over remaining levels
    {
      DomainDecompositionDOFMapper &finedddm = *dddmhierarchy[sourcelevel - 1];     // targetlevel
      for (size_t p = 0; p < nsubdomains[sourcelevel - 1]; p++)                     // loop over subdomains one level finer
        for (size_t k = 0; k < finedddm.basisvector_local_to_global[p].size(); k++) // loop over all basis vectors
        {
          auto icoarse = finedddm.basisvector_local_to_global[p][k];   // associated global degree of freedom on the coarser level
          auto n = finedddm.subdomainnumber_local_to_global[p].size(); // size of the basis vector
          for (size_t i = 0; i < n; i++)
            (*vcoarse[sourcelevel - 1])[finedddm.subdomainnumber_local_to_global[p][i]] += (*vcoarse[sourcelevel])[icoarse][0] * coarse_subdomain_basis_vectors[sourcelevel - 1][p][k][i];
        }
    }
    // prolongation 1(=sourcelevel)->0
    {
      DomainDecompositionDOFMapper &finedddm = *dddmhierarchy[0];                   // targetlevel
      for (size_t p = 0; p < nsubdomains[0]; p++)                                   // loop over subdomains one level finer
        for (size_t k = 0; k < finedddm.basisvector_local_to_global[p].size(); k++) // loop over all basis vectors
        {
          auto icoarse = finedddm.basisvector_local_to_global[p][k];   // associated global degree of freedom on the coarser level
          auto n = finedddm.subdomainnumber_local_to_global[p].size(); // size of the basis vector
          for (size_t i = 0; i < n; i++)
            v[finedddm.subdomainnumber_local_to_global[p][i]] += (*vcoarse[1])[icoarse][0] * fine_subdomain_basis_vectors[p][k][i];
        }
    }

  } // end of apply

  double get_prec_total_time() const
  {
    return prec_total_time;
  }

  // grab as many subdomains as possible and process them
  void fine_subdomain_solve_task(size_t rank, VEC &v, const VEC &d)
  {
    while (true)
    {
      size_t p; // the task to process in this iteration
      {
        std::lock_guard<std::mutex> lock{task_mutex}; // get exclusive access to task queue
        if (task_queue.empty())
          break;                // exit loop
        p = task_queue.front(); // read task
        task_queue.pop();       // remove task from queue
        if (verbose > 1)
          std::cout << "thread " << rank << " processing subdomain " << p << std::endl;
      }                                    // free lock
      fine_subdomain_solve(rank, p, v, d); // process task
    }
  }

  void fine_subdomain_solve(size_t rank, size_t p, VEC &v, const VEC &d)
  {
    // dof blocks in subdomain p
    auto n = dddmhierarchy[0]->subdomainnumber_local_to_global[p].size();

    // set up right hand side in subdomain
    VEC dlocal(n);
    for (size_t i = 0; i < n; i++)
    {
      dlocal[i] = d[dddmhierarchy[0]->subdomainnumber_local_to_global[p][i]];
      dlocal[i] *= subdomain_boundary_mask[0][p][i];
    }

    // solve subdomain problem
    Dune::InverseOperatorResult stat;
    VEC vlocal(n);
    vlocal = 0.0;
    finesubdomainsolvers[p]->apply(vlocal, dlocal, stat);

    // multiply with partition of unity -> restricted additive schwarz
    if (cycle == "aras" || cycle == "mras")
      for (size_t i = 0; i < n; i++)
        vlocal[i] *= partition_of_unity[0][p][i];

    // accumulate correction from subdomain
    std::lock_guard<std::mutex> lock{update_mutex}; // get exclusive access to global vector
    for (size_t i = 0; i < n; i++)
      v[dddmhierarchy[0]->subdomainnumber_local_to_global[p][i]] += vlocal[i];
  }

  /*!
    \brief Clean up.
  */
  virtual void post(VEC &x)
  {
  }

private:
  EigenproblemType eigenproblem_type_for_string(const std::string &type) {
    if (type == "geneo")
      return EigenproblemType::GenEO;
    else if (type == "aharmonic")
      return EigenproblemType::AHarmonic;
    else
      DUNE_THROW(Dune::NotImplemented, "Eigenproblem type '" + type + "' not implemented");
  }
};

#endif // HAVE_SUITESPARSE_UMFPACK || DOXYGEN

#endif // Udune_ftworth_HH
