#ifndef Udune_ftworth_schwarz_preconditioners_HH
#define Udune_ftworth_schwarz_preconditioners_HH

#include <queue>
#include <thread>
#include <mutex>

#include <dune/istl/umfpack.hh>

#include "matrixutilities.hh"

#define MULTITHREADED 1

// converts bool information in boundary dofs to floating point type
// true -> 0.0
// false -> 1.0
// but this is only for blocks being on the boundary and not for individual dofs in that block.
// Thus the name is a bit misleading
template <typename field_type>
std::vector<std::vector<field_type>>
get_boundary_mask(std::shared_ptr<std::vector<std::vector<size_t>>> local_to_global,
                  std::shared_ptr<std::vector<std::vector<bool>>> boundary_dofs)
{
  auto subdomains = local_to_global->size();
  std::vector<std::vector<field_type>> boundary_mask(subdomains);
  for (unsigned p = 0; p < subdomains; p++)
  {
    auto n = (*local_to_global)[p].size(); // dof blocks in subdomain p
    boundary_mask[p].resize(n);
    for (unsigned i = 0; i < n; i++)
      boundary_mask[p][i] = ((*boundary_dofs)[p][i]) ? 0.0 : 1.0;
  }
  return boundary_mask;
}

template <typename field_type>
std::vector<std::vector<field_type>>
get_partition_of_unity_standard(std::shared_ptr<std::vector<std::vector<size_t>>> local_to_global,
                                std::shared_ptr<std::vector<std::map<size_t, size_t>>> global_to_local,
                                std::shared_ptr<std::vector<std::vector<bool>>> boundary_dofs)
{
  // number of subdomains we have
  auto subdomains = local_to_global->size();

  // make a global vector holding the number of subdomains where dof is NOT on the boundary
  std::vector<int> k(global_to_local->size(), 0);
  for (unsigned p = 0; p < subdomains; p++)
    for (unsigned i = 0; i < (*local_to_global)[p].size(); i++)
      if (!(*boundary_dofs)[p][i])
        k[(*local_to_global)[p][i]] += 1;

  // now compute partition of unity
  std::vector<std::vector<field_type>> pu(subdomains);
  for (unsigned p = 0; p < subdomains; p++)
  {
    pu[p].resize((*local_to_global)[p].size());
    for (unsigned i = 0; i < (*local_to_global)[p].size(); i++)
      pu[p][i] = ((*boundary_dofs)[p][i]) ? 0.0 : (((field_type)1.0) / k[(*local_to_global)[p][i]]);
  }

  // return
  return pu;
}

#define DOXYGEN 1 // remove this again later

#if HAVE_SUITESPARSE_UMFPACK || DOXYGEN

/** \brief Two-level additive Schwarz preconditioner with "extended Nicolaides" coarse space
 *
 * \author Peter Bastian
 *
 * \tparam MAT a "Dune::BCRSMatrix<Dune::FieldMatrix>" type
 * \tparam VEC a "Dune::BlockVector<Dune::FieldVector>" type
 *
 */
template <typename MAT, typename VEC>
class ExtendedNicolaides : public Dune::Preconditioner<VEC, VEC>
{
  using VectorBlockType = typename VEC::block_type;
  using CBLOCKVEC = Dune::FieldVector<typename VEC::field_type, 1>;
  using CBLOCKMAT = Dune::FieldMatrix<typename VEC::field_type, 1, 1>;
  using CMAT = Dune::BCRSMatrix<CBLOCKMAT>;  // type for coarse level matrix
  using CVEC = Dune::BlockVector<CBLOCKVEC>; // type for coarse level vector

public:
  //! \brief The domain type of the preconditioner.
  using domain_type = VEC;
  //! \brief The range type of the preconditioner.
  using range_type = VEC;
  //! \brief The field type of the preconditioner.
  using field_type = typename VEC::field_type;

  // define the category
  virtual Dune::SolverCategory::Category category() const
  {
    return Dune::SolverCategory::sequential;
  }

private: // the local data of this class
  // input arguments or derived from that
  unsigned subdomains;      // the number of subdomains
  const unsigned blocksize; // size of the blocks in the input vectors
  std::shared_ptr<MAT> pAglobal;
  std::shared_ptr<VEC> global_dirichlet_dofs;                                // (*global_dirichlet_dofs)[i][j] is >0 if component j in block i is at the global dirichlet boundary
  std::shared_ptr<std::vector<std::vector<size_t>>> local_to_global;         // map local index to global index in each subdomain
  std::shared_ptr<std::vector<std::map<size_t, size_t>>> global_to_local;    // necessary for partition of unity
  std::shared_ptr<std::vector<std::vector<bool>>> subdomain_boundary_blocks; // flag for each degree of freedom BLOCK whether it is on the subdomain boundary
  std::string cycle;                                                         // type of preconditioner: single two ras mras
  int nthreads_setup;                                                        // number of threads used in setup
  int nthreads_solve;                                                        // number of threads used in the solver

  // things computed by the constructor for later use
  unsigned n_coarse;                                                 // total number of degrees of freedom in the coarse problem
  std::vector<std::vector<field_type>> boundary_mask;                // boundary_mask[p][i] is 0 if dof block i in subdomain p is at the subdomain boundary, otherwise it is 1
  std::vector<std::vector<field_type>> partition_of_unity;           // partition_of_unity[p][i] is *one* value for dof block i in subdomain p
  std::vector<std::vector<size_t>> coarse_vector_local_to_global;    // coarse_vector_local_to_global[p][i] gives the global dof number in the coarse problem of vector i in subdomain p
  std::vector<std::shared_ptr<Dune::UMFPack<MAT>>> subdomainsolvers; // the LU decomposition of the subdomain problems
  std::vector<std::vector<VEC>> coarse_basis_vectors;                // coarse_basis_vectors[p][k] gives the kth coarse grid vector in subdomain p
  std::shared_ptr<CMAT> pcmat;                                       // the coarse level matrix
  std::shared_ptr<Dune::UMFPack<CMAT>> coarselevelsolver;            // LU decomposition of coarse level matrix
  std::vector<std::set<size_t>> subdomain_graph;                     // subdomain_graph[p] gives the subdomains having joint degrees of freedom with subdomain p, including p
  std::queue<size_t> task_queue;                                     // queue of subdomain numbers
  std::mutex task_mutex;                                             // mutex to guard access to task queue
  std::mutex update_mutex;                                           // another mutex
#ifdef MULTITHREADED
  std::vector<std::thread> threads; // thread pool
#endif

public:
  //! \brief get number of basis vectors in subdomain
  size_t number_of_basis_vectors(int subdomain)
  {
    if (subdomain < 0 && subdomain >= subdomains)
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: subdomain out of range.");
    return coarse_basis_vectors[subdomain].size();
  }

  //! get basis vector k in subdomain p
  VEC prolongated_basis_vector(int p, int k)
  {
    if (p < 0 && p >= subdomains)
      DUNE_THROW(Dune::Exception, "number_of_basis_vectors: subdomain out of range.");

    VEC v(global_to_local->size()); // the global vector to be returned
    v = 0.0;
    auto n = (*local_to_global)[p].size();
    for (size_t i = 0; i < n; i++)
      v[(*local_to_global)[p][i]] = coarse_basis_vectors[p][k][i];
    return v;
  }

  //! \brief Constructor. It does all the work of setting up the components
  /**
   * \param pAglobal_ shared_ptr to the global system matrix
   * \param input_coarse_vectors input_coarse_vectors_ : a vector of pointers to coefficient vectors representing functions to be included in the coarse space, e.g. constants and linears
   * \param global_dirichlet_dofs_ global_dirichlet_dofs_[i][j] is 0 if global dof block i, component j is at Dirichlet boundary, 1 else
   * \param local_to_global_ (*local_to_global_)[p][i] maps dof block number i in subdomain p to its global number, component number is not changed and therefore not included
   * \param global_to_local_ (*global_to_local_)[i][p] maps global dof block number i to its local dof block number in subdomain p. Caution: may only be called if global dof i exists in subdomain p, otherwise use find!
   * \param subdomain_boundary_blocks_ (*subdomain_boundary_blocks_)[p][i] is true if dof block i in subdomain p is at the Dirichlet boundary
   * \param cycle_  different cycle forms: single (additive, no coarse grid), two (two level additive Schwarz), ras (restricted additive in subdomain no coarse), aras (ras + add coarse corr.), mras (ras with multiplicative coarse grid)
   */
  ExtendedNicolaides(std::shared_ptr<MAT> pAglobal_,
                     std::vector<std::shared_ptr<VEC>> input_coarse_vectors,
                     std::shared_ptr<VEC> global_dirichlet_dofs_,
                     std::shared_ptr<std::vector<std::vector<size_t>>> local_to_global_,
                     std::shared_ptr<std::vector<std::map<size_t, size_t>>> global_to_local_,
                     std::string cycle_,
                     int nthreads_setup_,
                     int nthreads_solve_)
      : subdomains(local_to_global_->size()),
        blocksize(VectorBlockType::dimension),
        pAglobal(pAglobal_),
        global_dirichlet_dofs(global_dirichlet_dofs_),
        local_to_global(local_to_global_),
        global_to_local(global_to_local_),
        cycle(cycle_),
        coarse_vector_local_to_global(local_to_global_->size()),
        subdomainsolvers(local_to_global_->size()),
        coarse_basis_vectors(local_to_global_->size()),
#ifdef MULTITHREADED
        threads(std::max(nthreads_setup_, nthreads_solve_)), // create non joinable threads
#endif
        nthreads_setup(nthreads_setup_),
        nthreads_solve(nthreads_solve_)
  {
    Dune::Timer timer_total;
    timer_total.reset();

    std::cout << "preparing two level preconditioner ... " << std::endl;

    // now we know where the Dirichlet constraints for subdomain boundaries are ...
    // the "boundary dofs" are actually boundary blocks and they are determined from the global stiffness matrix
    // by checking whether a complete row is available in one subdomain.
    // So this function is just about a dof being on a *subdomain* boundary (as the name implies) but it totally disregards Dirichlet boundary conditions!
    subdomain_boundary_blocks = std::make_shared<std::vector<std::vector<bool>>>(get_subdomain_boundary_dofs(*pAglobal, *local_to_global, *global_to_local));
    std::cout << "calculated boundary dofs" << std::endl;

    // compute mask vector to null out corrections on subdomain boundary
    boundary_mask = get_boundary_mask<field_type>(local_to_global, subdomain_boundary_blocks);

    // make a partition of unity
    partition_of_unity = get_partition_of_unity_standard<field_type>(local_to_global, global_to_local, subdomain_boundary_blocks);

    // make the coarse problem; this requires several steps

    // 1) build subdomain graph
    subdomain_graph.resize(subdomains);                                                           // subdomain_graph[p] gives the subdomains having joint degrees of freedom with subdomain p, including p
    for (size_t j = 0; j < global_to_local->size(); j++)                                          // loop over global dof numbers
      for (auto itp = (*global_to_local)[j].begin(); itp != (*global_to_local)[j].end(); ++itp)   // loop of subdomains sharing this degree of freedom
        for (auto itq = (*global_to_local)[j].begin(); itq != (*global_to_local)[j].end(); ++itq) // loop of subdomains sharing this degree of freedom
          subdomain_graph[itp->first].insert(itq->first);
    std::cout << "subdomain graph done" << std::endl;

    // 2) create the coarse vectors in each subdomain by
    //     a) picking out entries from the global input vectors
    //     b) multiplying with the partition of unity
    //     c) nulling out dofs on the global dirchlet boundary
    // this way, a prolongation to the global Dirichlet boundary will produce a zero contribution
    //
    // this is the only step that needs to be replaced by the eigenvector computation for GenEO!
    //
    auto n_coarse_basis_vectors_per_subdomain = input_coarse_vectors.size(); // here the number of coarse vectors in each subdomain is the same
    for (unsigned p = 0; p < subdomains; p++)
    {
      // create the subdomain vectors
      auto n = (*local_to_global)[p].size(); // size of the coarse vectors in subdomain p (in blocks)
      VEC vec(n);                            // make a template
      vec = 0.0;
      coarse_basis_vectors[p].resize(n_coarse_basis_vectors_per_subdomain, vec); // create the subdomain istl vectors; now coarse_basis_vectors[p][k] is an ISTL vector on subdomain p

      // fill them
      for (size_t k = 0; k < n_coarse_basis_vectors_per_subdomain; k++)
      {
        // pick out values from input vector
        for (size_t i = 0; i < n; i++)
          coarse_basis_vectors[p][k][i] = (*input_coarse_vectors[k])[(*local_to_global)[p][i]]; // pick out

        // normalize
        if (k == 0)
        {
          // the first is assumed to be the constant. Just rescale to norm 1
          coarse_basis_vectors[p][k] *= 1.0 / coarse_basis_vectors[p][k].two_norm();
        }
        else
        {
          // if it is not the first then we want to orthogonalize wrt. the constant. Let v_k be the current vector and v_0 the first (representing the constant).
          // Then seek z_k such that
          // so 1) z_k = beta*(v_k - alpha*v_0)
          //    2) z_k*v_0 = 0 => z_k*v_0/beta = 0 = (v_k - alpha*v_0)*v_0 => v_k*v_0 - alpha*v_0*v_0 = 0 => alpha = v_k*v_0/(v_0*v_0)
          //    3) z_k*z_k = 1
          auto alpha = coarse_basis_vectors[p][k] * coarse_basis_vectors[p][0];      // we assume that v_0 is already normalized
          coarse_basis_vectors[p][k].axpy(-alpha, coarse_basis_vectors[p][0]);       // orthogonalize
          coarse_basis_vectors[p][k] *= 1.0 / coarse_basis_vectors[p][k].two_norm(); // normalize
        }
      }

      // multiply with partition of unity and zero on Dirichlet boundary
      for (size_t k = 0; k < n_coarse_basis_vectors_per_subdomain; k++)
        for (size_t i = 0; i < n; i++)
        {
          auto iglobal = (*local_to_global)[p][i];
          coarse_basis_vectors[p][k][i] *= partition_of_unity[p][i]; // multiply with partition of unity, puts zero on subdomain boundary
          for (size_t j = 0; j < blocksize; j++)
            coarse_basis_vectors[p][k][i][j] *= (*global_dirichlet_dofs)[iglobal][j]; // zero out dof on global Dirichlet boundary
        }
    }

    // 3) create subdomain to global mapping for the coarse problem, i.e. from subdomain and local coarse vector number to global number of coarse vectors
    // from here on this should be generic in the number of coarse vectors per subdomain
    std::vector<size_t> coarse_vector_global_to_subdomain;
    std::vector<size_t> coarse_vector_global_to_local;
    for (unsigned p = 0; p < subdomains; p++)
    {
      coarse_vector_local_to_global[p].resize(coarse_basis_vectors[p].size()); // resize to number of basis vectors in this subdomain
      for (size_t k = 0; k < coarse_basis_vectors[p].size(); k++)              // loop over all coarse vectors in this subdomain
      {
        coarse_vector_global_to_subdomain.push_back(p);                                     // this coarse vector is in subdomain p; push_back assigns the global number!
        coarse_vector_global_to_local.push_back(k);                                         // it has the local number k; push_back assigns the same global number!
        coarse_vector_local_to_global[p][k] = coarse_vector_global_to_subdomain.size() - 1; // could also use coarse_vector_global_to_local.size()-1
      }
    }
    std::cout << coarse_vector_global_to_subdomain.size() << " coarse grid basis vectors" << std::endl;

    // 4) create the sparse coarse level matrix
    // This matrix in general has a different type than the input matrices since the blocksize is always 1 regardless of the blocksize of the input matrices.
    n_coarse = coarse_vector_global_to_subdomain.size(); // now we know the size of the coarse problem
    unsigned nz_coarse = 0;                              // compute number of nonzeroes
    for (unsigned p = 0; p < subdomain_graph.size(); p++)
      for (auto q : subdomain_graph[p])
        nz_coarse += coarse_basis_vectors[p].size() * coarse_basis_vectors[q].size();
    std::cout << nz_coarse << " nonzeroes in coarse level matrix" << std::endl;
    pcmat = std::shared_ptr<CMAT>(new CMAT(n_coarse, n_coarse, nz_coarse, CMAT::row_wise)); // allocate matrix in rowwise creation mode
    for (auto row = pcmat->createbegin(); row != pcmat->createend(); ++row)                 // fill nonzeroes
    {
      auto i = row.index();                                         // this is the global row index
      auto p = coarse_vector_global_to_subdomain[i];                // the subdomain this coarse vector comes from
      for (auto q : subdomain_graph[p])                             // loop over neighboring subdomains
        for (size_t k = 0; k < coarse_basis_vectors[q].size(); k++) // loop over coarse vectors in neighboring subdomain
        {
          auto j = coarse_vector_local_to_global[q][k]; // the global number of this coarse vector
          row.insert(j);
        }
    }

    // 5) compute entries of coarse level matrix
#ifdef MULTITHREADED
    std::cout << "Galerkin projection and factorization using " << nthreads_setup << " threads" << std::endl;

    for (size_t prow = 0; prow < subdomains; prow++) // loop over all subdomains (row index)
      task_queue.push(prow);
    for (int rank = 0; rank < nthreads_setup; ++rank)
      threads[rank] = std::thread{&ExtendedNicolaides<MAT, VEC>::subdomain_setup_task_dynamic, this, rank};
    for (int rank = 0; rank < nthreads_setup; ++rank)
      threads[rank].join();

      // for (int rank = 0; rank < nthreads_setup; ++rank)
      //   threads[rank] = std::thread{&ExtendedNicolaides<MAT, VEC>::subdomain_setup2, this, nthreads_setup, rank};
      // for (int rank = 0; rank < nthreads_setup; ++rank)
      //   threads[rank].join();
#else
    for (unsigned prow = 0; prow < subdomains; prow++) // loop over all subdomains (row index)
      subdomain_setup(0, prow);
#endif
    std::cout << "computed entries of coarse system" << std::endl;

    // now we may compute the LU decomposition of the coarse system
    std::cout << "LU decomposition of coarse system" << std::endl;
    coarselevelsolver = std::shared_ptr<Dune::UMFPack<CMAT>>(new Dune::UMFPack<CMAT>(*pcmat, false));
    std::cout << "... done" << std::endl;
    std::cout << "preconditioner ready, cycle=" << cycle << std::endl;

    auto setup_time_total = timer_total.elapsed();
    std::cout << "TIME SETUP PHASE total=" << setup_time_total << std::endl;
  }

  // grab as many subdomains as possible and process them
  void subdomain_setup_task_dynamic(size_t rank)
  {
    while (true)
    {
      size_t p; // the task to process in this iteration
      {
        std::lock_guard<std::mutex> lock{task_mutex}; // get exclusive access to task queue
        if (task_queue.empty())
          break;                // exit loop
        p = task_queue.front(); // read task
        task_queue.pop();       // remove task from queue
      } // free lock
      subdomain_setup(rank, p); // process task
    }
  }

  // handle complete coarse system setup and factorize subdomain matrix for subdomain prow in thread #rank
  void subdomain_setup_task_static(size_t rank)
  {
    for (unsigned prow = 0; prow < subdomains; prow++) // loop over all subdomains (row index)
      if (prow % nthreads_setup == rank)
        subdomain_setup(rank, prow);
  }

  // handle complete coarse system setup and factorize subdomain matrix for subdomain prow in thread #rank
  void subdomain_setup(size_t rank, size_t prow)
  {
    // create subdomain matrix from global matrix (copied from set_up_subdomain_matrices)
    int avg = pAglobal->nonzeroes() / pAglobal->N() + 2;                      // estimate for nonzeroes per row
    size_t n = (*local_to_global)[prow].size();                               // number of DOFs in subdomain prow
    auto pAsubdomain = std::make_shared<MAT>(n, n, avg, 0.15, MAT::implicit); // a new sparse matrix
    for (size_t i = 0; i < n; i++)
    {
      auto iglobal = (*local_to_global)[prow][i];
      auto cIt = (*pAglobal)[iglobal].begin();
      auto cEndIt = (*pAglobal)[iglobal].end();
      for (; cIt != cEndIt; ++cIt)
      {
        auto jglobal = cIt.index();
        auto it = (*global_to_local)[jglobal].find(prow); // lets see if column index is also in this subdomain
        if (it != (*global_to_local)[jglobal].end())
        {
          auto j = it->second; // global to local
          pAsubdomain->entry(i, j) = *cIt;
        }
      }
    }
    auto stats = pAsubdomain->compress();

    bool fl = assemble_dirichlet_constraints_in_subdomain((*local_to_global)[prow], *global_dirichlet_dofs, *pAsubdomain);

    // now that we have it already, compute factorization of subdomain matrix
    subdomainsolvers[prow] = std::make_shared<Dune::UMFPack<MAT>>(*pAsubdomain, false);

    // now do the triple matrix products
    VEC v1(pAsubdomain->N()); // two vectors on the subdomain to compute scalar products
    VEC v2(pAsubdomain->N());
    for (size_t krow = 0; krow < coarse_basis_vectors[prow].size(); krow++) // loop over all coarse grid functions in this subdomain
    {
      // row number in coarse system
      auto icoarse = coarse_vector_local_to_global[prow][krow];

      // matrix vector product
      // need to use transposed matrix here to work also for nonsymmetric matrices
      pAsubdomain->mtv(coarse_basis_vectors[prow][krow], v1);

      // now do all the scalar products with the other basis vectors
      for (auto pcol : subdomain_graph[prow])                                   // loop over all neighboring subdomains (including myself)
        for (size_t kcol = 0; kcol < coarse_basis_vectors[pcol].size(); kcol++) // loop over all coarse grid functions in the neighboring subdomain
        {
          // column number in coarse system
          auto jcoarse = coarse_vector_local_to_global[pcol][kcol];

          // produce restriction of basis vector from subdomain pcol to subdomain prow
          v2 = 0.0; // zero out
          for (size_t i = 0; i < pAsubdomain->N(); i++)
          {
            auto iglobal = (*local_to_global)[prow][i];       // global dof block for our local index
            auto it = (*global_to_local)[iglobal].find(pcol); // test if this global dof block exists in neighboring subdomain
            if (it != (*global_to_local)[iglobal].end())      // if yes, we have a common entry
            {
              auto index_in_other_subdomain = it->second;
              v2[i] = coarse_basis_vectors[pcol][kcol][index_in_other_subdomain]; // restriction of coarse vector in subdomain pcol in subdomain prow
            }
          }

          // now we have a matrix entry
          (*pcmat)[icoarse][jcoarse] = v2 * v1;
        }
    }
  }

  /*!
    \brief Prepare the preconditioner.
  */
  virtual void pre(VEC &x, VEC &b)
  {
  }

  /*!
    \brief Apply the precondioner.
  */
  virtual void apply(VEC &v, const VEC &d)
  {
    // do not need to zero out v as it is already zero

    // first we solve in each subdomain. This is independend of how the coarse space is treated
#ifdef MULTITHREADED
    task_queue = std::queue<size_t>(); // should be empty anyway
    for (unsigned p = 0; p < subdomains; p++)
      task_queue.push(p);
    for (int rank = 0; rank < nthreads_solve; ++rank)
      threads[rank] = std::thread{&ExtendedNicolaides<MAT, VEC>::subdomain_solve_task_dynamic, this, rank, std::ref(v), std::cref(d)};
    for (int rank = 0; rank < nthreads_solve; ++rank)
      threads[rank].join();
#else
    for (unsigned p = 0; p < subdomains; p++)
      subdomain_solve(0, p, v, d);
#endif

    // check if coarse level correction is requested
    if (cycle != "two" && cycle != "aras" && cycle != "mras")
      return;

    // update defect
    // unfortunately d is const, so we cant overwrite it
    VEC d2(d);
    if (cycle == "mras")
      pAglobal->mmv(v, d2);

    // prepare and solve coarse problem
    // the coarse grid defect
    CVEC dcoarse(n_coarse);
    dcoarse = 0.0;

    // restriction
    for (unsigned p = 0; p < subdomains; p++)
      for (size_t k = 0; k < coarse_basis_vectors[p].size(); k++)
      {
        auto icoarse = coarse_vector_local_to_global[p][k];
        auto n = (*local_to_global)[p].size();
        for (size_t i = 0; i < n; i++)
          dcoarse[icoarse] += d2[(*local_to_global)[p][i]] * coarse_basis_vectors[p][k][i];
      }

    // solve subdomain problem
    Dune::InverseOperatorResult stat;
    CVEC vcoarse(n_coarse);
    vcoarse = 0.0;
    coarselevelsolver->apply(vcoarse, dcoarse, stat);

    // prolongation
    for (unsigned p = 0; p < subdomains; p++)
      for (size_t k = 0; k < coarse_basis_vectors[p].size(); k++)
      {
        auto icoarse = coarse_vector_local_to_global[p][k];
        auto n = (*local_to_global)[p].size();
        for (size_t i = 0; i < n; i++)
          v[(*local_to_global)[p][i]] += coarse_basis_vectors[p][k][i] * vcoarse[icoarse][0];
      }
  }

  // grab as many subdomains as possible and process them
  void subdomain_solve_task_dynamic(size_t rank, VEC &v, const VEC &d)
  {
    while (true)
    {
      size_t p; // the task to process in this iteration
      {
        std::lock_guard<std::mutex> lock{task_mutex}; // get exclusive access to task queue
        if (task_queue.empty())
          break;                // exit loop
        p = task_queue.front(); // read task
        task_queue.pop();       // remove task from queue
      } // free lock
      subdomain_solve(rank, p, v, d); // process task
    }
  }

  void subdomain_solve(size_t rank, size_t p, VEC &v, const VEC &d)
  {
    // dof blocks in subdomain p
    auto n = (*local_to_global)[p].size();

    // set up right hand side in subdomain
    VEC dlocal(n);
    for (size_t i = 0; i < n; i++)
    {
      dlocal[i] = d[(*local_to_global)[p][i]];
      dlocal[i] *= boundary_mask[p][i];
    }

    // solve subdomain problem
    Dune::InverseOperatorResult stat;
    VEC vlocal(n);
    vlocal = 0.0;
    subdomainsolvers[p]->apply(vlocal, dlocal, stat);

    // multiply with partition of unity -> restricted additive schwarz
    if (cycle == "ras" || cycle == "aras" || cycle == "mras")
      for (size_t i = 0; i < n; i++)
        vlocal[i] *= partition_of_unity[p][i];

    // accumulate correction from subdomain
    std::lock_guard<std::mutex> lock{update_mutex}; // get exclusive access to global vector
    for (size_t i = 0; i < n; i++)
      v[(*local_to_global)[p][i]] += vlocal[i];
  }

  /*!
    \brief Clean up.
  */
  virtual void post(VEC &x)
  {
  }
};

#endif // HAVE_SUITESPARSE_UMFPACK || DOXYGEN

#endif // Udune_ftworth_HH
