#ifndef Udune_ftworth_convectiondiffusionproblemcheckerboard_HH
#define Udune_ftworth_convectiondiffusionproblemcheckerboard_HH

#include<dune/pdelab/localoperator/convectiondiffusionparameter.hh>

/** Parameter class for the stationary convection-diffusion equation of the following form:
 *
 * \f{align*}{
 *   \nabla\cdot(-A(x) \nabla u + b(x) u) + c(x)u &=& f \mbox{ in } \Omega,  \ \
 *                                              u &=& g \mbox{ on } \partial\Omega_D (Dirichlet)\ \
 *                (b(x) u - A(x)\nabla u) \cdot n &=& j \mbox{ on } \partial\Omega_N (Flux)\ \
 *                        -(A(x)\nabla u) \cdot n &=& o \mbox{ on } \partial\Omega_O (Outflow)
 * \f}
 * Note:
 *  - This formulation is valid for velocity fields which are non-divergence free.
 *  - Outflow boundary conditions should only be set on the outflow boundary
 *
 * The template parameters are:
 *  - GV a model of a GridView
 *  - RF numeric type to represent results
 */

template<typename GV, typename RF>
class ConvectionDiffusionProblem
{
  typedef Dune::PDELab::ConvectionDiffusionBoundaryConditions::Type BCType;
  const bool convection;
  Dune::ParameterTree& ptree;
  const RF D1;             // alternating values in checkerboard pattern
  const RF D2 = 1.0;
  const int nx = 8;        // no. of rectangles in x- and y-direction
  const int ny = 8;

public:
  typedef Dune::PDELab::ConvectionDiffusionParameterTraits<GV,RF> Traits;

  ConvectionDiffusionProblem() : convection(true), D1(1e-6) {}
  ConvectionDiffusionProblem(const bool convection_) : convection(convection_), D1(1e-6) {}
  ConvectionDiffusionProblem(Dune::ParameterTree& ptree_) : ptree(ptree_), convection(true), D1(ptree.get<double>("problem.D",(double)1e-6)) {}
  ConvectionDiffusionProblem(Dune::ParameterTree& ptree_, const bool convection_) : ptree(ptree_), convection(convection_), D1(ptree.get<double>("problem.D",(double)1e-6)) {}

  //! tensor diffusion constant per cell? return false if you want more than one evaluation of A per cell.
  static constexpr bool permeabilityIsConstantPerCell()
  {
    return true;
  }

  //! tensor diffusion coefficient
  typename Traits::PermTensorType
  A (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::DomainType xglobal = e.geometry().global(x);
    RF Dglobal = D1;
    if ((((int)(xglobal[0]*nx) % 2 != 0) && ((int)(xglobal[1]*ny) % 2 != 0))
        || (((int)(xglobal[0]*nx) % 2 == 0) && ((int)(xglobal[1]*ny) % 2 == 0)))
        Dglobal = D2;

    typename Traits::PermTensorType I;
    for (std::size_t i=0; i<Traits::dimDomain; i++)
      for (std::size_t j=0; j<Traits::dimDomain; j++)
        I[i][j] = (i==j) ? Dglobal : 0;
    return I;
  }

  //! velocity field
  typename Traits::RangeType
  b (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::RangeType v(0.0);
    if (convection)
    {
      //typename Traits::DomainType xglobal = e.geometry().global(x);
      v[0] = 2./3.;
      v[1] = 1.;
    }
    return v;
  }

  //! sink term
  typename Traits::RangeFieldType
  c (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! source term
  typename Traits::RangeFieldType
  f (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    //return 1.0;
    return 0.0;
  }

  //! boundary condition type function
  /* return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Dirichlet for Dirichlet boundary conditions
   * return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Neumann for flux boundary conditions
   * return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Outflow for outflow boundary conditions
   */
  BCType
  bctype (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    //return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Dirichlet;
    typename Traits::DomainType xglobal = is.geometry().global(x);
    if (xglobal[0]<1e-6) return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Dirichlet;
    if (xglobal[1]<1e-6) return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Dirichlet;
    //if (xglobal[1]>1.0-1e-6) return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Dirichlet;
    //if (xglobal[1]>3.0-1e-6) return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Outflow;
    //if (xglobal[1]<1e-6) return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Outflow;
    return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Outflow;
  }

  //! Dirichlet boundary condition value
  typename Traits::RangeFieldType
  g (const typename Traits::ElementType& e, const typename Traits::DomainType& xlocal) const
  {
    //return 100.0;
    //return 0.0;
    typename Traits::DomainType xglobal = e.geometry().global(xlocal);
    //return (1.0-xglobal[0])*(1.0-xglobal[0]) + xglobal[1]*xglobal[1]*xglobal[1];
    if (xglobal[0]<1e-6) return 1.0;
    if (xglobal[1]<1e-6) return 0.0;
    //if (xglobal[1]<1.0-1e-6) return 0.0;
    return 0.0;
  }

  //! flux boundary condition
  typename Traits::RangeFieldType
  j (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }

  //! outflow boundary condition
  typename Traits::RangeFieldType
  o (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
    /*typename Traits::DomainType xglobal = is.geometry().global(x);
    if (xglobal[1]<1e-6) return 1.0;
    return 1.0;*/
  }
};

#endif // Udune_ftworth_convectiondiffusionproblemcheckerboard_HH
