#ifndef Udune_ftworth_convectiondiffusionproblemgeneo_HH
#define Udune_ftworth_convectiondiffusionproblemgeneo_HH

#include<dune/pdelab/localoperator/convectiondiffusionparameter.hh>

/** Parameter class for the stationary convection-diffusion equation of the following form:
 *
 * \f{align*}{
 *   \nabla\cdot(-A(x) \nabla u + b(x) u) + c(x)u &=& f \mbox{ in } \Omega,  \ \
 *                                              u &=& g \mbox{ on } \partial\Omega_D (Dirichlet)\ \
 *                (b(x) u - A(x)\nabla u) \cdot n &=& j \mbox{ on } \partial\Omega_N (Flux)\ \
 *                        -(A(x)\nabla u) \cdot n &=& o \mbox{ on } \partial\Omega_O (Outflow)
 * \f}
 * Note:
 *  - This formulation is valid for velocity fields which are non-divergence free.
 *  - Outflow boundary conditions should only be set on the outflow boundary
 *
 * The template parameters are:
 *  - GV a model of a GridView
 *  - RF numeric type to represent results
 */

template<typename GV, typename RF>
class ConvectionDiffusionProblem
{
  // examples taken from Bootland et al. (2023): https://doi.org/10.1093/imanum/drac036
  typedef Dune::PDELab::ConvectionDiffusionBoundaryConditions::Type BCType;
  Dune::ParameterTree& ptree;
  const bool convection;
  // factor in velocity fields
  const RF b_cst;
  // factor scaling the decay of the "point" source
  const RF source_decay = 1e-1;
  const RF source_mag = 1e0;
  // skyscraper pattern
  const RF a_min = 1.0;    // minimal diffusion coefficient (outside of skyscrapers and inclusions)
  const RF a_max;          // maximal diffusion coefficient
  const RF a_sky;
  const int nx = 9;        // no. of rectangles in x- and y-direction
  const int ny = 9;
  // inclusion 1
  const RF incl1_x_left = 0.15;      // lower left corner
  const RF incl1_y_left = 0.63;
  const RF incl1_x_right = 0.72;     // lower right corner
  const RF incl1_y_right = 0.42;
  const RF incl1_width = 0.045;       // width of inclusion
  const RF incl1_m = (incl1_y_right - incl1_y_left)/(incl1_x_right - incl1_x_left);
  const RF incl1_n = (incl1_x_right*incl1_y_left - incl1_x_left*incl1_y_right)/(incl1_x_right - incl1_x_left);
  // inclusion 2
  const RF incl2_x_left = 0.22;      // lower left corner
  const RF incl2_y_left = 0.03;
  const RF incl2_x_right = 0.94;     // lower right corner
  const RF incl2_y_right = 0.15;
  const RF incl2_width = 0.08;       // width of inclusion
  const RF incl2_m = (incl2_y_right - incl2_y_left)/(incl2_x_right - incl2_x_left);
  const RF incl2_n = (incl2_x_right*incl2_y_left - incl2_x_left*incl2_y_right)/(incl2_x_right - incl2_x_left);

public:
  typedef Dune::PDELab::ConvectionDiffusionParameterTraits<GV,RF> Traits;

  ConvectionDiffusionProblem() : convection(true), b_cst(1e4), a_max(5.0), a_sky(a_max/10.0) {}
  ConvectionDiffusionProblem(const bool convection_) : convection(convection_), b_cst(1e4), a_max(5.0), a_sky(a_max/10.0) {}
  ConvectionDiffusionProblem(Dune::ParameterTree& ptree_) : ptree(ptree_), convection(true), b_cst(ptree.get<double>("problem.b_cst",(double)1e4)), a_max(ptree.get<double>("problem.a_max",(double)5)), a_sky(a_max/10.0) {}
  ConvectionDiffusionProblem(Dune::ParameterTree& ptree_, const bool convection_) : ptree(ptree_), convection(convection_), b_cst(ptree.get<double>("problem.b_cst",(double)1e4)), a_max(ptree.get<double>("problem.a_max",(double)5)), a_sky(a_max/10.0) {}

  //! tensor diffusion constant per cell? return false if you want more than one evaluation of A per cell.
  static constexpr bool permeabilityIsConstantPerCell()
  {
    return true;
  }

  //! tensor diffusion coefficient
  typename Traits::PermTensorType
  A (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::DomainType xglobal = e.geometry().global(x);
    // set to minimum
    RF Aglobal = a_min;

    /*// skyscraper pattern
    if (((int)(xglobal[0]*nx) % 2 == 0) && ((int)(xglobal[1]*ny) % 2 == 0))
        Aglobal = a_sky*((int)(xglobal[1]*ny) + 1.0);

    // inclusion 2
    if (isInInclusion2(xglobal))
        Aglobal = 0.5*a_max;

    // inclusion 1
    if (isInInclusion1(xglobal))
        Aglobal = a_max;*/

    // set value
    typename Traits::PermTensorType I;
    for (std::size_t i=0; i<Traits::dimDomain; i++)
      for (std::size_t j=0; j<Traits::dimDomain; j++)
        I[i][j] = (i==j) ? Aglobal : 0;
    return I;
  }

  //! velocity field
  typename Traits::RangeType
  b (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::RangeType v(0.0);
    if (convection)
    {
      typename Traits::DomainType xglobal = e.geometry().global(x);
      // (5.3), divergence-free
      RF factor = b_cst*(1 + sin(2*M_PI*(2*xglobal[1] - xglobal[0])));
      // (5.4), non-vanishing divergence
      // RF factor = b_cst*(1 + sin(2*M_PI*(2*xglobal[0] + xglobal[1])));
      v[0] = 2.0*factor;
      v[1] = 1.0*factor;
    }
    return v;
  }

  //! sink term
  typename Traits::RangeFieldType
  c (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! source term
  typename Traits::RangeFieldType
  f (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    //return 0.0;
    typename Traits::DomainType xglobal = e.geometry().global(x);
    return source_mag*exp(-((xglobal[0] - 0.5)*(xglobal[0] - 0.5) + (xglobal[1] - 0.5)*(xglobal[1] - 0.5))/source_decay);
  }

  //! boundary condition type function
  /* return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Dirichlet for Dirichlet boundary conditions
   * return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Neumann for flux boundary conditions
   * return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Outflow for outflow boundary conditions
   */
  BCType
  bctype (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    /*typename Traits::DomainType xglobal = is.geometry().global(x);
    if (xglobal[0]>1-1e-6) return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Outflow;
    if (xglobal[1]>1-1e-6) return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Outflow;*/
    return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Dirichlet;
  }

  //! Dirichlet boundary condition value
  typename Traits::RangeFieldType
  g (const typename Traits::ElementType& e, const typename Traits::DomainType& xlocal) const
  {
    /*typename Traits::DomainType xglobal = e.geometry().global(xlocal);
    if (xglobal[0]<1e-6) return 1.0;*/
    return 0.0;
  }

  //! flux boundary condition
  typename Traits::RangeFieldType
  j (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }

  //! outflow boundary condition
  typename Traits::RangeFieldType
  o (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }

  bool isInInclusion(const RF x_left, const RF x_right, const RF width, const RF m, const RF n, const typename Traits::DomainType& xglobal) const {
    RF yline = m*xglobal[0] + n;
    return ((xglobal[0] >= x_left) && (xglobal[0] <= x_right)
        && (xglobal[1] >= yline) && (xglobal[1] <= yline + width));
  }

  bool isInInclusion1(const typename Traits::DomainType& xglobal) const {
    return isInInclusion(incl1_x_left, incl1_x_right, incl1_width, incl1_m, incl1_n, xglobal);
  }

  bool isInInclusion2(const typename Traits::DomainType& xglobal) const {
    return isInInclusion(incl2_x_left, incl2_x_right, incl2_width, incl2_m, incl2_n, xglobal);
  }
};

#endif // Udune_ftworth_convectiondiffusionproblemgeneo_HH
