#ifndef Udune_ftworth_convectiondiffusionproblemrotating_HH
#define Udune_ftworth_convectiondiffusionproblemrotating_HH

#include<dune/pdelab/localoperator/convectiondiffusionparameter.hh>

/** Parameter class for the stationary convection-diffusion equation of the following form:
 *
 * \f{align*}{
 *   \nabla\cdot(-A(x) \nabla u + b(x) u) + c(x)u &=& f \mbox{ in } \Omega,  \ \
 *                                              u &=& g \mbox{ on } \partial\Omega_D (Dirichlet)\ \
 *                (b(x) u - A(x)\nabla u) \cdot n &=& j \mbox{ on } \partial\Omega_N (Flux)\ \
 *                        -(A(x)\nabla u) \cdot n &=& o \mbox{ on } \partial\Omega_O (Outflow)
 * \f}
 * Note:
 *  - This formulation is valid for velocity fields which are non-divergence free.
 *  - Outflow boundary conditions should only be set on the outflow boundary
 *
 * The template parameters are:
 *  - GV a model of a GridView
 *  - RF numeric type to represent results
 */

template<typename GV, typename RF>
class ConvectionDiffusionProblem
{
  typedef Dune::PDELab::ConvectionDiffusionBoundaryConditions::Type BCType;
  const bool convection;
  Dune::ParameterTree& ptree;
  const RF twoSigmaSquared = 1e-2;//0.004;
  const RF x_c = 0.25;//0.25;//-0.25;
  const RF y_c = 0.5;//0.0;
  // factor in velocity fields
  const RF b_cst;
  const RF D1;             // alternating values in checkerboard pattern
  const RF D2 = 1.0;
  const int nx = 32;        // no. of rectangles in x- and y-direction
  const int ny = 32;

public:
  typedef Dune::PDELab::ConvectionDiffusionParameterTraits<GV,RF> Traits;

  ConvectionDiffusionProblem() : convection(true), D1(1e-4), b_cst(1.0) {}
  ConvectionDiffusionProblem(const bool convection_) : convection(convection_), D1(1e-4), b_cst(1.0) {}
  ConvectionDiffusionProblem(Dune::ParameterTree& ptree_) : ptree(ptree_), convection(true), D1(ptree.get<double>("problem.D",(double)1e-4)), b_cst(ptree.get<double>("problem.b_cst",(double)1.0)) {}
  ConvectionDiffusionProblem(Dune::ParameterTree& ptree_, const bool convection_) : ptree(ptree_), convection(convection_), D1(ptree.get<double>("problem.D",(double)1e-4)), b_cst(ptree.get<double>("problem.b_cst",(double)1.0)) {}

  //! tensor diffusion constant per cell? return false if you want more than one evaluation of A per cell.
  static constexpr bool permeabilityIsConstantPerCell()
  {
    return true;
  }

  //! tensor diffusion coefficient
  typename Traits::PermTensorType
  A (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    /*typename Traits::PermTensorType I;
    for (std::size_t i=0; i<Traits::dimDomain; i++)
      for (std::size_t j=0; j<Traits::dimDomain; j++)
        I[i][j] = (i==j) ? D1 : 0;
    return I;*/
    typename Traits::DomainType xglobal = e.geometry().global(x);
    RF Dglobal = D1;
    if ((((int)(xglobal[0]*nx) % 2 != 0) && ((int)(xglobal[1]*ny) % 2 != 0))
        || (((int)(xglobal[0]*nx) % 2 == 0) && ((int)(xglobal[1]*ny) % 2 == 0)))
        Dglobal = D2;

    typename Traits::PermTensorType I;
    for (std::size_t i=0; i<Traits::dimDomain; i++)
      for (std::size_t j=0; j<Traits::dimDomain; j++)
        I[i][j] = (i==j) ? Dglobal : 0;
    return I;
  }

  //! velocity field
  typename Traits::RangeType
  b (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    typename Traits::RangeType v(0.0);
    if (convection)
    {
      typename Traits::DomainType xglobal = e.geometry().global(x);
      /*v[0] = -4*xglobal[1];
      v[1] = 4*xglobal[0];*/
      xglobal -= 0.5;
      // (5.6)
      v[0] = b_cst*2*xglobal[1]*(1 - xglobal[0]*xglobal[0]);
      v[1] = -b_cst*2*xglobal[0]*(1 - xglobal[1]*xglobal[1]);
      v[0] -= 0.15*b_cst*xglobal[0];
      v[1] -= 0.15*b_cst*xglobal[1];
      // (5.7)
      // v *= sin(8*M_PI*(1 - xglobal[0]*xglobal[0])*(1 - xglobal[1]*xglobal[1]));
    }
    return v;
  }

  //! sink term
  typename Traits::RangeFieldType
  c (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    return 0.0;
  }

  //! source term
  typename Traits::RangeFieldType
  f (const typename Traits::ElementType& e, const typename Traits::DomainType& x) const
  {
    /*typename Traits::DomainType x_global = e.geometry().global(x);
    typename Traits::RangeFieldType u_global = g(e, x);
    // div(bu)*twoSigmaSquared/u
    typename Traits::RangeFieldType div1 = 8*x_global[1]*(x_global[0] - x_c) - 8*x_global[0]*(x_global[1] - y_c);
    // div(D1 \nabla u)*twoSigmaSquared/u
    typename Traits::RangeFieldType div2 = 4*D1*((x_global[0] - x_c)*(x_global[0] - x_c)/twoSigmaSquared
                                        + (x_global[1] - y_c)*(x_global[1] - y_c)/twoSigmaSquared
                                        - 1.0);
    return (div1 - div2)*u_global/twoSigmaSquared;*/
    typename Traits::DomainType xglobal = e.geometry().global(x);
    return 1.0*exp(-((xglobal[0] - x_c)*(xglobal[0] - x_c) + (xglobal[1] - y_c)*(xglobal[1] - y_c))/twoSigmaSquared);

  }

  //! boundary condition type function
  /* return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Dirichlet for Dirichlet boundary conditions
   * return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Neumann for flux boundary conditions
   * return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Outflow for outflow boundary conditions
   */
  BCType
  bctype (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return Dune::PDELab::ConvectionDiffusionBoundaryConditions::Dirichlet;
  }

  //! Dirichlet boundary condition value
  typename Traits::RangeFieldType
  g (const typename Traits::ElementType& e, const typename Traits::DomainType& xlocal) const
  {
    return 0.0;
    /*typename Traits::DomainType xglobal = e.geometry().global(xlocal);
    return exp(-((xglobal[0] - x_c)*(xglobal[0] - x_c) + (xglobal[1] - y_c)*(xglobal[1] - y_c))/twoSigmaSquared);*/
  }

  //! flux boundary condition
  typename Traits::RangeFieldType
  j (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }

  //! outflow boundary condition
  typename Traits::RangeFieldType
  o (const typename Traits::IntersectionType& is, const typename Traits::IntersectionDomainType& x) const
  {
    return 0.0;
  }
};

#endif // Udune_ftworth_convectiondiffusionproblemrotating_HH
