// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

// always include the config file
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
// C++ includes
#include<math.h>
#include<iostream>
#include<vector>
#include<set>
#include<sstream>
// dune-common includes
#include<dune/common/parallel/mpihelper.hh>
#include<dune/common/parametertreeparser.hh>
#include<dune/common/timer.hh>
#include <dune/common/exceptions.hh> // We use exceptions
// dune-istl includes
#include<dune/istl/preconditioners.hh>
// dune-geometry includes
#include<dune/geometry/referenceelements.hh>
#include<dune/geometry/quadraturerules.hh>
// dune-grid includes
#include<dune/grid/onedgrid.hh>
#include<dune/grid/io/file/vtk/vtkwriter.hh>
#include<dune/grid/io/file/gmshreader.hh>
#if HAVE_UG || HAVE_DUNE_UGGRID
#include<dune/grid/uggrid.hh>
#endif
#if HAVE_DUNE_ALUGRID
#include<dune/alugrid/grid.hh>
#include<dune/alugrid/dgf.hh>
#include<dune/grid/io/file/dgfparser/dgfparser.hh>
#endif
// dune-pdelab includes
#include<dune/pdelab/common/function.hh>
#include<dune/pdelab/common/vtkexport.hh>
#include<dune/pdelab/function/callableadapter.hh>
#include<dune/pdelab/gridfunctionspace/vtk.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include<dune/pdelab/constraints/common/constraints.hh>
// ftworth includes
#include"../dune/ftworth/assemblewrapper.hh"
#include"../dune/ftworth/subdomainutilities.hh"
#include"../dune/ftworth/cdediscretizer.hh"
#include"../dune/ftworth/eigensolver.hh"
#include"../dune/ftworth/multilevel_geneo_preconditioner.hh"
// includes from this directory
//#include"../problems/convectiondiffusionconst.hh"
//#include"../problems/convectiondiffusioncheckerboard.hh"
//#include"../problems/convectiondiffusionrotating.hh"
#include"../problems/convectiondiffusiongeneo.hh"

#define DIMENSION 2
#define DEGREE 1
//#define USECG 1
//#define USEDG 1
#define USEFV 1
//#define UGGRID 1
#define YASPGRID 1
#define USEMLGENEO 1
//#define USEEXTNIC 1
//#define USEILU 1
//#define USENOPREC 1

int main(int argc, char** argv)
{
  try{
    // Maybe initialize MPI
    Dune::MPIHelper& helper = Dune::MPIHelper::instance(argc, argv);
    std::cout << "Hello World! This is dune-ftworth." << std::endl;
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout<<"I am rank "<<helper.rank()<<" of "<<helper.size()
               <<" processes!"<<std::endl;

    // Read parameters from ini file
    Dune::ParameterTree ptree;
    Dune::ParameterTreeParser ptreeparser;
    ptreeparser.readINITree("convectiondiffusion.ini",ptree);
    ptreeparser.readOptions(argc,argv,ptree);
   
#if (HAVE_UG || HAVE_DUNE_UGGRID) // check if we have the UG grid manager

    //===================
    // grid generation and partitioning
    //===================
    
    // load a grid file and refine
#ifdef UGGRID
    using Grid = Dune::UGGrid<DIMENSION>;  
    std::string mshfile = ptree.get("grid.filename","circle.msh");
    std::shared_ptr<Grid> pgrid;
    pgrid = std::shared_ptr<Grid>(Dune::GmshReader<Grid>::read(mshfile,true,false));
#endif
    
    // make a YaspGrid
#ifdef YASPGRID
    //using Grid = Dune::YaspGrid<DIMENSION>;
    using Grid = Dune::YaspGrid<DIMENSION,Dune::EquidistantOffsetCoordinates<double,DIMENSION>>;
    Dune::FieldVector<double,DIMENSION> lower_left;
    lower_left[0] = ptree.get("grid.LLX",(double)0.0);
    lower_left[1] = ptree.get("grid.LLY",(double)0.0);
    if (DIMENSION>2)
      lower_left[2] = ptree.get("grid.LLZ",(double)0.0);
    Dune::FieldVector<double,DIMENSION> upper_right;
    upper_right[0] = lower_left[0] + ptree.get("grid.LX",(double)1.0);
    upper_right[1] = lower_left[1] + ptree.get("grid.LY",(double)1.0);
    if (DIMENSION>2)
      upper_right[2] = lower_left[2] + ptree.get("grid.LZ",(double)1.0);
    std::array<int,DIMENSION> cells;
    cells[0] = ptree.get("grid.NX",(int)10);
    cells[1] = ptree.get("grid.NY",(int)10);
    if (DIMENSION>2)
      cells[2] = ptree.get("grid.NZ",(int)10);
    //std::shared_ptr<Grid> pgrid = std::shared_ptr<Grid>(new Grid(upper_right,cells));
    std::shared_ptr<Grid> pgrid = std::shared_ptr<Grid>(new Grid(lower_left,upper_right,cells));
#endif

    // refine grid
    int refine = ptree.get("grid.refine",(int)0);
    pgrid->globalRefine(ptree.get("grid.refine",refine));

    // extract the leaf grid view
    auto gv = pgrid->leafGridView();
    using GV = decltype(gv); // and get the type
    const int dim = GV::dimension;
    auto& indexset = gv.indexSet(); // to attach data to elements
    std::cout << indexset.size(0) << " elements" << std::endl;
    std::cout << indexset.size(dim) << " vertices" << std::endl;
#ifdef USEMLGENEO
    std::vector<size_t> subdomains = ptree.get<std::vector<size_t>>("geneo.subdomains");
#endif
#ifdef USEEXTNIC
    std::vector<size_t> subdomains = ptree.get<std::vector<size_t>>("extendednicolaides.subdomains");
#endif
#if defined(USEMLGENEO) || defined(USEEXTNIC)
    std::cout << subdomains[0] << " subdomains" << std::endl;
#endif

    //===================
    // set up a discretization scheme
    //===================

    // solve a convection diffusion problem
    using DF = Grid::ctype;
    using RF = double;

    // set up the solver class
    using Problem = ConvectionDiffusionProblem<GV,RF>;
#ifdef USECG
#ifdef UGGRID
    using Discretization = CGConvectionDiffusionProblemSimplex<GV,Problem,DEGREE>;
#endif
#ifdef YASPGRID
    using Discretization = CGConvectionDiffusionProblemCube<GV,Problem,DEGREE>;
#endif
    bool hasSkeleton = false;
#endif
#ifdef USEDG
#ifdef UGGRID
    using VBE = Dune::PDELab::ISTL::VectorBackend<Dune::PDELab::ISTL::Blocking::none,Dune::PB::PkSize<DEGREE,GV::dimension>::value>;
    using Discretization = DGConvectionDiffusionProblemSimplex<GV,Problem,DEGREE,VBE>;
    //using Discretization = DGConvectionDiffusionProblemSimplex<GV,Problem,DEGREE>;
#endif
#ifdef YASPGRID
    using VBE = Dune::PDELab::ISTL::VectorBackend<Dune::PDELab::ISTL::Blocking::none,Dune::QkStuff::QkSize<DEGREE,GV::dimension>::value>;
    using Discretization = DGConvectionDiffusionProblemCube<GV,Problem,DEGREE,VBE>;
    //using Discretization = DGConvectionDiffusionProblemCube<GV,Problem,DEGREE>;
#endif
    bool hasSkeleton = true;
#endif

#ifdef USEFV
#ifndef YASPGRID
    std::cerr << "USEFV only works with YASPGRID" << std::endl;
#endif
    using Discretization = FVConvectionDiffusionProblemCube<GV,Problem>;
    bool hasSkeleton = true;
#endif

    //typename Dune::FieldVector<RF,GV::dimensionworld> b_vec = ptree.get<typename Dune::FieldVector<RF,GV::dimensionworld>>("problem.b_vec");
    //Problem problem(b_vec);
    Problem problem(ptree);
    Discretization disc(gv,problem,ptree);

    //===================
    // set up global matrix and solution vector
    //===================

    // set up linear system in PDELab components
    using VEC = typename Discretization::VEC;
    using ISTLV = Dune::PDELab::Backend::Native<VEC>;
    using MAT = typename Discretization::MAT;
    using ISTLM = Dune::PDELab::Backend::Native<MAT>;

    VEC x(*disc.gfs);
    ISTLV& istlx = Dune::PDELab::Backend::native(x);
    Dune::PDELab::interpolate(*disc.g,*disc.gfs,x);
    /*for (size_t i=0; i<istlx.N(); i++)
      for (size_t j=0; j<istlx[i].N(); j++)
        istlx[i][j] = (rand() % 10000)/10000.0;
    set_constrained_dofs(*disc.cc,0.0,x);*/ // dirichlet dofs are zero, others are 1, so we can use it as a mask!
    
    // asymmetric system matrix, A
    MAT A(*disc.go); // this makes the sparsity pattern
    A = 0.0;
    ISTLM& istlA = Dune::PDELab::Backend::native(A);
    std::shared_ptr<ISTLM> pistlA = stackobject_to_shared_ptr(Dune::PDELab::Backend::native(istlA));

    // residual and correction
    VEC r(*disc.gfs);
    r = 0.0;
    disc.go->residual(x,r);
    ISTLV& istlr = Dune::PDELab::Backend::native(r);
//=====================
// multilevel GenEO
//=====================
#ifdef USEMLGENEO
    // symmetric system matrix, A_plus
    Problem problemA_plus(ptree, false);  // sets convection to false, i.e., b = 0
    Discretization discA_plus(gv,problemA_plus,ptree);
    MAT A_plus(*discA_plus.go); // this makes the sparsity pattern
    A_plus = 0.0;
    ISTLM& istlA_plus = Dune::PDELab::Backend::native(A_plus);
    std::shared_ptr<ISTLM> pistlA_plus = stackobject_to_shared_ptr(Dune::PDELab::Backend::native(istlA_plus));

    size_t overlap = ptree.get("geneo.overlap",(size_t)1);
    std::string extensionmethod = ptree.get("geneo.extensionmethod","vertex");
    size_t drop_small_overlap = ptree.get("geneo.drop_small_overlap",(size_t)1);
    bool coord_partitioning = ptree.get("geneo.coordinate_partitioning",(bool)false);
    std::vector<size_t> parts = ptree.get<std::vector<size_t>>("geneo.parts");

    // make domain decomposition on fine grid
    std::shared_ptr<DomainDecomposition> pdd;
    if (coord_partitioning)
      pdd = std::make_shared<DomainDecomposition>(gv,subdomains[0],parts,overlap,extensionmethod,drop_small_overlap,false);
    else
      pdd = std::make_shared<DomainDecomposition>(gv,Dune::MPIHelper::getLocalCommunicator(),subdomains[0],overlap,extensionmethod,drop_small_overlap,false);

    //===================
    // build subdomain information & test matrices in subsets
    //==================
    
    auto pdddm = std::shared_ptr<DomainDecompositionDOFMapper>(new DomainDecompositionDOFMapper(*disc.gfs,pdd));

    std::cout << "create sparse matrices for snippets and subdomains" << std::endl;

    // A
    auto pAvolume_snippet_matrices = std::shared_ptr<std::vector<std::shared_ptr<ISTLM>>>( new std::vector<std::shared_ptr<ISTLM>>(set_up_local_matrices(istlA,pdddm->volumesnippetnumber_local_to_global,pdddm->volumesnippetnumber_global_to_local)) );
    for (auto matptr : *pAvolume_snippet_matrices) *matptr = 0.0; // clear subdomain matrices

    std::shared_ptr<std::vector<std::shared_ptr<ISTLM>>> pAskeleton_snippet_matrices; // the empty pointer
    if (hasSkeleton)
      {
        pAskeleton_snippet_matrices = std::shared_ptr<std::vector<std::shared_ptr<ISTLM>>>( new std::vector<std::shared_ptr<ISTLM>>(set_up_local_matrices(istlA,pdddm->skeletonsnippetnumber_local_to_global,pdddm->skeletonsnippetnumber_global_to_local)) );
        for (auto matptr : *pAskeleton_snippet_matrices) *matptr = 0.0; // clear subdomain matrices
      }

    // A_plus
    auto pA_plus_volume_snippet_matrices = std::shared_ptr<std::vector<std::shared_ptr<ISTLM>>>( new std::vector<std::shared_ptr<ISTLM>>(set_up_local_matrices(istlA_plus,pdddm->volumesnippetnumber_local_to_global,pdddm->volumesnippetnumber_global_to_local)) );
    for (auto matptr : *pA_plus_volume_snippet_matrices) *matptr = 0.0; // clear subdomain matrices

    std::shared_ptr<std::vector<std::shared_ptr<ISTLM>>> pA_plus_skeleton_snippet_matrices; // the empty pointer
    if (hasSkeleton)
      {
        pA_plus_skeleton_snippet_matrices = std::shared_ptr<std::vector<std::shared_ptr<ISTLM>>>( new std::vector<std::shared_ptr<ISTLM>>(set_up_local_matrices(istlA_plus,pdddm->skeletonsnippetnumber_local_to_global,pdddm->skeletonsnippetnumber_global_to_local)) );
        for (auto matptr : *pA_plus_skeleton_snippet_matrices) *matptr = 0.0; // clear subdomain matrices
      }

    //===================
    // matrix assembly
    //==================

    std::cout << "assembling subdomain matrices" << std::endl;
    // now we may wrap the local operator
    using WLOP = SnippetAssemblerWrapper<typename Discretization::LOP,ISTLM,typename Discretization::GFS>;
    // A
    WLOP wlopA(*disc.gfs,*disc.lop,pdddm,pAvolume_snippet_matrices,pAskeleton_snippet_matrices);
    // A_plus
    WLOP wlopA_plus(*discA_plus.gfs,*discA_plus.lop,pdddm,pA_plus_volume_snippet_matrices,pA_plus_skeleton_snippet_matrices);

    // and set up a new grid operator
    using WGOP = Dune::PDELab::GridOperator<typename Discretization::GFS,typename Discretization::GFS,WLOP,
                                            typename Discretization::MBE,typename Discretization::RF,typename Discretization::RF,typename Discretization::RF,
                                            typename Discretization::CC,typename Discretization::CC>;
    // A
    WGOP wgopA(*disc.gfs,*disc.cc,*disc.gfs,*disc.cc,wlopA,*disc.mbe);
    // A_plus
    WGOP wgopA_plus(*discA_plus.gfs,*discA_plus.cc,*discA_plus.gfs,*discA_plus.cc,wlopA_plus,*discA_plus.mbe);
    Dune::Timer timer;
    timer.reset();
    wgopA.jacobian(x,A);  // assembles the global stiffness matrix and the snippets
                          // we have global dirichlet constraints in the global matrix A but *not* in the snippets!
    wgopA_plus.jacobian(x,A_plus);  // same for A_plus
    auto assemble_time = timer.elapsed();
    std::cout << "ASSEMBLE TIME=" << assemble_time << std::endl;
    
    // global vector with dirichlet boundary conditions
    // contains 0.0 for dof on Dirichlet boundary, otherwise 1.0
    VEC dirichlet_mask(*disc.gfs);
    dirichlet_mask = 1.0;
    set_constrained_dofs(*disc.cc,0.0,dirichlet_mask); // dirichlet dofs are zero, others are 1, so we can use it as a mask!
    std::shared_ptr<ISTLV> pistldirichlet_mask = stackobject_to_shared_ptr(Dune::PDELab::Backend::native(dirichlet_mask));


    //===================
    // now set up solver and solve
    //==================
    
    size_t coarse_overlap = ptree.get("geneo.coarse_overlap",(size_t)0);
    std::string pum = ptree.get("geneo.pum","standard");
    std::string fineGEVPrhs = ptree.get("geneo.fineGEVPrhs","pu");
    std::string coarseGEVPrhs = ptree.get("geneo.coarseGEVPrhs","pu");
    std::string coarseeigensolver = ptree.get("geneo.coarseeigensolver","eigen");
    std::string cycle = ptree.get("geneo.cycle","additive");
    // size_t n_eigenvectors_fine_computed = ptree.get("geneo.n_eigenvectors_fine_computed",(size_t)10);
    // size_t n_eigenvectors_fine_used = ptree.get("geneo.n_eigenvectors_fine_used",(size_t)5);
    size_t n_eigenvectors_coarse_computed = ptree.get("geneo.n_eigenvectors_coarse_computed",(size_t)10);
    size_t n_eigenvectors_coarse_used = ptree.get("geneo.n_eigenvectors_coarse_used",(size_t)5);
    size_t mlgeneo_verbose = ptree.get("geneo.verbose",(size_t)0);
    double abs_zero_ker = ptree.get("geneo.abs_zero_ker",(double)1e-10);
    double regularization_ker = ptree.get("geneo.regularization_ker",(double)0.0);
    // double arpack_tolerance = ptree.get("geneo.arpack_tolerance",(double)0.0);
    // double shift = ptree.get("geneo.arpack_shift",(double)0.001);
    // double eigenvalue_fine_threshold = ptree.get("geneo.eigenvalue_fine_threshold",(double)-1.0);
    double eigenvalue_coarse_threshold = ptree.get("geneo.eigenvalue_coarse_threshold",(double)-1.0);
    bool merge_disconnected = ptree.get("geneo.merge_disconnected",(bool)false);
    int nthreads_setup = ptree.get<int>("geneo.nthreads_setup");
    int nthreads_solve = ptree.get<int>("geneo.nthreads_solve");
    size_t shrink = ptree.get<size_t>("geneo.shrink"); // shrink is only effective with distance based pu
    // std::string fineeigensolver = ptree.get<std::string>("geneo.fineeigensolver");
    std::string eigenproblem_type = ptree.get<std::string>("geneo.eigenproblem");

    // Parameters for fine-level eigensolver
    EigensolverParameters eigensolver_params;
    eigensolver_params.type = ptree.get("fine_eigensolver.type", "arpack");
    eigensolver_params.nev_target = ptree.get<size_t>("fine_eigensolver.nev_target", 10);
    eigensolver_params.nev_max = ptree.get<size_t>("fine_eigensolver.nev_max", 10);
    eigensolver_params.shift = ptree.get<double>("fine_eigensolver.shift", 0.001);
    eigensolver_params.tolerance = ptree.get<double>("fine_eigensolver.tolerance", 0.001);
    eigensolver_params.threshold = ptree.get<double>("fine_eigensolver.threshold", 0.5);
    eigensolver_params.regularisation = ptree.get<double>("fine_eigensolver.regularisation", 0.0);
    eigensolver_params.a = ptree.get<double>("fine_eigensolver.search_interval_left", 0.0);
    eigensolver_params.b = ptree.get<double>("fine_eigensolver.search_interval_right", 10.0);
#endif
//=====================
// extended Nicolaides
//=====================
#ifdef USEEXTNIC
    // partition the elements
    bool coord_partitioning = ptree.get("extendednicolaides.coordinate_partitioning",(bool)false);
    std::vector<size_t> parts = ptree.get<std::vector<size_t>>("extendednicolaides.parts");

    std::vector<size_t> partition;
    if (coord_partitioning)
      partition = coordinate_partitioning(gv,parts);
    else
      partition = parmetis_partitioning(gv,helper,subdomains[0]);

    // generate overlap: need to decide on size of overlap as well as extension method
    unsigned overlap = ptree.get("extendednicolaides.overlap",(unsigned)1);
    std::string extensionmethod = ptree.get("extendednicolaides.extensionmethod","vertex");
    auto overlappingsubdomains = grow_subdomains(gv,subdomains[0],partition,overlap,extensionmethod);

    // test overlap generation
    std::vector<unsigned> k0(indexset.size(0));
    size_t k_0 = 0;
    unsigned count_elements=0;
    for (const auto& e : elements(gv))
      {
        k0[indexset.index(e)] = overlappingsubdomains[indexset.index(e)].size();
        k_0 = std::max(k_0,overlappingsubdomains[indexset.index(e)].size());
        count_elements += overlappingsubdomains[indexset.index(e)].size();
      }
    std::cout << k_0 << " k_0" << std::endl;
    std::cout << indexset.size(0) << " elements, " << count_elements << " in all subdomains, ratio=" << ((double)count_elements)/indexset.size(0) << std::endl;

    disc.go->jacobian(x, A);

    //===================
    // build subdomain information
    //==================

    std::cout << "building subdomain information" << std::endl;
    using GFS = typename Discretization::GFS;
    using FLSDI = FineLevelSubdomainInformation<GFS>;
    FLSDI flsdi(*disc.gfs,subdomains[0],overlappingsubdomains);
    unsigned summed_dofs=0;
    for (unsigned i=0; i<flsdi.global_to_subdomains->size(); i++)
      summed_dofs += (*flsdi.global_to_subdomains)[i].size();
    std::cout << flsdi.global_to_subdomains->size() << " dof blocks, " << summed_dofs << " dof blocks in all subdomains, ratio=" << ((double)summed_dofs)/flsdi.global_to_subdomains->size() << std::endl;

    unsigned dofspersubdomain=0;
    for (unsigned p=0; p<subdomains[0]; ++p)
      if ((*flsdi.local_to_global)[p].size()>dofspersubdomain) dofspersubdomain = (*flsdi.local_to_global)[p].size();
    std::cout << "maximum number of dofs (blocks) per subdomain " << dofspersubdomain << std::endl;

    // still need to set constraints in subdomain matrices
    // we do the same trick: make a coefficient vector indicating where the constraints are
    std::cout << "assembling dirichlet constraints" << std::endl;
    VEC dirichlet(*disc.gfs);
    dirichlet = 1.0;
    set_constrained_dofs(*disc.cc, 0.0, dirichlet); // dirichlet dofs are zero, others are 1, so we can use it as a mask!
    std::shared_ptr<ISTLV> pistldirichlet = stackobject_to_shared_ptr(Dune::PDELab::Backend::native(dirichlet));

    //===================
    // now set up the preconditioner
    //==================

    // define number of coarse vectors per subdomain; must be >0
    // these are the constants for convection-diffusion
    int nvec = 1;
    std::vector<std::shared_ptr<VEC>> coarse_vectors(nvec);
    for (int i=0; i<nvec; i++) coarse_vectors[i] = std::shared_ptr<VEC>(new VEC(*disc.gfs,0.0));
    std::vector<std::shared_ptr<ISTLV>> istl_coarse_vectors(nvec);
    for (int i=0; i<nvec; i++) istl_coarse_vectors[i] = stackobject_to_shared_ptr(Dune::PDELab::Backend::native(*coarse_vectors[i]));

    auto constlambda = [](const auto& x){ return 1.0;};
    auto constfunc = Dune::PDELab::makeGridFunctionFromCallable(gv,constlambda);
    Dune::PDELab::interpolate(constfunc,*disc.gfs,*coarse_vectors[0]);

    if (nvec>=2)
      {
        auto xlambda = [](const auto& x){ return x[0];};
        auto xfunc = Dune::PDELab::makeGridFunctionFromCallable(gv,xlambda);
        Dune::PDELab::interpolate(xfunc,*disc.gfs,*coarse_vectors[1]);
      }
    if (nvec>=3)
      {
        auto ylambda = [](const auto& x){ return x[1];};
        auto yfunc = Dune::PDELab::makeGridFunctionFromCallable(gv,ylambda);
        Dune::PDELab::interpolate(yfunc,*disc.gfs,*coarse_vectors[2]);
      }
    if (nvec>=4 && dim==3)
      {
        auto zlambda = [](const auto& x){ return x[2];};
        auto zfunc = Dune::PDELab::makeGridFunctionFromCallable(gv,zlambda);
        Dune::PDELab::interpolate(zfunc,*disc.gfs,*coarse_vectors[3]);
      }

    int nthreads_setup = ptree.get<int>("extendednicolaides.nthreads_setup");
    int nthreads_solve = ptree.get<int>("extendednicolaides.nthreads_solve");
    std::string cycle = ptree.get("extendednicolaides.cycle","single");
#endif
#if !defined(USEMLGENEO) && !defined(USEEXTNIC)
    disc.go->jacobian(x,A);
#endif
    Dune::MatrixAdapter<ISTLM,ISTLV,ISTLV> op(istlA);
#ifdef USEMLGENEO
    std::cout << "creating multilevel geneo preconditioner" << std::endl;
    MultiLevelGenEO<ISTLM,ISTLV> prec(Dune::MPIHelper::getLocalCommunicator(),
                                      nthreads_setup,
                                      nthreads_solve,
                                      pistlA,
                                      pistlA_plus,
                                      pistldirichlet_mask,
                                      pAvolume_snippet_matrices,
                                      pAskeleton_snippet_matrices,
                                      pA_plus_volume_snippet_matrices,
                                      pA_plus_skeleton_snippet_matrices,
                                      pdddm,
                                      subdomains,
                                      pum,
                                      fineGEVPrhs,
                                      coarseGEVPrhs,
                                      shrink,
                                      coarse_overlap,
                                      eigenproblem_type,
                                      eigensolver_params,
                                      coarseeigensolver,
                                      n_eigenvectors_coarse_computed,n_eigenvectors_coarse_used,
                                      eigenvalue_coarse_threshold,
                                      abs_zero_ker,
                                      regularization_ker,
                                      merge_disconnected,
                                      cycle,
                                      mlgeneo_verbose);
#endif
#ifdef USEEXTNIC
    std::cout << "constructing extended Nicolaides preconditioner" << std::endl;
    ExtendedNicolaides<ISTLM,ISTLV> prec(pistlA,
                                        istl_coarse_vectors,
                                        pistldirichlet,
                                        flsdi.local_to_global,
                                        flsdi.global_to_local,
                                        cycle,
                                        nthreads_setup,
                                        nthreads_solve);
#endif
#ifdef USEILU
    size_t ilun = ptree.get("ilu.n",(size_t)0);
    double iluw = ptree.get("ilu.w",(double)1.0);
    Dune::SeqILU<ISTLM,ISTLV,ISTLV> prec(istlA,ilun,iluw,false);
#endif
#ifdef USENOPREC
    Dune::Richardson<ISTLV,ISTLV> prec;
#endif

    size_t itmax = ptree.get("solver.itmax",(size_t)100);
    std::string solvertype = ptree.get("solver.type","gmres");
    double reduction = ptree.get<double>("solver.reduction",(double)1e-8);
    Dune::CGSolver<ISTLV> cgsolver(op,prec,reduction,itmax,2);
    size_t restart = ptree.get("solver.restart",(size_t)50);
    Dune::RestartedGMResSolver<ISTLV> gmressolver(op,prec,reduction,restart,itmax,2);
    Dune::LoopSolver<ISTLV> loopsolver(op,prec,reduction,100,2);
    Dune::InverseOperatorResult stat;
    ISTLV istlv(istlx);
    istlv = (double)0.0;
    if (solvertype=="gmres")
      gmressolver.apply(istlv,istlr,stat);
    else
      loopsolver.apply(istlv,istlr,stat);
    istlx -= istlv;
    std::cout << "LOG10 " << std::log10(stat.conv_rate) << " " << -10.0/std::log10(stat.conv_rate) <<  std::endl;

    //===================
    // visualization
    //==================

#ifdef USEMLGENEO
    // generate some vectors for visualization
    std::vector<std::vector<unsigned>> k(subdomains.size());
    for (auto& v : k) v.resize(indexset.size(0));
    size_t k_0 = 0;
    unsigned count_elements=0;
    for (size_t i=0; i<indexset.size(0); ++i)
      {
        for (size_t l=0; l<subdomains.size(); l++)
          k[l][i] = prec.get_overlappingsubdomains(l)[i].size();
        k_0 = std::max(k_0,prec.get_overlappingsubdomains(0)[i].size());
        count_elements += prec.get_overlappingsubdomains(0)[i].size();
      }
    std::cout << k_0 << " k_0" << std::endl;
    std::cout << indexset.size(0) << " elements, " << count_elements << " in all subdomains, ratio=" << ((double)count_elements)/indexset.size(0) << std::endl;
#endif

    bool save_output = ptree.get("output.save",(bool)true);
    if (!save_output) return 0; // then we are done

    // visualize result
    std::string outfile = ptree.get("output.filename","output");
    Dune::VTKWriter<GV> vtkwriter(gv,Dune::VTK::nonconforming);
    using DGF = typename Discretization::DGF;
    DGF dgfx(*disc.gfs,x);
    using VTKF = Dune::PDELab::VTKGridFunctionAdapter<DGF>;
    vtkwriter.addVertexData(std::shared_ptr<VTKF>(new VTKF(dgfx,"fesol")));

    // true solution (only for rotating pulse!)
    VEC sol_vec(*disc.gfs);
    auto sol_lambda = [&](const auto& e, const auto& x){
      return problem.g(e,x);
    };
    auto sol = Dune::PDELab::makeGridFunctionFromCallable(gv,sol_lambda);
    Dune::PDELab::interpolate(sol,*disc.gfs,sol_vec);
    DGF dgfsol(*disc.gfs,sol_vec);
    vtkwriter.addVertexData(std::shared_ptr<VTKF>(new VTKF(dgfsol,"sol")));

    // right-hand side
    VEC f_vec(*disc.gfs);
    auto f_lambda = [&](const auto& e, const auto& x){
      return problem.f(e,x);
    };
    auto f = Dune::PDELab::makeGridFunctionFromCallable(gv,f_lambda);
    Dune::PDELab::interpolate(f,*disc.gfs,f_vec);
    DGF dgff(*disc.gfs,f_vec);
    vtkwriter.addVertexData(std::shared_ptr<VTKF>(new VTKF(dgff,"rhs")));

    // diffusion coefficient
    VEC D_vec(*disc.gfs);
    auto D_lambda = [&](const auto& e, const auto& x){
      return problem.A(e,x)[0][0];
    };
    auto D = Dune::PDELab::makeGridFunctionFromCallable(gv,D_lambda);
    Dune::PDELab::interpolate(D,*disc.gfs,D_vec);
    DGF dgfD(*disc.gfs,D_vec);
    vtkwriter.addVertexData(std::shared_ptr<VTKF>(new VTKF(dgfD,"diff.coeff.")));

    // velocity field
    /*using VDGF = typename Discretization::VDGF;
    using VVTKF = Dune::PDELab::VTKGridFunctionAdapter<VDGF>;
    typename Discretization::VVEC b_vec(*disc.vgfs);
    auto b_lambda = [&](const auto& e, const auto& x){
      return problem.b(e,x);
    };
    auto b = Dune::PDELab::makeGridFunctionFromCallable(gv,b_lambda);
    Dune::PDELab::interpolate(b,*disc.vgfs,b_vec);
    VDGF dgfb(*disc.vgfs,b_vec);*/

    using Entity = typename GV::template Codim<0>::Entity;
    struct VectorFunction : Dune::VTKFunction<GV> {
      VectorFunction(Problem* problem) : Dune::VTKFunction<GV>{}, _problem{problem} {}
      int ncomps () const {return DIMENSION;};
      double evaluate (int comp, const Entity& e,
                              const Dune::FieldVector<double,dim>& xi) const{
                                return _problem->b(e, xi)[comp];
                              };
      //! get name
      virtual std::string name () const {return "b";};
      //! virtual destructor
      virtual ~VectorFunction () {}

      Problem* _problem;
    };
    vtkwriter.addVertexData(std::make_unique<VectorFunction>(&problem));

#ifdef USEMLGENEO
    vtkwriter.addCellData(pdd->partition,"partition 0");
    vtkwriter.addCellData(pdd->vizpartition,"vizpartition 0");
    vtkwriter.addCellData(prec.get_partition(1),"partition 1");
    vtkwriter.addCellData(prec.get_vizpartition(1),"vizpartition 1");
    for (size_t l=0; l<subdomains.size(); l++)
      {
        std::ostringstream s1;
        s1 << "volumesnippetnumber" << l ;
        vtkwriter.addCellData(prec.get_element_to_volumesnippetnumber(l),s1.str());
      }
    for (size_t l=0; l<subdomains.size(); l++)
      {
        std::ostringstream s1;
        s1 << "k" << l ;
        vtkwriter.addCellData(k[l],s1.str());
      }
    bool vizbasisvectors = ptree.get("geneo.vizbasisvectors",false);
    std::vector<std::vector<size_t>> viz_vectors;
    for (int level=0; level<1; level++)
      for (int p=0; p<std::min((size_t)1024,prec.number_of_subdomains(level)); p++)
        for (int i=0; i<std::min((size_t)20,prec.number_of_basis_vectors(level,p)); i++)
          {
            std::vector<size_t> v; v.push_back(level);  v.push_back(p);  v.push_back(i);
            viz_vectors.push_back(v);
          }
    int n_bv=0;
    if (vizbasisvectors)
      n_bv = viz_vectors.size();
    std::vector<std::shared_ptr<VEC>> basisvectors(n_bv); // empty vector
    std::vector<std::shared_ptr<DGF>> basisvectordgfs(n_bv); // the discrete grid functions
    std::vector<std::string> basisvectornames(n_bv); // the names
    if (vizbasisvectors)
      for (int i=0; i<n_bv; i++)
        {
          int l = viz_vectors[i][0];
          int p = viz_vectors[i][1];
          int k = viz_vectors[i][2];
          basisvectors[i] = std::shared_ptr<VEC>( new VEC(*disc.gfs,0.0) ); // add a vector
          Dune::PDELab::Backend::native(*basisvectors[i]) = prec.prolongated_basis_vector(l,p,k); // fill value
          basisvectordgfs[i] = std::shared_ptr<DGF>( new DGF(*disc.gfs,*basisvectors[i]) );
          std::ostringstream s1;
          s1 << "basisvector_" << l << "_" << p << "_" << k;
          basisvectornames[i] = s1.str();
        }
    for (int i=0; i<basisvectors.size(); i++)
      {
        std::cout << "add output for " << basisvectornames[i] << std::endl;
        vtkwriter.addVertexData(std::shared_ptr<VTKF>(new VTKF(*basisvectordgfs[i],basisvectornames[i])));
      }
#endif
    vtkwriter.write(outfile,Dune::VTK::appendedraw);
#else
    std::cout << "Needs UGGrid" << std::endl;
#endif
    return 0;
  }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
  }
}
