// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

// always include the config file
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
// C++ includes
#include<math.h>
#include<iostream>
#include<vector>
#include<set>
#include<sstream>
// dune-common includes
#include<dune/common/parallel/mpihelper.hh>
#include<dune/common/parametertreeparser.hh>
#include<dune/common/timer.hh>
#include <dune/common/exceptions.hh> // We use exceptions
// dune-istl includes
#include<dune/istl/preconditioners.hh>
// dune-geometry includes
#include<dune/geometry/referenceelements.hh>
#include<dune/geometry/quadraturerules.hh>
// dune-grid includes
#include<dune/grid/onedgrid.hh>
#include<dune/grid/io/file/vtk/vtkwriter.hh>
#include<dune/grid/io/file/gmshreader.hh>
#if HAVE_UG || HAVE_DUNE_UGGRID
#include<dune/grid/uggrid.hh>
#endif
#if HAVE_DUNE_ALUGRID
#include<dune/alugrid/grid.hh>
#include<dune/alugrid/dgf.hh>
#include<dune/grid/io/file/dgfparser/dgfparser.hh>
#endif
// dune-pdelab includes
#include<dune/pdelab/common/function.hh>
#include<dune/pdelab/common/vtkexport.hh>
#include<dune/pdelab/function/callableadapter.hh>
#include<dune/pdelab/gridfunctionspace/vtk.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include<dune/pdelab/constraints/common/constraints.hh>
#include<dune/pdelab/boilerplate/pdelab.hh>
// ftworth includes
#include"../dune/ftworth/partitioner.hh"
#include"../dune/ftworth/assemblewrapper.hh"
#include"../dune/ftworth/subdomainutilities.hh"
#include"../dune/ftworth/cdediscretizer.hh"
#include"../dune/ftworth/schwarz_preconditioners.hh"

#include"../dune/ftworth/matrixutilities.hh"
// includes from this directory
#include"poisson.hh"
#include"heterogeneous_problem.hh"

int main(int argc, char** argv)
{
  try{
    // Maybe initialize MPI
    Dune::MPIHelper& helper = Dune::MPIHelper::instance(argc, argv);
    std::cout << "Hello World! This is dune-ftworth." << std::endl;
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout<<"I am rank "<<helper.rank()<<" of "<<helper.size()
               <<" processes!"<<std::endl;

    // Read parameters from ini file
    Dune::ParameterTree ptree;
    Dune::ParameterTreeParser ptreeparser;
    ptreeparser.readINITree("extendednicolaides.ini",ptree);
    ptreeparser.readOptions(argc,argv,ptree);

#if (HAVE_UG || HAVE_DUNE_UGGRID) // check if we have the UG grid manager

    //===================
    // grid generation and partitioning
    //===================
    
    // load a grid file and refine
    std::string mshfile = ptree.get("grid.filename","circle.msh");
    using Grid = Dune::UGGrid<2>;  
    std::shared_ptr<Grid> pgrid;
    if (helper.rank()==0)
      // the third argument must be false, otherwise it does not work ... not documented!
      pgrid = std::shared_ptr<Grid>(Dune::GmshReader<Grid>::read(mshfile,true,false));
    
    // refine grid
    int refine = ptree.get("grid.refine",(int)0);
    pgrid->globalRefine(ptree.get("grid.refine",refine));

    // // create structured grid
    // typedef Dune::YaspGrid<2> GM;
    // typedef Dune::PDELab::StructuredGrid<GM> Grid;
    // Grid pgrid(Dune::GeometryType::cube,16,1);
    // pgrid->globalRefine(2);

    // extract the leaf grid view
    auto gv = pgrid->leafGridView();
    using GV = decltype(gv); // and get the type
    const int dim = GV::dimension;
    auto& indexset = gv.indexSet(); // to attach data to elements
    std::vector<unsigned> subdomains = ptree.get<std::vector<unsigned>>("extendednicolaides.subdomains");
    std::cout << indexset.size(0) << " elements" << std::endl;
    std::cout << indexset.size(dim) << " vertices" << std::endl;
    std::cout << subdomains[0] << " subdomains" << std::endl;

    // partition the elements using ParMetis
    auto partition = parmetis_partitioning(gv,helper,subdomains[0]);

    // generate overlap: need to decide on size of overlap as well as extension method
    unsigned overlap = ptree.get("extendednicolaides.overlap",(unsigned)1);
    std::string extensionmethod = ptree.get("extendednicolaides.extensionmethod","vertex");
    auto overlappingsubdomains = grow_subdomains(gv,subdomains[0],partition,overlap,extensionmethod);

    // test overlap generation
    std::vector<unsigned> k0(indexset.size(0));
    size_t k_0 = 0;
    unsigned count_elements=0;
    for (const auto& e : elements(gv))
      {
        k0[indexset.index(e)] = overlappingsubdomains[indexset.index(e)].size();
        k_0 = std::max(k_0,overlappingsubdomains[indexset.index(e)].size());
        count_elements += overlappingsubdomains[indexset.index(e)].size();
      }
    std::cout << k_0 << " k_0" << std::endl;
    std::cout << indexset.size(0) << " elements, " << count_elements << " in all subdomains, ratio=" << ((double)count_elements)/indexset.size(0) << std::endl;

    //===================
    // set up a discretization scheme
    //===================

    // some important types
    using DF = Grid::ctype;
    using RF = double;

    // set up the problem parameters
    using Problem = PoissonProblem<GV,RF>;
    // using Problem = HeterogeneousProblem<GV,RF>;    
    Problem problem;

    // set up the discretization
    //using Discretization = DGConvectionDiffusionProblemSimplex<GV,Problem,1>;
    using Discretization = CGConvectionDiffusionProblemSimplex<GV,Problem,1>;
    Discretization disc(gv,problem,ptree);

    //===================
    // assemble the global problem
    //===================

    // set up linear system in PDELab components
    using VEC = typename Discretization::VEC;
    using MAT = typename Discretization::MAT;

    VEC x(*disc.gfs);
    Dune::PDELab::interpolate(*disc.g,*disc.gfs,x);
    // choose random initial guess (ok this overwrites)
    for (size_t i=0; i<Dune::PDELab::Backend::native(x).N(); i++)
      for (size_t j=0; j<Dune::PDELab::Backend::native(x)[i].N(); j++)
        Dune::PDELab::Backend::native(x)[i][j] = (rand() % 10000)/10000.0;
    set_constrained_dofs(*disc.cc,0.0,x); // dirichlet dofs are zero, others are 1, so we can use it as a mask!

    // set up global stiffness matrix and resdual vector
    // now it is sufficient to do standard assembly !
    Dune::Timer timer_total;
    timer_total.reset();
    MAT A(*disc.go);
    A = 0.0;
    std::cout << "assemble matrix and residual" << std::endl;
    disc.go->jacobian(x,A); // this assembles the matrix
    VEC r(*disc.gfs);
    r = 0.0;
    disc.go->residual(x,r);
    std::cout << "took " << timer_total.elapsed() << " seconds" << std::endl;

    // extract ISTL objects
    using ISTLV = Dune::PDELab::Backend::Native<VEC>;
    ISTLV& istlx = Dune::PDELab::Backend::native(x);
    ISTLV& istlr = Dune::PDELab::Backend::native(r);
    using ISTLM = Dune::PDELab::Backend::Native<MAT>;
    ISTLM& istlA = Dune::PDELab::Backend::native(A);
    std::shared_ptr<ISTLM> pistlA = stackobject_to_shared_ptr(Dune::PDELab::Backend::native(istlA));

    //===================
    // build subdomain information
    //==================

    std::cout << "building subdomain information" << std::endl;
    using GFS = typename Discretization::GFS;
    using FLSDI = FineLevelSubdomainInformation<GFS>;
    FLSDI flsdi(*disc.gfs,subdomains[0],overlappingsubdomains);
    unsigned summed_dofs=0;
    for (unsigned i=0; i<flsdi.global_to_subdomains->size(); i++)
      summed_dofs += (*flsdi.global_to_subdomains)[i].size();
    std::cout << flsdi.global_to_subdomains->size() << " dof blocks, " << summed_dofs << " dof blocks in all subdomains, ratio=" << ((double)summed_dofs)/flsdi.global_to_subdomains->size() << std::endl;

    unsigned dofspersubdomain=0;
    for (unsigned p=0; p<subdomains[0]; ++p)
      if ((*flsdi.local_to_global)[p].size()>dofspersubdomain) dofspersubdomain = (*flsdi.local_to_global)[p].size();
    std::cout << "maximum number of dofs (blocks) per subdomain " << dofspersubdomain << std::endl;

    //===================
    // process dirichlet constraints
    //==================

    // still need to set constraints in subdomain matrices
    // we do the same trick: make a coefficient vector indicating where the constraints are
    std::cout << "assembling dirichlet constraints" << std::endl;
    VEC dirichlet(*disc.gfs);
    dirichlet = 1.0;
    set_constrained_dofs(*disc.cc,0.0,dirichlet); // dirichlet dofs are zero, others are 1, so we can use it as a mask!
    std::shared_ptr<ISTLV> pistldirichlet = stackobject_to_shared_ptr(Dune::PDELab::Backend::native(dirichlet));
            
    //===================
    // now set up the preconditioner
    //==================

    // make coarse grid vectors for two level method
    int nvec = 3; // define number of coarse vectors per subdomain; must be >0
    std::vector<std::shared_ptr<VEC>> coarse_vectors(nvec);
    for (int i=0; i<nvec; i++) coarse_vectors[i] = std::shared_ptr<VEC>(new VEC(*disc.gfs,0.0));
    std::vector<std::shared_ptr<ISTLV>> istl_coarse_vectors(nvec);
    for (int i=0; i<nvec; i++) istl_coarse_vectors[i] = stackobject_to_shared_ptr(Dune::PDELab::Backend::native(*coarse_vectors[i]));
      
    auto constlambda = [](const auto& x){ return 1.0;};
    auto constfunc = Dune::PDELab::makeGridFunctionFromCallable(gv,constlambda);
    Dune::PDELab::interpolate(constfunc,*disc.gfs,*coarse_vectors[0]);

    if (nvec>=2)
      {
        auto xlambda = [](const auto& x){ return x[0];};
        auto xfunc = Dune::PDELab::makeGridFunctionFromCallable(gv,xlambda);
        Dune::PDELab::interpolate(xfunc,*disc.gfs,*coarse_vectors[1]);
      }
    if (nvec>=3)
      {
        auto ylambda = [](const auto& x){ return x[1];};
        auto yfunc = Dune::PDELab::makeGridFunctionFromCallable(gv,ylambda);
        Dune::PDELab::interpolate(yfunc,*disc.gfs,*coarse_vectors[2]);
      }
    if (nvec>=4 && dim==3)
      {
        auto zlambda = [](const auto& x){ return x[2];};
        auto zfunc = Dune::PDELab::makeGridFunctionFromCallable(gv,zlambda);
        Dune::PDELab::interpolate(zfunc,*disc.gfs,*coarse_vectors[3]);
      }
 
    // set up an ISTL solver and solve
    int nthreads_setup = ptree.get<int>("extendednicolaides.nthreads_setup");
    int nthreads_solve = ptree.get<int>("extendednicolaides.nthreads_solve");
    std::string cycle = ptree.get("extendednicolaides.cycle","single");
    int L = subdomains.size();
    Dune::MatrixAdapter<ISTLM,ISTLV,ISTLV> op(istlA);
    ExtendedNicolaides<ISTLM,ISTLV> prec(pistlA,istl_coarse_vectors,pistldirichlet,flsdi.local_to_global,flsdi.global_to_local,cycle,nthreads_setup,nthreads_solve);
    //Dune::SeqILU<ISTLM,ISTLV,ISTLV> prec(istlA,1.0);
    //Dune::Richardson<ISTLV,ISTLV> prec;
    //Dune::CGSolver<ISTLV> solver(op,prec,1e-8,100,2,true);
    Dune::RestartedGMResSolver<ISTLV> solver(op,prec,1e-8,50,2500,2);
    Dune::InverseOperatorResult stat;
    ISTLV istlv(istlx);
    istlv = (double)0.0;
    solver.apply(istlv,istlr,stat);
    istlx -= istlv;

    //===================
    // visualization
    //==================
    return 0;
    
    // visualize result
    using DGF = typename Discretization::DGF;
    DGF dgfx(*disc.gfs,x);
    std::string outfile = ptree.get("output.filename","output");
    Dune::VTKWriter<GV> vtkwriter(gv,Dune::VTK::nonconforming);
    vtkwriter.addCellData(partition,"partition");
    vtkwriter.addCellData(k0,"k0");
    using VTKF = Dune::PDELab::VTKGridFunctionAdapter<DGF>;
    vtkwriter.addVertexData(std::shared_ptr<VTKF>(new VTKF(dgfx,"fesol")));
    std::cout << "writing vtk file" << std::endl; 
    vtkwriter.write(outfile,Dune::VTK::appendedraw);
#else
    std::cout << "Needs UGGrid" << std::endl;
#endif
    return 0;
  }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
  }
}
